;
;                                 FSS.ASM
;         "Faden Spannungs Sensor Programm" based on:
;          -MpK 0.8- with interrupt handler for PIC16LC71
;
;*******************************************************************
;
;Creator : M.Kellenberger, 18.11.93, NTB-Buchs CH
;System  : Boxer Editor, MPALC, MPSTART
;
;History MpK:
;V 0.0    21.10.93 mik        Systemlabel start with "_"!
;V 0.1    22.10.93 mik        System design and description
;V 0.1a   25.10.93 mik        Add "#" to user-changeable lines
;V 0.1b   15.11.93 mik        Made systembody faster (20MHz)
;V 0.1c   16.11.93 mik        MpK Kernel via simulater tested!
;V 0.2    25.11.93 mik        RxError, RxBusy and Cmds!
;V 0.3    13. 1.94 mik        RS-484 Enable implemented
;V 0.4    21. 1.94 mik        Compiler flags _Tst,_MHz,_RS implemented
;V 0.5    27. 1.94 mik        Compiler flags _Tst into TstDbg implemented
;V 0.6    28. 1.94 mik        Compiler flags _WDT implemented
;V 0.7     8. 2.94 mik        Ext_Cmd is 001T'0000 (Flag T is no also for EXT Cmds)
;         10. 2.94 mik        reset condition into Cmd
;         16. 2.94 mik        _TxInv, _RxInv implemented
;V 0.8    21. 2.94 mik        _Peven, _Parity deleted
;
;User programmers revision:
;V 1.0    18.11.93 mik        First attempt
;V 1.0b   25.11.93 mik        Test run, programmed into PIC "FSS V1.0b",OK
;V 1.1    26.11.93 mik        Cmd_Hdr,Address,Cmd Select added
;V 1.2    30.11.93 mik        Cmd-Process, SubInpExe, Output Sub skeleton
;                             and IntAD added
;V 1.3    01.12.93 mik        Daisy Chain, FGS-Cmd, Shoot imppplemented
;                             programmed into PIC "FSS V1.3/B/nP", OK
;V 1.4    02.12.93 mik        Reduced to simple prototyp I,Cmd Sht implemented
;         07.12.93 mik        Multiplier at GetBufW1 added, bugs fixed
;         08.12.93 mik        Bugs fixed, Cmd SOG implemented
;         09.12.93 mik        Cmd GDX implemented
;         16.12.93 mik        Cmd GER, Cut implemented
;V 1.5    11.01.94 mik        Pin EnTx,Extended commands, SNV added and GER moved
;         17.01.94 mik        Benninger Protokoll completed with tail
;         18.01.94 mik        Redesign of code, Test Sht
;         19.01.94 mik        Redesign of code, Test GDX-OV/M
;         20.01.94 mik        GBufW, _IntAD corrected, Test GDX-AV/M,
;         21.01.94 mik        _IntRTCC corrected, Test GDX-AV/M,
;         24.01.94 mik        _IntAD corrected, Test SOG,SNV,EXT-GER
;         25.01.94 mik        Test DChPin,EXT-GER,LA-GER
;                             Test Sulzer: Sht,GER
;         26.01.94 mik        programmed into PIC
;V 1.6    31.01.94 mik        several errors corrected, SCX implemented
;          3. 2.94 mik        24MHz clock, HexOutput for test implemented
;          7. 2.94 mik        Hex, programmed
;V 1.7     9. 2.94 mik        Ext, CmdTst-Flag ,Dbg corrected
;         10. 2.94 mik        Invers A/D input, programmed
;V 1.8    11. 2.94 mik        Restart of AD after every Cmd,Cut into bits,
;                             SCX inverts ADRES, _Inv added, ZRO implemented
;         10. 2.94 mik        programmed
;V 1.8l   31. 3.94 mik        Test programmed to 16MHz
;V 1.9     6. 4.94 mik        ChNivSel corrected, not invert,
;          8. 8.94 mik        Listen All branch corrected
;           . 8.94 mik        programmed
;
;*******************************************************************
;
; Description:
; ============
;   1. Multi process Kernel System body:
;   ------------------------------------
;         The MpK consists of
;           predefined register- and macrosets
;           initialization of HW and status/option registers
;           one (empty) MAIN (user) process,
;           an INTerruptHANDLER, which branches to the 4
;           handling processes of each interrupt source.
;           RTCC-int handler cycles through 4 processes: Tx, Rx0, ?? and Rx1.
;           and may launch the AD-conversion repeatedly.
;
;         All this empty modules may be filled by the user
;         (programmer) and described in the following paragraph!
;
;    2. FSS
;    ------
;         This program is design to sample 8 sensores via 8 resistore bridges,
;         multiplexed trough a analog switcher, amplified in one integrated
;         op-amp and converted to 8bit at one AD-input with reference.
;
;         The sampled data's are stored in a simple buffer (due to lake of more memory)
;         and calculated with offset and gain. The mean od the data is also calcable.
;         The PIC controller is connected to the machine by 2 (RS485:Benninger) or
;         4 (RS422:Sulzer) wired lines via a bus driver. All this components (plus supply)
;         are located close to the 8 sensors.
;
;       Calculation:
;       -----------
;         The sample rate is about 2kHz for every sensor. There are two ways of collecting
;         the samples:
;           Value:  The samples are stored in one buffer for every sensor as they are (8bit).
;                   To send this value to the host, the offset is subtracted and a gain is
;                   multiplied to it.
;           Mean:   The same way as above, but the sample is added to the old buffer value
;                   and then divided by 2 at storing. This causes a mean value of every sensor
;                   sample with descending weight. The most recent sample has most weight (half)
;                   and the 9th sample behind has 1/1024th weight.
;
;       Protocoll:
;       ---------
;         The protocoll is designed for two ways of connection:
;           Sulzer: A 4-wire-line is used for this point-to-point connction (RS-422 or
;                   RS-232(C)) Only one FSS is at the line and it needs no addressing and no
;                   cmd tailing. If there is a new Cmd sending down, the last respond of the FSS
;                   will be cutted off without a final ACK! This procedure must be used for the
;                   Sht-X Cmd, because it does send data endlessly. Every occured error causes
;                   a NACK at any time during the protocoll. This may be used to detect transmission
;                   and "cut" error immediatly.
;           Benninger: A 2-wire-bidirectional line is used for this bus connection
;                   (RS-485). Up to 100 FSSs are at the line and it needs addressing and cmd tailing.
;                   Every message must end with a final ACK or NACK! Only errors which occure
;                   during addressed response do cause a NACK directly. At any other
;                   (unaddressed) time a transmission error is ignored and a "cut" error
;                   pulles the daisy chain line to request a LA-GER command.
;
;       Commands:
;       ---------
;         Sht-X:    Get Endlessly the value of sensor X (until next Cmd).(Shoot;only Sulzer)
;
;         GDX-OV[S]:Get value of one sensor.  (Get data one value; S<8)
;         GDX-OV[S]:Get value of one offset.  (Get data one value; S>7)
;
;         GDX-OM[S]:Get mean  of one sensor.  (Get data one mean;  S<8)
;         GDX-OM[S]:Get value of one gain.    (Get data one mean;  S>7)
;
;         GDX-AV:   Get values of all sensors.(Get data all value)
;         GDX-AM:   Get "mean" of all sensors.(Get data all mean)
;                   This cmd is not correct!
;
;         SOG[V]:   Sets offset or gain to value V. (Set Offset/Gain)
;
;         SNV[N]:   Set new niveau N to detect cut-slope. (Set niveau)
;
;         EXT-SCX:  Sets channel x permanently or resets to normal operation.(Extended cmd set channel)
;
;         EXT-GER:  Gets last (two) cut-errored sensor(s).(Extended cmd get error)
;
;         EXT-ZRO:  Sets Offset to read values and gain 1.00.(Extended zero)
;
;         LA-GER:   Gets last (two) cut-errored, daisy-chained sensor(s).(Listen all cmd get error)
;
;
;       Remark:
;       -------
;         Compiler vars are Firma and P2nd
;         Firma = 'S' is ment for Sulzer version: One FSS at ser. IF
;               = 'B' is ment for Benninger vers.: up to 128 FSS at ser. IF
;         P2nd  = 'No' is ment for the first programming attempt.
;                    All offsets/gains are FFh. Resolve correct values via SW.
;               = 'Yes'is ment for the second progr. att.:
;                    Offset/Gain may be overwritten by the right values.
;
;       Future (ST6265):
;       ----------------
;         - Niveau = 1Byte              -> EEPROM  1B
;         - Offset/Gain                 -> EEPROM 16B
;         - Lauferkennung: Rippel detektieren
;            Lauf  = 1 Byte
;            Select= 1 Byte
;         - PCM ouput fr 8 Bremsen (8Pin)
;         - Adressierung per Software   -> EEPROM  1B
;         - Checksumme ins Protokol
;                  = 1 Byte
;         - Parity check
;                  = 1 Byte
;         - Linearisierung              -> EEPROM 8x128B
;         - CmdExt = 1Byte
;         -
;         - evtl. Verstrkungsumschaltung (Sulzer/Benninger) (1-2Pin)
;         - evtl. Nullpunktkorrektur Elektronik (FET ber Kannal 0)
;
; Notes to the source:
; --------------------
;         1. Labels with the starter "_" are system labels
;         2. # mark changeable values
;         3. Compiler/Assembler vars are:
;            _RS      = 232/485
;            _MHz     = 20,16,...
;            _Tst     = Yes/No
;            _WDT     = Yes/No
;            _RxInv   = Yes/No
;            _TxInv   = Yes/No
;            _ADInv   = Yes/No
;
;*******************************************************************
;
;Assembler options:
;

          LIST C=100,E=0,N=60,F=INHX8M,P=16C71,R=DEC,L=ON

Yes       set       1         ; Do use this for compiler directives!
No        set       0         ; Do not use "YES" or "NO" for them!

Firma     set       'S'       ; 'S' for Sulzer    (One at the line, def.=nomean calc)
                              ; 'B' for Benninger (Addressable, default=mean calc)

          IF        Firma='S'
_RS        set      232       ; 232 for point-to-point-connection via RS232 connection
          ELSE
_RS        set      485       ; 485 for multy-2-wire-RS485-bus with output pin TxEn
          ENDIF

_Tst      set       Yes       ; Yes for test sample: Adds 0Dh0Ah if testflag in Cmd is set
                              ; No for serie!

_RxInv    set       No        ; Yes for Rx Bits to be inverted
                              ; No for Rx Bits not to be inverted

_TxInv    set       No        ; Yes for Tx Bits to be inverted
                              ; No for Tx Bits not to be inverted

_ADInv    set       No        ; Yes for ADRES to be inverted
                              ; No for ADRES not to be inverted

_WDT      set       Yes       ; Yes for active WDT and reset sending code
                              ; No for no WDT!

P2nd      set       No        ; No for filling ff's into Gain and Offset
                              ; Yes for overwriting correct offset and gain values

_MHz      set       16        ; 24 for 24MHz quarz connection
                              ; 20 for 20MHz quarz connection
                              ; 16 for 16MHz quarz connection
;
;*******************************************************************
;
;
; ID VALUES                   ; Burn this within "MPSTART.EXE"
;
; 2000h   3FFFh               ; Code Revision Number
; 2001h   3FFFh               ; Code Manufacturing Number
; 2002h   3FFFh               ; Code ...
; 2003h   3FFFh               ; Code ...
;
;CONFIGURATION WORD: (2007H)
;
;Select   FOSC1     FOSC0     ; OSC Selection Fuses
;         0         0         ;  LP
;         0         1         ;  XT
;  >      1         0         ;  HS _MHz=20..5
;         1         1         ;  RC (default)
;
;         WDTE                ; WDT Enable Fuse
;  #      0                   ;  disable
;         1                   ;  enable (default)
;
;         PWRTE               ; Power Up Timer Fuse
;         0                   ;  disable
;  >      1                   ;  enable (default)
;
;         CP                  ; Code Protection Fuse
;         0                   ;  On
;  >      1                   ;  Off (default)
;
;
;*******************************************************************
;
;Includes
;

          INCLUDE   "PicReg.equ"
          INCLUDE   "Pic71.mac"

;
;*******************************************************************
;
;Constants
;

Presc     equ       b'000'              ; Prescaler Value   000=1:2
                                        ;                   001=1:4
                                        ;                      =
                                        ;                   111=1:256

                                        ; RTCC set value (minus corr. cycles for HW (2)delay)
                                        ; Calculation: (# for serial IF baudrate)
                                        ; ------------
          IF        _MHz=24
Rtcc       equ      (78)-2              ; ...for 24MHz
                                        ; >>>>>>>>> 24MHz t=0.16666us <<<<<<<<<<<<
                                        ; 78 *  0.167us= 13us
                                        ;  4 * 13us    = 52us => 19.2kBaud
          ENDIF
          IF        _MHz=20
Rtcc       equ      (65)-2              ; ...for 20MHz
                                        ; >>>>>>>>> 20MHz t=0.2us <<<<<<<<<<<<
                                        ; 65 *  0.2us  = 13us
                                        ;  4 * 13us    = 52us => 19.2kBaud
          ENDIF
          IF        _MHz=16
Rtcc       equ      (52)-2              ; ...for 16MHz
                                        ; >>>>>>>>> 16MHz t=0.25us <<<<<<<<<<<
                                        ;104 * 0.25us  = 26us
                                        ;  4 * 26us    =104us => 9.6kBaud

                                        ; 52 * 0.25us  = 13us
                                        ; ##
                                        ;  4 * 13us    = 52us => 19.2kBaud

          ELSE
                                        ; ...else ???
          ENDIF




Adj       equ       (0)+2               ; Adjust value (X * 0.2us plus 2)

SRate     equ       100h-(3)*10h        ; Sample rate for (multiplexed) AD-conversion
                                        ;
                                        ; Calculation:
                                        ; ------------
                                        ; Process timeslice t=26us
                                        ;  2 *  26us => (19,2kHz)
                                        ;  3 *  26us => (12,8kHz)
                                        ;  ##

                                        ; Process timeslice t=13us
                                        ;  4 *  13us => (19,2kHz)
                                        ;  5 *  13us => (15,4kHz)

                                        ; Process timeslice t=6.5us
                                        ;  8 *  6.5us => (19,2kHz)
                                        ; 10 *  6.5us => (15,4kHz)
                                        ; 11 *  6.5us => (13,9kHz)

PCnt0     equ       00h                 ; PStat Reg. Process counter bit 0
PCnt1     equ       01h                 ; PStat Reg. Process counter bit 1
RxFlag    equ       02h                 ; PStat Reg. 1=Rx is active
TxFlag    equ       03h                 ; PStat Reg. 1=Tx is active
                                        ; PStat Reg. Bit 4-7: Countdown for sample rate

DataRdy   equ       ADCnusd             ; DataReady Flag in reg. ADCON0:
                                        ; # Must be reset via user
                                        ;   when data is read !!!

RxStart   equ       00h                 ; TxRxFlg Reg. 1=Rec. start bit
RxStop    equ       01h                 ; TxRxFlg Reg. 1=Rec. stop bit
RxSel     equ       02h                 ; TxRxFlg Reg. Select Rx process 0=01,1=11
TxStart   equ       03h                 ; TxRxFlg Reg. 1=Tx is active
TxStop    equ       04h                 ; TxRxFlg Reg. 1=Tx is active
                                        ; TxRxFlg Reg. Bit 5-7: bit counter Tx

                                        ; Bits in STATUS (Main!) or SaveST (Interrupt routine!):
RxError   equ       RP1                 ; Flag to mark Rx Error start/stop/parity
RxBusy    equ       IRP                 ; Flag to mark Rx busy (INTterupt while DataRdy=1)

                                        ; FSS #:
Mult_Ofl  equ       RP1                 ; Flag to mark multiply overflow

                                        ; Port Pins:
                                        ; ----------

                                        ; PORTB:
RxPin     equ       00h                 ; Pin RS-232-C Receiver
TxPin     equ       01h                 ; Pin RS-232-C Transmitter
          IF        _RS=485
TxEn       equ      02h                 ; Pin Enable Tx of RS485 port
          ENDIF

                                        ; FSS #:
                                        ; Benninger: RS-485 2-Draht mit Enable
                                        ; Sulzer:    RS-485 4-Draht Fullduplex
CutPin    equ       03h                 ; Pin Cut

          IF        Firma='S'
MLfPin     equ      02h                 ; Pin MitLauf
          ENDIF

Mux0      equ       05h                 ; Pins for the MUX DG407 A0
Mux1      equ       06h                 ;                   "    A1
Mux2      equ       07h                 ;                   "    A2

          IF        Firma='B'           ; PORTA:
DChPin     equ      04h                 ; Pin Daisy Chain low->High (ext. NPN Trans!)
          ENDIF



                                        ; Protokol:
                                        ; =========

                                        ; RxData:
                                        ; -------
Rx_Hdr    equ       0FFh                ; Unique data on Rx line to indicate Cmd start
                                        ; ( It must be 0FFh !!)

          IF        _RS=485             ; Only Benninger has addressable nodes!
Rx_Tail    equ      0FEh                ; Data on Rx line to indicate Cmd end
                                        ; ( It must be 0FEh !!)

Rx_Addr    equ      20h                 ; Pic node address #
Rx_LAll    equ      00h                 ; Pic node address for all to listen (no ACK)
                                        ; ( It must be 00h !!)
          ENDIF

                                        ; Cmd bits: (Compared in the order bit 4 to bit 0)
                                        ; ==============================================

                                        ; Process flow bits:
                                        ; ------------------
Cmd_Exe   equ       07h                 ;  Command is executed
Cmd_NACK  equ       06h                 ;  NACK must be sent immediatly
Cmd_ACK   equ       05h                 ;  ACK must be sent immediatly

                                        ; Command bits (001T xxxx):
                                        ; -------------------------
                                        ;  Extended Command:TXh=Test of the following
Cmd_Tst   equ       04H                 ;                       routines by calling TstDbg routine
                                        ;                       after every 'Data-out'
                                        ;                   T0h=Extended commands:
Cmd_Sht   equ       03h                 ;  GetShoot-Xh;OK/NACK > Data's ...

Cmd_GDX   equ       02h                 ;  GetData-All/One-Mean/Val(Sensor);OK/NACK > (8x) Data
GDX_All   equ       01H                 ;    Set for All, Reset for One
GDX_Val   equ       00H                 ;    Set for Value OR Offset, Reset for Mean or Gain

Cmd_SOG   equ       01h                 ;  Set-Offset/Gain(Sensor;Value);OK/NACK > Offset/Gain
SOG_Off   equ       00H                 ;    Set for Offset,Reset for Gain

Cmd_SNV   equ       00h                 ;  GetError;OK/NACK > Error-Sensor


                                        ; Extended commands:
                                        ; ------------------
EXT_SCX   equ       03h                 ;    SelectChannel-X

EXT_ZRO   equ       01h                 ;    Zero
ZRO_GRst  equ       00h                 ;     Set bit if Gain Reset is desired

EXT_GER   equ       00h                 ;    GetError

                                        ; Listen All commands:
                                        ; --------------------
LAC_GER   equ       00h                 ;    GetError

                                        ; Reset conditions:
                                        ; -----------------
_RstC     equ       'R'                 ; Reset
_WdtSlpC  equ       'S'                 ; WDT occured in sleep mode
_WdtToC   equ       'T'                 ; WDT occured during running
_MClrSC   equ       'M'                 ; Reset pin was pulled in sleep mode

                                        ; TxData:
                                        ; -------
Tx_Rst    equ       080h+_RstC          ; Reset
Tx_WDTSLP equ       080h+_WdtSlpC       ; WDT occured in sleep mode
Tx_WDTTO  equ       080h+_WdtToC        ; WDT occured during running
Tx_MCLRS  equ       080h+_MClrSC        ; Reset pin was pulled in sleep mode

Tx_SErr   equ       0FEh                ; NACK: Syntax Error occured
Tx_RxBusy equ       0FDh                ; NACK: Rx is still busy while new data is pushing
Tx_RxErr  equ       0FCh                ; NACK: Rx start/stop bit error occured
Tx_ACK    equ       0FBh                ; ACK:  ''
Tx_Cut    equ       0FAh                ; NACK: A Cut occured
Tx_MLf    equ       0F9h                ; NACK: A MLf occured
Tx_NErr   equ       0F8h                ; OK :  No Error (for GER)
                                        ; ...

DataOfl   equ       0F9h                ; Data overflow
DataUfl   equ       0F8h                ; Data underflow

Tx_DataMx equ       0F8h                ; 00H-F7H=data

;
;*******************************************************************
;
;Variables
;

SaveW     equ       0Ch                 ; Interrupt save register W
SaveST    equ       0Dh                 ;    "       "   STATUS
RxData    equ       0Eh                 ; Received byte
TxData    equ       0Fh                 ; Byte to send
TxRxFlg   equ       10h                 ; Tx & Ryx Flags and counter
Cmd       equ       11h                 ; Command register:           00H=no cmd

PStat     equ       12h                 ; Process register:
                                        ; Bit
                                        ; 0,1:    RTCC Process Counter:         00=Tx
                                        ;                                       01=[Rx]
                                        ;                                       10=#(Adjust)
                                        ;                                       11=[Rx]
                                        ; 2,3:    Rx/TxFlags: 1 if receiving/sending
                                        ; 4:      Rx process pointer 0=RTCC proc 01
                                        ;                                1=RTCC proc 11
                                        ; 5-7:    Sample rate count down
                                        ;

                                        ; FSS regs:
                                        ; ---------
ChNivSel  equ       13h                 ; Selects channel by nibble values 0-7
                                        ; Niveau value at high nibble
ChCut     equ       14h                 ; Cut var for:      Bit 7-0: 1 for cutted

W1        equ       15h                 ; Working Regs
W2        equ       16h                 ; "
W3        equ       17h                 ; "

Buf0      equ       18h                 ; Data buffers for calc/control         Channel 0
Buf1      equ       19h                 ;                   "                   Channel 1
Buf2      equ       1Ah                 ;                   "                   Channel 2
Buf3      equ       1Bh                 ;                   "                   Channel 3
Buf4      equ       1Ch                 ;                   "                   Channel 4
Buf5      equ       1Dh                 ;                   "                   Channel 5
Buf6      equ       1Eh                 ;                   "                   Channel 6
Buf7      equ       1Fh                 ;                   "                   Channel 7

Off0      equ       20h                 ; Offset of Channel 0
Off1      equ       21h                 ; Offset of Channel 1
Off2      equ       22h                 ; Offset of Channel 2
Off3      equ       23h                 ; Offset of Channel 3
Off4      equ       24h                 ; Offset of Channel 4
Off5      equ       25h                 ; Offset of Channel 5
Off6      equ       26h                 ; Offset of Channel 6
Off7      equ       27h                 ; Offset of Channel 7

Gain0     equ       28h                 ; Gain of Channel 0
Gain1     equ       29h                 ; Gain of Channel 1
Gain2     equ       2Ah                 ; Gain of Channel 2
Gain3     equ       2Bh                 ; Gain of Channel 3
Gain4     equ       2Ch                 ; Gain of Channel 4
Gain5     equ       2Dh                 ; Gain of Channel 5
Gain6     equ       2Eh                 ; Gain of Channel 6
Gain7     equ       2Fh                 ; Gain of Channel 7
;
;*******************************************************************
;
;Vectors (PIC16C71)
;

          ORG       PIC71
_Reset    goto      _StartUp            ; Reset startup routine

                                        ; PIC 71: new defined vectors:
                                        ; ----------------------------
_WDTSLP   goto      _WdtSlp             ; # Reenter after WDT wake-up from SLEEP
_WDTTO    goto      _WdtTo              ; # Reenter after WDT time-out
_MCLRSLP  goto      _MClrSlp            ; # Reenter after /MCLR wake-up from Sleep

                                        ; (HINT: user must firstly clear flags with
                                        ; "clrwdt" and set GIE!!)

_IntV     goto      _IntHndlr           ; Interrupt handler (Reenter after INT
                                        ;   wake-up from Sleep)

_Vers     data      "MpK V0.8/FSS V1.9/"; Mark "Version/...

          IF        Firma='S'           ; ...Firma/...
           data     "(c)NTB+Sulzer,Rueti/"
          ELSE
           data     "(c)NTB+Benninger,Uzwil/"
          ENDIF

          IF        _MHz=24             ; ...xxMHz/..
           data     "24"
          ELSE

          IF        _MHz=20
           data     "20"
          ELSE

          IF        _MHz=16
           data     "16"
          ELSE
           data     "??"                ; #
          ENDIF

          ENDIF

          ENDIF

           data     "MHz/RS"            ; ..RS.../..

          IF        _RS=232
           data     "232C/"
          ELSE
           data     "485/"
          ENDIF

_Baud     data      "19.2kBd/"          ; # mark also baud rate

          IF        _Tst=Yes
           data     "Tst/"
          ENDIF

          IF        _RxInv=Yes
           data     "RxINV/"
          ENDIF
          IF        _TxInv=Yes
           data     "TxINV/"
          ENDIF
          IF        _ADInv=Yes
           data     "ADINV"
          ENDIF
;
;*******************************************************************
;
; Subroutines which must be at PCH=0
; ----------------------------------
;

                                        ; Gets Buf(ChNivSel 0-7)-Offset*Gain to W1
                                        ; =====================================
GBufW     bcf       STATUS,CARRY        ; Clear for rotation
          rlf       ChNivSel,W          ; Get ChNivSel * 2 as jump
          andlw     0eh                 ; Mask Select bits
          addwf     PC,F                ; Jump to location

          movf      Buf0,W              ; Get bufdata to set into W
          goto      SetW1
          movf      Buf1,W
          goto      SetW1
          movf      Buf2,W
          goto      SetW1
          movf      Buf3,W
          goto      SetW1
          movf      Buf4,W
          goto      SetW1
          movf      Buf5,W
          goto      SetW1
          movf      Buf6,W
          goto      SetW1
          movf      Buf7,W

SetW1     movwf     W1                  ; Set W1 with converted data

          rlf       ChNivSel,W          ; Get ChNivSel * 2 as jump
          andlw     0eh                 ; Mask Select bits
          addwf     PC,F                ; Jump to location

          movf      Off0,W              ; Get Offset data to set into W
          goto      SubOff
          movf      Off1,W
          goto      SubOff
          movf      Off2,W
          goto      SubOff
          movf      Off3,W
          goto      SubOff
          movf      Off4,W
          goto      SubOff
          movf      Off5,W
          goto      SubOff
          movf      Off6,W
          goto      SubOff
          movf      Off7,W

SubOff    subwf     W1,F                ; Subtract Offset data from W1

          skpnc                         ; Test for underflow!!
          goto      Gain

                                        ; Underflow!!
Ufl       movlw     DataUfl             ; ===========
          movwf     W1                  ; Set Ufl-Error
          setc                          ; Mark it with Carry
          return                        ; No gain calculation needed


Gain      bcf       STATUS,CARRY        ; Clear for rotation
          rlf       ChNivSel,W          ; Get ChNivSel * 2 as jump
          andlw     0eh                 ; Mask Select bits
          addwf     PC,F                ; Jump to location

          movf      Gain0,W             ; rotate gain to carry
          goto      MultGain
          movf      Gain1,W             ; Get Gain data to set into W
          goto      MultGain
          movf      Gain2,W
          goto      MultGain
          movf      Gain3,W
          goto      MultGain
          movf      Gain4,W
          goto      MultGain
          movf      Gain5,W
          goto      MultGain
          movf      Gain6,W
          goto      MultGain
          movf      Gain7,W

MultGain  movwf     W2                  ; Save Gain to W2
          clrf      W3                  ; W3 is result (init with 0)

          setc                          ; Round up half
          rlf       W1,F                ; Expand range
          skpnc                         ; Test Overflow ?
          goto      Ofl                 ;  Yes: Overflow, Carry is already set!

          bcf       STATUS,Mult_Ofl     ; Clear Mult_Ofl-Flag(RxError)
MGLoop    bsf       INTCON,GIE          ; Enable Interrupt
          clrc                          ; Clear carry
          rlf       W2,F                ; Get next bit of gain into carry
          movf      W1,W                ; Get data (prefetch) to W
          skpnc                         ; Is there an Overflow into carry ?
          addwf     W3,F                ;  Yes: Add data (W) to result
          skpnc                         ; Is there an new Overflow into carry ?
          bsf       STATUS,Mult_Ofl     ;   Yes: Set Mult_Ofl Flag

NOvr      clrc
          rrf       W1,F                ; Data=data/2
          movf      W2,F                ; Test Gain for zero
          bcf       INTCON,GIE          ; Disable Interrupt
          skpz                          ; If zero then end
          goto      MGLoop              ; else loop

MultEnd   bsf       INTCON,GIE          ; Enable Interrupt
          clrc                          ; Reset carry
          rrf       W3,W                ; Shrink range
          clrc                          ; Reset carry
          btfsc     STATUS,Mult_Ofl     ; Test Mult_Ovl
          addlw     80h                 ; Correct result (set C???)


TstOfl    skpnc                         ; Carry=Overflow or error!
          goto      Ofl                 ;   Yes: Overflow

          addlw     ~Tx_DataMx          ; Test Range
          skpnc
          goto      Ofl                 ;

          addlw     Tx_DataMx           ;
          addlw     01h                 ; Reset old value!
          clrc                          ; No error
          goto      NOfl                ;

Ofl       movlw     DataOfl             ;   Yes: Mark Overflow if neccessary

NOfl      bcf       STATUS,Mult_Ofl     ; Clear Mult_Ofl-Flag(RxError)
          return

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

                                        ; Gets Offset/Gain(ChNivSel 8-17h) to W
                                        ; ==================================
GetOW     bcf       STATUS,CARRY        ; Clear for rotation
          rlf       ChNivSel,W          ; Get ChNivSel * 2 as jump
          andlw     1eh                 ; Mask Select bits
          addlw     00-8-4              ; Calc position
          addwf     PC,F                ; Jump to location

GetGW     bcf       STATUS,CARRY        ; Clear for rotation
          rlf       ChNivSel,W          ; Get ChNivSel * 2 as jump
          andlw     1eh                 ; Mask Select bits
          addwf     PC,F                ; Jump to location

OffTab    movf      Off0,W              ; Get Offdata to set into W
          return
          movf      Off1,W              ; Get Offdata to set into W
          return
          movf      Off2,W              ; Get Offdata to set into W
          return
          movf      Off3,W              ; Get Offdata to set into W
          return
          movf      Off4,W              ; Get Offdata to set into W
          return
          movf      Off5,W              ; Get Offdata to set into W
          return
          movf      Off6,W              ; Get Offdata to set into W
          return
          movf      Off7,W              ; Get Offdata to set into W
          return

GainTab   movf      Gain0,W             ; Get Gaindata to set into W
          return
          movf      Gain1,W             ; Get Gaindata to set into W
          return
          movf      Gain2,W             ; Get Gaindata to set into W
          return
          movf      Gain3,W             ; Get Gaindata to set into W
          return
          movf      Gain4,W             ; Get Gaindata to set into W
          return
          movf      Gain5,W             ; Get Gaindata to set into W
          return
          movf      Gain6,W             ; Get Gaindata to set into W
          return
          movf      Gain7,W             ; Get Gaindata to set into W
          return

;
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
                                        ; Gets pure Buf(ChNivSel 0-7h) to W
                                        ; ==================================
GetBW     bcf       STATUS,CARRY        ; Clear for rotation
          rlf       ChNivSel,W          ; Get ChNivSel * 2 as jump
          andlw     0eh                 ; Mask Select bits
          addwf     PC,F                ; Jump to location

BufTab    movf      Buf0,W              ; Get Bufdata to set into W
          return
          movf      Buf1,W              ; Get Bufdata to set into W
          return
          movf      Buf2,W              ; Get Bufdata to set into W
          return
          movf      Buf3,W              ; Get Bufdata to set into W
          return
          movf      Buf4,W              ; Get Bufdata to set into W
          return
          movf      Buf5,W              ; Get Bufdata to set into W
          return
          movf      Buf6,W              ; Get Bufdata to set into W
          return
          movf      Buf7,W              ; Get Bufdata to set into W
          return

;
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;

MarkCut   bsf       PORTB,CutPin        ; Set output pin Cut
          swapf     PORTB,W             ; Mark Cut
          andlw     b'00001110'         ; ========

          addwf     PC,F                ; Jump to location

SetBit    bsf       ChCut,0             ; Set Bit 0 of ChCut
          return
          bsf       ChCut,1             ; Set Bit 1 of ChCut
          return
          bsf       ChCut,2             ; Set Bit 2 of ChCut
          return
          bsf       ChCut,3             ; Set Bit 3 of ChCut
          return
          bsf       ChCut,4             ; Set Bit 4 of ChCut
          return
          bsf       ChCut,5             ; Set Bit 5 of ChCut
          return
          bsf       ChCut,6             ; Set Bit 6 of ChCut
          return
          bsf       ChCut,7             ; Set Bit 7 of ChCut
          return
;
;*******************************************************************
;
; Start up after any reset or power up...
;

                                        ; Detects Reset conditions
                                        ; and reenters accordingly
                                        ; ------------------------
_StartUp  btfsc     STATUS,PD           ; IF sleep detected THEN
          goto      nosleep
          btfss     STATUS,TO           ;  IF WDT time-out THEN
          goto      _WDTSLP             ;   Goto WDT-Sleep reentrance
          goto      _MCLRSLP            ;  ELSE Goto /MCLR-Sleep reentrance
                                        ;  ENDIF
nosleep   btfss     STATUS,TO           ; ELSIF WDT time-out THEN
          goto      _WDTTO              ;  Goto WDT time-out reentrance
                                        ; ENDIF

          movlw     Tx_Rst              ; Send Tx_Rst at Reset if bit 7 is set
          movwf     TxData
          movlw     _RstC               ; Mark Reset Condition
          movwf     Cmd
          goto      _PwrOnRst

; -----------------------------------------------------------------------

_WdtSlp   movlw     Tx_WDTSLP           ; Send Tx_WDTSLP at Reset if bit 7 is set
          movwf     TxData
          movlw     _WdtSlpC            ; Mark WDT in sleep mode Condition
          movwf     Cmd
          goto      _Dummy

; -----------------------------------------------------------------------

_WdtTo    movlw     Tx_WDTTO            ; Send Tx_WDTTO at Reset if bit 7 is set
          movwf     TxData
          movlw     _WdtToC             ; Mark WDT timeout Condition
          movwf     Cmd
          goto      _Dummy

; -----------------------------------------------------------------------

_MClrSlp  movlw     Tx_MCLRS            ; Send Tx_MCLRS at Reset if bit 7 is set
          movwf     TxData
          movlw     _MClrSC             ; Mark Reset pin pulled in sleep Condition
          movwf     Cmd
          goto      _Dummy

;
;***************************************************************
;
; PowerOn- or Reset-Condition
; ---------------------------
; Put system into defined condition
; Hardware initialization:
;

_Dummy    IF        _WDT=Yes            ; and unused entry for other power up cond.
           clrwdt
          ENDIF
                                        ; Ordinary entry...
;
; Hardware initialization:
;

                                        ; TRISA Register: #
                                        ; -----------------
_PwrOnRst movlw     b'11111'            ; Out:    -
          tris      TRISA               ; In:     0-3:      AD-Inputs,
                                        ;         4:        Daisy Chain Pin
          movlw     b'11111'            ; Reset Port A: DCh=1 (Preset for output)
          movwf     PORTA               ; -------------
          movf      PORTA,W
                                        ; TRISB Register: (#)
                                        ; -------------------
          movlw     b'00000001'         ; Out:    7,6,5,  Mux Pins
                                        ;         4,3,2,  Cut, MLf, TxEn pins
                                        ;         1:      TxPin
          tris      TRISB               ; In:     0:      RxPin
          IF        _TxInv=Yes          ; Reset Port B: MUX=000,Cut=0,MLf=0,DCh=0
           movlw    b'00000001'                             ; -------------
          ELSE                ;               TxEn=0,TxPin=1
           movlw    b'00000011'
          ENDIF
          movwf     PORTB
          movf      PORTB,W

          movlw     Buf0                ; Load Indirect Pointer to first buffer
          movwf     FSR

          movlw     b'10001000'         ; OPTION Register init: (#)
          OPTION                        ; RBPU=dis, INTEDG=falling, RTS=Intern,
                                        ; RTE=X, PSA=WDT,(Presc=000)

          movlw     b'11000001'         ; ADCON0 Register: #
          movwf     ADCON0              ; ADCS=11,x,CHS=00,GO=0,ADIF=0,ADON=1


;         movlw     b'01'               ; ADCON1 Register: #
          Page1                         ; PCFG=xxxxxx01:RA0-2=ADin,RA3=ARef
          movwf     ADCON1
          Page0

;
; Init system regs.:
;

          movlw     SRate+b'0011'
          movwf     PStat               ; Bit 7-4:          Countdown sample rate
                                        ; Bit 3,2:          TxFlag, RxFlag,
                                        ; Bit 1,0:          Proc 11 (->Tx 00)

          movlw     b'00001000'
          movwf     TxRxFlg             ; Init Tx / Rx flags
                                        ; Bit 7-5:          Tx bit counter=000
                                        ; Bit 4:  Flag TxStop
                                        ; Bit 3:  Flag TxStart  =1
                                        ; Bit 2:  Flag RxSel proc 01,11 select
                                        ; Bit 1:  Flag RxStop
                                        ; Bit 0:  Flag Rxtart

;
;
; RTCC and prescaler initialization
;

          clrwdt                        ; Set defined conditions to flags, WDT
                                        ; and the prescaler
          Page1
          movf      OPT,W               ; Set prescaler:
          Page0
          iorlw     Presc               ; Internal,low-high,RTCC,'constant'
          OPTION                        ; (OPT was set to xx000000b within _InitHW)
                                        ; First time RTCC runs up from 00h to 0FFh

;
;
; Enable and disable interrupts:
;
          movlw     b'11110000'         ; # En-/disables Interrupts:
          movwf     INTCON              ;   GIE,ADIE,RTIE,INTE,RBIE,rtcc,int,rb

          goto      _Main               ; Goto Main Programm

;
;*******************************************************************
;
; Main interrupt handler which branches depending on interrupt source
; -------------------------------------------------------------------
; All interrupts are branching to this routine. It firstly saves the
; work- and the statusfile, than detects the occured interrupt and
; jumps to the according subroutine.
; The way the user queues the calls that makes the priority. Last has
; most priority, if one after the other is tested ("_RetIntX").
; If the user wants equal priority, just reenter always at "_RetInt"!
; To maintain 100% STATUSreg recovery, the zeroflag must be testet
; and re-set after reloading the workfile!
;

                                        ; Interrupt Handler (GIE=0)
                                        ; -----------------
                                        ; higher priority interrupts which occur
                                        ; during handler are executed if detected
                                        ; after RetIntX of the previous interrupt!

_IntHndlr movwf     SaveW               ; Save W and STATUS
          movf      STATUS,W
          movwf     SaveST

                                        ; Interrupt branch depending on source
                                        ; and reenter-test:
                                        ; (Most important proc last!)
                                        ; ------------------------------------
_RetInt   btfsc     ADCON0,ADIF         ; IF AD completion THEN
          goto      _IntAD              ;  goto AD-Handler

_RetInt1
;         btfsc     INTCON,RBIF         ; ELSEIF Port B change THEN
;         goto      _IntPORTRB          ;  goto RB-Handler

_RetInt2  btfsc     INTCON,INTF         ; ELSEIF INT THEN
          goto      _INT                ;  goto INT-Handler

_RetInt3  btfsc     INTCON,RTIF         ; IF RTCC overflow THEN
          goto      _IntRTCC            ;  goto RTCC-Handler
                                        ; ENDIF

_RetInt4  movf      SaveST,W            ; Reload old STATUS and
          movwf     STATUS
          movf      SaveW, W            ; Z-Flag changed!!
;         btfsc     SaveST,Z            ; Test old Z
;         setz                          ; and re-set it...
;         btfss     SaveST,Z
;         clrz
EOI       retfie                        ; Enable global interrupt and
                                        ; return to processes

;
;*******************************************************************
;
; Interrupt on completion of A/D-conversion (wake-up)
; ---------------------------------------------------
; #

_IntAD    bcf       ADCON0,ADIF         ; Reset flag
          IF        _Tst=Yes
           btfss    INTCON,ADIE         ; IF AD is disabled THEN
           goto     _RetInt1            ;   return
          ENDIF                         ; ENDIFF

          IF        _ADInv=Yes
InvRes     movlw    0FFh                ; Invert ADRES
           xorwf    ADRES,F             ;
          ENDIF

TstCut    movf      ADRES,W             ; Test cut
          subwf     INDADR,W            ; ========
          skpc                          ; IF Buf-Res<0(NC) THEN
          goto      NoCut               ;  Incline: no cut

TstNiv    rrf       ChNivSel,W          ; ELSE...(Dec)
          andlw     b'01111000'         ;   Get/mask Niveau bits
          addwf     ADRES,W             ; ...IF Res+Niv > FFH(C) THEN
          skpnc
          goto      NoCut               ;   Step too small: no cut

          subwf     INDADR,W            ; ELSEIF Buf-(Res+Niv)<0(NC) THEN
          skpc                          ;   no cut
          goto      NoCut               ; ELSE

Cut       Call      MarkCut             ; Mark Cut into ChCut
          bsf       Cmd,Cmd_NACK        ; Mark into Cmd-Word NACK

NoCut     movf      ADRES,W             ; Store Result to buffer
                                        ; ======================
          IF        Firma='S'
           btfsc    Cmd,Cmd_Sht         ; Is Cmd Shoot ?
           goto     Value
          ENDIF
          btfss     Cmd,Cmd_GDX         ; Is Cmd GDX?
          IF        Firma='S'
           goto     Value               ; Sulzer is value important
          ELSE
           goto     Mean                ; Benninger is mean important
          ENDIF
          btfsc     Cmd,GDX_All         ; Is GetDataAll?
          IF        Firma='S'
           goto     Value               ; Sulzer is value important
          ELSE
           goto     Mean                ; Benninger is mean important
          ENDIF
          btfsc     Cmd,GDX_Val         ; Is GetDataValue?
          goto      Value               ;  Yes...

                                        ; else
Mean      incf      INDADR,F            ; round up
          skpnz                         ; Was an overflow?
          decf      INDADR,F            ;   Yes: do not round up!
          addwf     INDADR,F            ; calc mean value
          rrf       INDADR,F            ; (/2)
          goto      Incr

Value     movwf     INDADR              ; only value

                                        ; Increment MUX
Incr      movlw     b'00100000'         ; =============
          addwf     PORTB,F             ; MUX=MUX+1

                                        ; Increment Buffer Pointer
          incf      FSR,F               ; ========================
          btfss     FSR,05h             ; Last buffer reached?
          goto      _RetInt1            ;  No...

          movlw     Buf0                ;  Yes:
          movwf     FSR                 ;   Set to first buffer
          goto      _RetInt1            ; End of AD interrupt

;
;*******************************************************************
;
; Interrupt on changes on port RB4-7 (wake-up)
; --------------------------------------------
; #

;_IntPORTRB movwf   PORTB,W
;         bcf       INTCON,RBIF         ; Reset flag
;         btfss     INTCON,RBIE         ; IF RB is disabled THEN
;         goto      _RetInt2            ;   return
                                        ; ENDIF

                                        ; #

;         goto      _RetInt2            ; End of Port RB interrupt

;
;*******************************************************************
;
; Extern INTerrupt on RB0 pin input (wake-up)
; -------------------------------------------
; Is used to determine RxStart. A falling edge activates the interrupt.
; The sub disables itself during next 8 bits until data is received.
; The RxProcess is set free by TxFlag and TxStart. To catch the input
; level at the right time next RxProcess time slice is counted and set
; to flag RxSel.
;

_INT      bcf       INTCON,INTF         ; Reset flag
          btfss     INTCON,INTE         ; IF INT is disabled THEN
          goto      _RetInt3            ;   return
                                        ; ENDIF

          bcf       INTCON,INTE         ; Disable next interrupts until _RxStop

          bsf       PStat,RxFlag        ; Set RxFlag
          bsf       TxRxFlg,RxStart     ; Set pointer to start bit

          incf      PStat,F             ; Because _INT is before _IntRTCC
          btfsc     PStat,PCnt1         ; If Proc00/01 is active then set RxSel=0
          bsf       TxRxFlg,RxSel
          btfss     PStat,PCnt1         ; else set RxSel=1
          bcf       TxRxFlg,RxSel
          decf      PStat,F

          goto      _RetInt3            ; End of INT interrupt

;
;*******************************************************************
;
; Interrupt on Real Time Clock overflow:
; --------------------------------------
; Every (#) Xus the RTC has its overflow to detemine the right time intervall
; for Tx- and RxProcesses. The user must consider of having enough spare
; cycle to avoid the RTC overflowing twice in case of several interrupt
; sources occur!
; Before jumping to the 4 intervall processes a separate AD-Start sub is
; testet for its call. This allows to start the AD-conversion at every
; 1*Xus up to 16*Xus (sample rate)!
; If a uncommon X-tal is used, an adjust process may be provided to inc or dec
; some RTC-steps to adjust the roundabout time of the 4 processes!
; Otherwise this process may be used by the user. #
;

_IntRTCC  bcf       INTCON,RTIF         ; Reset flag
;         btfss     INTCON,RTIE         ; IF RTCC is disabled THEN
;         goto      _RetInt4            ;   return
                                        ; ENDIF

          movlw     100H-Rtcc           ; Set RTCC start value from constant (Xus)
          addwf     RTCC,f              ; #

          skpc                          ; Test if there was a cycle lost?
          goto      NoLostC
                                        ; LostCycle
          movlw     0ffh                ; ---------
          movwf     RTCC                ; Set it near before INT

NoLostC   movlw     11h                 ; Countdown of sample rate and
          addwf     PStat,F             ; set process counter + 1 (overflow bit 2?)
          skpnc
          call      _ADProc             ; Proc AD: Start conversion

                                        ; Check processes 00 to 11
                                        ; ------------------------
          btfsc     PStat,PCnt0         ; Test
          goto      _RxProc             ; Proc 01 or 11: Rx depending on RxSel
          btfsc     PStat,PCnt1
          goto      _RetInt4            ; Not used:
;         goto      _AdjProc            ; Proc 10: # (Adjust)
          movlw     4h                  ; fix overflow bit 2
          subwf     PStat,F             ;
          goto      _TxProc             ; Proc 00: Tx

;
;*******************************************************************
;
; AD process:
; -----------
; Starts AD-conversion and resets sample rate countdown
;

_ADProc   bsf       ADCON0,GO           ; Start conversion
          movlw     SRate               ; Re-set countdown value
          addwf     PStat,F
          return

;
;*******************************************************************
;
; Tx process 00:If TxFlag is set data TxData will be sent via pin RB1.
;                   According to TxRxFlgs:
;                   TxStart:start(low)
;                   TxData: (Count at TxRxFlg)
;                   1xH:      bit0
;                   3xH:      bit1
;                   5xH:      bit2
;                   7xH:      bit3
;                   9xH:      bit4
;                   BxH:      bit5
;                   DxH:      bit6
;                   FxH:      bit7 -> Carry sets TxStop
;                   TxStop:   stop(high),reset TxFlag to mark end of Tx.
;                                          and sets TxStart for next transmission
; To set a new TxData check if TxFlag=0!
;

                                        ; Transmitter every 4*Xus
_TxProc   btfsc     PStat,TxFlag        ; Is transmitter active?
          goto      _TxOK               ;   yes: send next data

          IF        _RS=485
           bcf      PORTB,TxEn          ;   no:  Disable output
          ENDIF
          goto      _RetInt4            ;        return

_TxOK     btfsc     TxRxFlg,TxStart     ; Is start bit to be sent?
          goto      _TxStart            ; Yes
          btfss     TxRxFlg,TxStop      ; Is data to be sent?
          goto      _TxData             ; Yes

_TxStop   IF        _TxInv=Yes
           bcf      PORTB,TxPin         ; Set output high (STOP)
          ELSE
           bsf      PORTB,TxPin         ; Set output high (STOP)
          ENDIF
          bcf       TxRxFlg,TxStop      ; Clear TxStop
          bcf       PStat,TxFlag        ; Clear TxFlag
          bsf       TxRxFlg,TxStart     ; Set TxStart for next transfer
          IF        _RS=485
           bsf      INTCON,INTE         ; Enable Rx Int-process
          ENDIF
          goto      _RetInt4

                                        ; Send START
_TxStart  IF        _RS=485
           btfsc    PStat,RxFlag        ; Is Rx receiving ?
           goto     _RetInt4            ;   Yes: Wait for its end OR set Error???

           bcf      INTCON,INTE         ; Disable Rx Int-process
           bsf      PORTB,TxEn          ; Enable output
          ENDIF

          IF        _TxInv=Yes
           bsf      PORTB,TxPin         ; Set output low (START)
          ELSE
           bcf      PORTB,TxPin         ; Set output low (START)
          ENDIF
           bcf      TxRxFlg,TxStart     ; Start is sent
          goto      _RetInt4

                                        ; Send DATA
_TxData   IF        _Tst=Yes
           clrc                         ; Set rotatebits to zero
          ENDIF
          rrf       TxData,F            ; bit to carry
          IF        _TxInv=Yes
           skpc
           bsf      PORTB,TxPin         ; Set output according to carry
           skpnc
           bcf      PORTB,TxPin         ;
          ELSE
           skpnc
           bsf      PORTB,TxPin         ; Set output according to carry
           skpc
           bcf      PORTB,TxPin         ;
          ENDIF
          movlw     20h                 ; Increment counter...
          addwf     TxRxFlg,F           ; count up to bit 8...
          skpnc                         ; is last bit sent?
          bsf       TxRxFlg,TxStop      ; Yes->Set Stop flag
          goto      _RetInt4

;
;*******************************************************************
;
; Rx process 01,11: If RxFlag is ser and the correct process according
; to RxSel is running, data RxData will be sent via pin RB1:
;                   According to TxRxFlags:
;                   RxStart:start(low), set first bit of RxData high
;                   RxData: bit0..7, shift RxData till first bit is high
;                                        ->set RxStop
;                   RxStop:   stop(high),reset RxFlag, enable INTerrupt,
; # User must clear flag DataRdy after datafetch at file ADCON0
; to mark data is read!
;

                                        ; Receiver process every (other) 4*Xus
_RxProc   btfss     PStat,RxFlag        ; Is receiver active?
          goto      _RetInt4            ; no:return
          btfsc     PStat,PCnt1         ; Is it the selected timeslice?
          goto      _Rx11               ;
_Rx01     btfss     TxRxFlg,RxSel       ; PCnt=0:Sel=?
          goto      _RxOK               ; Sel=0   =>ok, go ahead
          goto      _RetInt4            ; Sel=1 =>return
_Rx11     btfss     TxRxFlg,RxSel       ; PCnt1=1:Sel=?
          goto      _RetInt4            ; Sel=0 =>return

_RxOK     btfsc     TxRxFlg,RxStart     ; Is start bit first
          goto      _RxStart            ; Yes
          btfss     TxRxFlg,RxStop      ; Is stop bit next?
          goto      _RxData             ; No: ->Data

_RxStop   IF        _RxInv=Yes
           btfsc    PORTB,RxPin         ; Is signal high?
          ELSE
           btfss    PORTB,RxPin         ; Is signal high?
          ENDIF
          goto      _RxError            ; Low: error
                                        ; High:ok...
_RxStopE  bsf       INTCON,INTE         ; Enable INT interrupt for new data
          bsf       ADCON0,DataRdy      ; Set data ready flag
          bcf       PStat,RxFlag        ; Clear RxFlag
          bcf       TxRxFlg,RxStop      ; Clear RxStop
          goto      _RetInt4
                                        ; An error occured:
_RxError  bsf       SaveST,RxError      ; Mark Rx Error (Reset in Main after sent)
          bsf       Cmd,Cmd_NACK        ; Mark into Cmd-Word NACK
          goto      _RxStopE            ; # Re-enable RxProcess after sending NACK!!


_RxStart  bcf       TxRxFlg,RxStart     ; Start bit passed
                                        ; ===============================
                                        ; # Execute "bcf ADCON0,DataRdy"
                                        ;   after data is read and before
                                        ;   new databits are comming in!!
                                        ; ===============================
          btfsc     ADCON0,DataRdy      ; Test for old data is read?
          goto      _RxBusy             ; Too slow.... >>> error!!!
          IF        _RxInv=Yes
           btfsc    PORTB,RxPin         ; Is signal low?
          ELSE
           btfsc    PORTB,RxPin         ; Is signal low?
          ENDIF
          goto      _RxError            ; High: error
          movlw     80H
          movwf     RxData              ; set rotate flag
          goto      _RetInt4

                                        ; An error occured:
_RxBusy   bsf       Cmd,Cmd_NACK        ; Mark into Cmd-Word NACK
          bsf       SaveST,RxBusy       ; Mark Rx Busy (Reset in Main after sent)
          goto      _RxStopE            ; # Re-enable RxProcess after sending NACK!!


_RxData   btfsc     PORTB,RxPin         ; Receive data
          IF        _RxInv=Yes
           clrc
          ELSE
           setc
          ENDIF
          btfss     PORTB,RxPin
          IF        _RxInv=Yes
           setc
          ELSE
           clrc
          ENDIF
          rrf       RxData,F            ; Carry to file
          skpnc
          bsf       TxRxFlg,RxStop      ; Last bit is received -> get stop bit now
          goto      _RetInt4

;
;*******************************************************************
;
; Adjust process 10:
; ------------------
; This process may be used for adjustment of the roundabout time of
; the 4 Process (and the sample rate). By inc- or decrementing the RTCC
; any adjustment may be done!
; # If not used this is a free user definable process:
;

;_AdjProc movlw     Adj                 ; Adjust RTCC to the av. of Xus/process
;         addwf     RTCC,F
                                        ; OR #
;         goto      _RetInt4

;
;*******************************************************************
;
; Subroutines:
; ============

TestCut   clrc                          ; Test if there is a cut => set carry bit
                                        ; =======================================
          bcf       INTCON,GIE          ; Disable Interrupt
          movf      ChCut,W             ; Z if no cut
          skpz                          ; If NZ (=cut) then
          setc                          ;  set carry
          bsf       INTCON,GIE          ; Enable Interrupt
          return

;
;*******************************************************************
;
; Main Processes
; --------------
; MpK Rem.:         7 stack levels are free. Z-Flag may not be used without
;                   disabling GIE!

; Zwo pseude processes are conncted and runned by a roundabout angle
; point: SUB SubInpExe. during the wait of cmd-process for cmd input,
; cmd flags and cmd output are executed via output sub processes.


_Main     IF        _Tst=Yes
           btfsc    TxData,07h          ; Send Reset Info if WDT is used
           bcf      TxData,07h          ; clear that bit
           bsf      PStat,TxFlag
          ENDIF
                                        ; Clear used Variables
                                        ; ====================
          clrf      W1                  ; Clear all work regs
          clrf      W2
          clrf      W3

;         movlw     _RstC               ; Was other than normal reset?
;         subwf     Cmd
;         skpz                          ; Yes:
;         goto      EndRstCnd           ;   skip settings

          movlw     2*50h               ; Clear FSS regs:
          movwf     ChNivSel            ; Niveau=50(08h-78h), no Channel selected
          clrf      ChCut               ; No Cut

                                        ; Set First fix values
                                        ; ====================
          IF        P2nd=Yes            ; PROGRAMM second (overwrite)
           movlw    080h                ; Gain = 1.00h
           movwf    Gain0               ; Set Gain Channel 0
           movlw    090h                ; Gain = 1.00h
           movwf    Gain1               ; Set Gain Channel 1
           movlw    0A0h                ; Gain = 1.00h
           movwf    Gain2               ; Set Gain Channel 2
           movlw    0B0h                ; Gain = 1.00h
           movwf    Gain3               ; Set Gain Channel 3
           movlw    0C0h                ; Gain = 1.00h
           movwf    Gain4               ; Set Gain Channel 4
           movlw    0D0h                ; Gain = 1.00h
           movwf    Gain5               ; Set Gain Channel 5
           movlw    0E0h                ; Gain = 1.00h
           movwf    Gain6               ; Set Gain Channel 6
           movlw    0F0h                ; Gain = 1.00h
           movwf    Gain7               ; Set Gain Channel 7

           movlw    000h                ; Offset = 00h
           movwf    Off0                ; Set Offset of Channel 0
           movlw    008h                ; Offset = 00h
           movwf    Off1                ; Set Offset of Channel 1
           movlw    010h                ; Offset = 00h
           movwf    Off2                ; Set Offset of Channel 2
           movlw    018h                ; Offset = 00h
           movwf    Off3                ; Set Offset of Channel 3
           movlw    020h                ; Offset = 00h
           movwf    Off4                ; Set Offset of Channel 4
           movlw    028h                ; Offset = 00h
           movwf    Off5                ; Set Offset of Channel 5
           movlw    030h                ; Offset = 00h
           movwf    Off6                ; Set Offset of Channel 6
           movlw    038h                ; Offset = 00h
           movwf    Off7                ; Set Offset of Channel 7
           ELSE                         ; PROGRAMM first:
           movlw    0ffh                ; Gain = 1.FEh
           movwf    Gain0               ; Set Gain Channel 0
           movlw    0ffh                ; Gain = 1.FEh
           movwf    Gain1               ; Set Gain Channel 1
           movlw    0ffh                ; Gain = 1.FEh
           movwf    Gain2               ; Set Gain Channel 2
           movlw    0ffh                ; Gain = 1.FEh
           movwf    Gain3               ; Set Gain Channel 3
           movlw    0ffh                ; Gain = 1.FEh
           movwf    Gain4               ; Set Gain Channel 4
           movlw    0ffh                ; Gain = 1.FEh
           movwf    Gain5               ; Set Gain Channel 5
           movlw    0ffh                ; Gain = 1.FEh
           movwf    Gain6               ; Set Gain Channel 6
           movlw    0ffh                ; Gain = 1.FEh
           movwf    Gain7               ; Set Gain Channel 7

           movlw    0FFh                ; Offset = FFh
           movwf    Off0                ; Set Offset of Channel 0
           movlw    0FFh                ; Offset = FFh
           movwf    Off1                ; Set Offset of Channel 1
           movlw    0FFh                ; Offset = FFh
           movwf    Off2                ; Set Offset of Channel 2
           movlw    0FFh                ; Offset = FFh
           movwf    Off3                ; Set Offset of Channel 3
           movlw    0FFh                ; Offset = FFh
           movwf    Off4                ; Set Offset of Channel 4
           movlw    0FFh                ; Offset = FFh
           movwf    Off5                ; Set Offset of Channel 5
           movlw    0FFh                ; Offset = FFh
           movwf    Off6                ; Set Offset of Channel 6
           movlw    0FFh                ; Offset = FFh
           movwf    Off7                ; Set Offset of Channel 7
          ENDIF
                                        ; End of reset conditions:
EndRstCnd clrf      Cmd                 ; Rem.: TxData, RxData are undef...


;--------------------------------------------------------------------
;
; Cmd-Process
; -----------
; Receives Cmd via Rx-Line, checks Header (0FFH), Address (Benninger)
; and saves Cmd. If there is an direct Cmd-action to do at a specific
; cmd, it will be performed immediatly.

GetHdr    nop                           ; Wait for Header
          IF        _RS=485             ; ===============
           bsf      INTCON,GIE          ; Enable Interrupts (used if addresse was wrong)
          ENDIF

RetNACK   call      SubInpExe           ; Wait for Cmd_Hdr

          movlw     Rx_Hdr
          subwf     RxData,W            ; IF Rx_Hdr->FFh-FFH=00H(C) ....

          bcf       ADCON0,DataRdy      ; ( Free RxProcess

          IF        _RS=232             ; Sulzer:
           btfsc    Cmd,Cmd_NACK        ; Is there a NACK?
           goto     BrnchNACK           ; Send NACK and Restart CmdExecution

          ELSE                          ; Benninger:
           btfss    Cmd,Cmd_NACK        ; Is there a NACK?
           goto     HdrOK               ;  No...continue
           btfsc    Cmd,Cmd_Exe         ; Is still a command running?
           goto     BrnchNACK           ;  Yes: Send NACK
           call     TestCut             ; Test for separate cut flag output
           skpc                         ;  C set for cut
           goto     HdrOK               ;  No: go ahead

           movlw    b'01111'            ; Preset for DChPin output
           btfss    PORTA,DChPin        ; Test for not active Daisy Chain
           goto     GetHdr              ;  No: wait for next header
           tris     TRISA               ;  Yes:   Set Daisy chain pin
           bcf      Cmd,Cmd_NACK        ;       Clear flag (Don't send any NACK!...)
           goto     GetHdr              ;  Wait for next header
          ENDIF                         ; )

HdrOK     skpc                          ; THEN get next data ...
          goto      GetHdr              ; ELSE...wait for new header (sync)

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
          IF        _RS=485             ; Only Benninger supports addresses:
                                        ; Wait for Address
                                        ; ================
GetAdr     call     SubInpExe           ; Wait for address data

           movlw    ~Rx_LAll            ; Test for listen all address (00H)
Invers     addwf    RxData,W            ; IF LAll->00H+FFH=FFH,NC ....

           bcf      ADCON0,DataRdy      ;   (Free RxProcess
           movf     RxData,W            ;   Save address W := RxData)
           skpc                         ;
           goto     ListenAll           ;     BRANCH (reset DataRdy there!)
                                        ; ENDIF
TestAdr    bcf      INTCON,GIE          ; Disable Interrupts (Z)
           sublw    Rx_ADDR,W           ; IF RxData=Rx_Addr (Z-Flag!!) THEN
           skpz
           goto     GetHdr              ;   Wait for new header (sync)
           bsf      INTCON,GIE          ; ENDIF,Enable Interrupts (Z)

           btfss    Cmd,Cmd_NACK        ; Is there a NACK?
           goto     AdrOK               ;  No...continue
           btfsc    Cmd,Cmd_Exe         ; Is still a command running?
           goto     BrnchNACK           ;  Yes: Send NACK
           call     TestCut             ; Test for separate cut flag output
           skpc                         ;  C set for cut
           goto     AdrOK               ;  No: go ahead

           movlw    b'01111'            ; Preset for DChPin output
           btfss    PORTA,DChPin        ; Test for not active Daisy Chain
           goto     GetHdr              ;  No: wait for next header
           tris     TRISA               ;  Yes:   Set Daisy chain pin
           bcf      Cmd,Cmd_NACK        ;       Clear flag (Don't send any NACK!...)
           goto     GetHdr              ;  Wait for next header

AdrOK      nop
          ENDIF
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

                                        ; Wait for Cmd
                                        ; ============
GetCmd    movlw     00h                 ; Clear Cmd for new Command. This stops ExeSubs!
          movwf     Cmd
          bcf       STATUS,RxError      ; Reset Rx Errors..
          bcf       STATUS,RxBusy

          call      SubInpExe           ; Wait for Cmd

          movfw     RxData              ; Get RxData
          bcf       ADCON0,DataRdy      ; Free RxProcess
          btfsc     Cmd,Cmd_NACK        ; Is there a NACK?
          goto      BrnchNACK           ; Restart CmdExecution

          bsf       INTCON,ADIE         ; AD (re-)enable after SCX

                                        ; (Default is '001x xxxx')
          movwf     Cmd                 ; Save Cmd
          btfsc     Cmd,07h             ; Test for Syntax Error
          goto      SyntaxErr
          btfsc     Cmd,06h             ; Test for Syntax Error
          goto      SyntaxErr
          btfss     Cmd,05h             ; Test for Syntax Error
          goto      SyntaxErr
                                        ; No SyntaxError:
          movlw     b'00011111'         ; Clear first 3 bits anyway...
          andwf     Cmd,F               ; ... of Cmd
          goto      TstActCmd           ; Test for action command now!

RetTAC    btfsc     Cmd,Cmd_NACK        ; Test for Syntax Error
          goto      BrnchNACK
          goto      GetHdr              ; Endless loop


;
;--------------------------------------------------------------------
;
; Syntax Error occured:
; ---------------------
; Only the NACK-Flag is set!
;

SyntaxErr bsf       Cmd,Cmd_NACK        ; Imaginary set NACK for Syntax error
          bcf       STATUS,RxError      ; Reset Rx Errors..
          bcf       STATUS,RxBusy
;         goto      BrnchNACK           ; (no other flag is set)


;
;--------------------------------------------------------------------
;
; Error branch if Cmd_NACK is set:
;
; If Cmd_NACK is set this routine should be accessed.
; Tree errors may be sent back to host:
;         if RxError is set, start/stop bit failure occured
;         if RxBusy is set, Tx did send to fast
;         else a Syntax error were found
; The according Error-Data will be sent back to host and
; Cmd-Process restarts at GetHdr...



BrnchNACK bcf       Cmd,Cmd_NACK        ; Clear NACK Flag

          IF        _RS=485             ; Wait for Cmd-Protokoll to be finish
Wait00H    call     SubInpExe           ; Wait for Cmd-tail

           movlw    ~Rx_Tail+1          ; Test for Cmd-tail (FEh)
           addwf    RxData,W            ; IF tail->FEh+2=00h->C ....
           movf     RxData,W
           bcf      ADCON0,DataRdy      ; Free RxProcess
           skpnc                        ; Was it Rx-Tail?
           goto     TestErr             ;   Yes: send Errorcode

           movlw    Rx_Hdr
           subwf    RxData,W            ; IF Rx_Hdr->FFh-FFH=00H->C ....
           skpc                         ;   Yes: Leave Error, get Adress
           goto     Wait00H             ;   No:  Wait for next data

           movlw    00h                 ; Leave Error, read new Adress!
           movwf    Cmd
           bcf      STATUS,RxError
           bcf      STATUS,RxBusy
           goto     GetAdr              ;
          ENDIF

TestErr   btfsc     STATUS,RxError      ; Did RxError occur ?
          goto      TxRxError           ;  Yes
          btfsc     STATUS,RxBusy       ; Did Rxbusy occur ?
          goto      TxRxBusy            ;  Yes
          call      TestCut             ; Test for separate cut flag output
          skpnc                         ;  C set for cut
          goto      TxCut               ;  Yes
          goto      TxSxErr             ; Else Syntax Error... !

TxCut     bsf       PORTB,CutPin        ; Set output pin Cut
          movlw     Tx_Cut              ; Send NACK for Cut
          goto      SendErr

                                        ; MPK:
TxRxError movlw     Tx_RxErr            ; Mark for RxError
          bcf       STATUS,RxError      ; Reset Flag
          goto      RxEn                ; Free Rx Process again

TxRxBusy  movlw     Tx_RxBusy           ; Mark for RxBusy
          bcf       STATUS,RxBusy       ; Reset Flag

RxEn      bsf       INTCON,INTE         ; Enable INT interrupt for new data
          goto      SendErr


TxSxErr   movlw     Tx_SErr             ; Mark for syntax error

SendErr   call      SubSendW
          movlw     00h                 ; Resets Exe-, ACK-(, NACK-) and Cmd-Flags
          movwf     Cmd
          goto      RetNACK             ; Restarts Cmd receiving process!


;
;****************************************************************************
;
; Action Command routines
; -----------------------
; If a Cmd as an action part(*), it will no be performed once. After that
; the Cmd-Process waits for new Cmd.
;
; (*) The action part may include additional Data fetching, action execution
; and/or first CmdSub call.


TstActCmd btfsc     Cmd,Cmd_Sht         ; Shoot
          IF        Firma='B'
           goto     SErr                ;  Benninger: Not implemented
          ELSE
           goto     Sht_Cmd             ;  Sulzer: Yes
          ENDIF

          btfsc     Cmd,Cmd_GDX         ; Get Data ... ?
          goto      GDX_Cmd             ;  Yes

          btfsc     Cmd,Cmd_SOG         ; Set Offset/Gain ?
          goto      SOG_Cmd             ;  Yes

          btfsc     Cmd,Cmd_SNV         ; Get Error ?
          goto      SNV_Cmd             ;  Yes

                                        ; Extended commands ('xxxT 0000')?
          goto      EXT_Cmd             ; Spezial routine

                                        ; SyntaxError when Cmd = 00h.

;
;--------------------------------------------------------------------
;
; Get data and execute direct action commands
;
;
;

          IF        Firma='S'
                                        ; Shoot X
Sht_Cmd    movlw    0f0h                ; =======
           andwf    ChNivSel,F          ; Mask and clear Ch---Sel
           movf     Cmd,W               ; Get X of Cmd
           andlw    7h                  ; Mask sensor number 0..7
           addwf    ChNivSel,F          ; Set Channel Selection to low nibble

           call     SubRetACK           ; Set ACK-Data to TXProcess
                                        ; Calc first data immediatly
Sht_Ret    call     GBufW               ; Gets buffer into W1 according to ChNivSel
           movwf    W1
           bsf      Cmd,Cmd_Exe         ; Set Exe Flag
           goto     RetTAC              ; goto GetHdr
          ENDIF

;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

                                        ; Get Data X
                                        ; ==========
GDX_Cmd   btfsc     Cmd,GDX_All         ; IF GD-All THEN
          goto      SetACKEXE           ;  Get all

          movlw     0f0h                ; ELSE (just one channel)
          andwf     ChNivSel,F          ;  Mask and clear Ch---Sel
          call      SubInpExe           ;  Wait for next data
          movf      RxData,W            ;  Get Data into W
          bcf       ADCON0,DataRdy      ;  Free RxProcess
          btfsc     Cmd,Cmd_NACK        ;  Is there a NACK?
          goto      RetTAC              ;  Restart CmdExecution

GDOne     andlw     0Fh                 ;  Mask channel     (0-Fh)
          addwf     ChNivSel,F          ;  Save data into ChNivSel(0-Fh)
          goto      SetACKEXE           ;  Set Cmd_Exe,Cmd_ACK flags and goto GetHdr
                                        ; ENDIF

;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

                                        ; Set Offset/Gain
                                        ; ===============
SOG_Cmd   clrf      W1                  ; Set counter value for 8 data into W1

LoopSOG   call      SubInpExe           ; Wait for next data
          movf      RxData,W            ; Get Data into W
          bcf       ADCON0,DataRdy      ; Free RxProcess
          btfsc     Cmd,Cmd_NACK        ; Is there a NACK?
          goto      RetTAC              ; Restart CmdExecution

          movwf     W2                  ; Save data into W2
          bcf       INTCON,GIE          ; Disable Interrupt because ...
          movf      FSR,W               ; ... of FSR usage
          movwf     W3                  ; Save FSR into W3
          btfss     Cmd,SOG_Off         ; If NOT Offset then
          movlw     Gain0               ;   Get Gain addr
          btfsc     Cmd,SOG_Off         ; If Offset then
          movlw     Off0                ; Get Offset addr
          addwf     W1,W                ; Select right channel (8..1)
          movwf     FSR                 ; Set this address as indirect address
          movf      W2,W                ; Get data
          movwf     INDADR              ; Store data into indirect address
          movf      W3,W                ; Restore old FSR
          movwf     FSR
          bsf       INTCON,GIE          ; Free interrupts

NextSOG   incf      W1,F                ; Do it 8 times
          btfss     W1,3h               ; Is bit 3 set (=8)?
          goto      LoopSOG             ;   No: loop
          goto      SetACK              ;   Yes: Set Cmd_ACK flags and goto GetHdr

;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

                                        ; Set Niveau
                                        ; ==========
SNV_Cmd   call      SubInpExe           ; Wait for next data
          movf      RxData,W            ; Get Data into W
          bcf       ADCON0,DataRdy      ; Free RxProcess
          btfsc     Cmd,Cmd_NACK        ; Is there a NACK?
          goto      RetTAC              ; Restart CmdExecution

          movwf     W1                  ; Save niveau into W1
          movlw     0fh                 ; Mask Select
          andwf     ChNivSel,F          ; Clear Niveau
          rlf       W1,W                ; Niveau only bit 6-3 are used->into high nibble
          andlw     0f0h                ; Mask high nibble
          addwf     ChNivSel,F          ; Set Niveau into ChNivSel
          goto      SetACK              ; Set Cmd_ACK flags and goto GetHdr


;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

                                        ; Extended commands
                                        ; =================
EXT_Cmd   call      SubInpExe           ; Wait for next data
          movf      RxData,W            ; Get Data into W
          bcf       ADCON0,DataRdy      ; Free RxProcess
          btfsc     Cmd,Cmd_NACK        ; Is there a NACK?
          goto      RetTAC              ; Restart CmdExecution

          movwf     W1                  ; Save Ext into W1

                                        ; Test Ext action
TstActExt                               ; ---------------
          btfsc     W1,EXT_SCX          ; SelectChannel-X ?
          goto      SCX_Ext             ;  Yes

          btfsc     W1,EXT_ZRO          ; ZERO ?
          goto      ZRO_Ext             ;  Yes

          btfsc     W1,EXT_GER          ; Get Error ?
          goto      GER_Ext             ;  Yes

                                        ; ELSE SyntaxError...
SErr      bsf       Cmd,Cmd_NACK        ; Imaginary set NACK for Syntax error
          bcf       STATUS,RxError      ; Reset Rx Errors..
          bcf       STATUS,RxBusy
          goto      RetTAC              ; (no other flag is set)

;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

                                        ; SelectChannel-X
                                        ; ===============
SCX_Ext   movlw     0F0h                ; Mask Sel=0
          andwf     ChNivSel,F          ; Clr Sel
          movf      W1,W                ; Get Channel
          andlw     07h                 ; Mask Selected channel
          addwf     ChNivSel,F          ; Put into ChNivSel

          rlf       W1,W                ; Get channel x2
          andlw     b'00001110'         ; mask channel
          movwf     W2                  ; Save into W2
          swapf     W2,F                ; put into high nibble

SCXWait   bsf       INTCON,GIE
          movf      PORTB,W             ; get Mux channels
          andlw     b'11100000'         ; Mask mux
          bcf       INTCON,GIE
          xorwf     W2,W                ; Test Mux value
          skpz
          goto      SCXWait

SCXCatch  bsf       INTCON,GIE
          bcf       INTCON,ADIE         ; AD disable
          goto      SCXEnd

SCXEnd    goto      SetACKEXE           ; Set Cmd_Exe,Cmd_ACK flags and goto GetHdr


;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

                                        ; ZERO
                                        ; ====
ZRO_Ext   nop                           ; Set MEAN-Flag
;         goto      SetACKEXE           ; Set Cmd_Exe,Cmd_ACK flags and goto GetHdr


;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

                                        ; Get Error
                                        ; =========
GER_Ext
;         nop                           ; "No action"
;         goto      SetACKEXE           ; Set Cmd_Exe,Cmd_ACK flags and goto GetHdr


;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

SetACKEXE bsf       Cmd,Cmd_Exe         ; Set Exe Flag
SetACK    nop                           ; Finish command action with ACK and Exe flag set

          IF        _RS=485             ; Wait for Cmd-Protokoll to be finish
           call     SubInpExe           ; Wait for Cmd-tail

           movlw    ~Rx_Tail+1          ; Test for Cmd-tail (FEh)
           addwf    RxData,W            ; IF tail->FEh+2=00h->C ....
           movf     RxData,W
           bcf      ADCON0,DataRdy      ; Free RxProcess
           skpnc                        ; Was it Rx-Tail?
           goto     SetA_               ;   Yes: set ACK

           movlw    Rx_Hdr
           subwf    RxData,W            ; IF Rx_Hdr->FFh-FFH=00H->C ....
           skpc                         ;   Yes: Leave Error, get Adress
           goto     RetTAC              ;   No:  Return to Hdr

           movlw    00h                 ; Leave Error, read new Adress!
           movwf    Cmd
           goto     GetAdr              ;
          ENDIF

SetA_     bsf       Cmd,Cmd_ACK         ; Set ACK Flag
          goto      RetTAC


          IF        _RS=485             ; [ Listen All
;
;****************************************************************************
;
; Listen All BRANCH;
;
;
                                        ; Wait for LACmd
                                        ; ==============
ListenAll  movlw    00h                 ; Clear Cmd for new Command. This stops ExeSubs!
           movwf    Cmd

           call     SubInpExe           ; Wait for LACmd

           movfw    RxData              ; Get RxData
           bcf      ADCON0,DataRdy      ; Free RxProcess
           btfss    Cmd,Cmd_NACK        ; Is there a NACK?
           goto     TestLA              ;  No...continue
           call     TestCut             ; Test for separate cut flag output
           skpc                         ;  C set for cut
           goto     TestLA              ;  No: go ahead

           movlw    b'01111'            ; Preset for DChPin output
           btfss    PORTA,DChPin        ; Test for not active Daisy Chain
           goto     TestLA              ;  No: go ahead
           tris     TRISA               ;  Yes:   Set Daisy chain pin
           bcf      Cmd,Cmd_NACK        ;       Clear flag (Don't send any NACK!...)

                                        ; (Default is '001x xxxx')
TestLA     movwf    Cmd                 ; Save Cmd
           btfsc    Cmd,07h             ; Test for Syntax Error
           goto     RetTALAC
           btfsc    Cmd,06h             ; Test for Syntax Error
           goto     RetTALAC
           btfss    Cmd,05h             ; Test for Syntax Error
           goto     RetTALAC
                                        ; No SyntaxError:
           movlw    b'00011111'         ; Clear first 3 bits anyway...
           andwf    Cmd,F               ; ... of Cmd
           goto     TstActLAC           ; Test for action command now!

RetTALAC   goto     GetHdr              ; Error return with no mark of SyntaxError

;
;--------- -------------------------------------------------------------------
;
; SUB: Act ion Command routines
; -------- --------------------
; If a Cmd  as an action part(*), it will no be performed once. After that
; the Cmd- Process waits for new Cmd.
;
; (*) The  action part may include additional Data fetching, action execution
; and/or f irst CmdSub call.


TstActLAC  btfsc    Cmd,LAC_GER         ; LACmd GetError ?
           goto     LAGER_Out           ;  Yes

           bsf      Cmd,Cmd_NACK        ; Imaginary set NACK for Syntax error
           goto     RetTALAC            ; (no other flag is set)

;
; - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
                                        ; Get Error
                                        ; =========
LAGER_Out  call     TestCut             ; Test for separate cut flag output
           skpc                         ;  C set for cut
           goto     RetTALAC            ; No cut: => Return without any output
                                        ; Cut is marked: finish normal Cmd GER...
           movlw    b'00000001'         ; Cmd_GER
           movwf    W1                  ; Save it into ExtCmd W1
           movlw    b'10010000'         ; Ext_Cmd and EXE-Flag
           movwf    Cmd                 ; Store EXT into Cmd because of GER_Out
;          goto     SetLAE_A_           ; Set EXE and ACK Flag, return


;- - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

SetLAE_A_  bsf      Cmd,Cmd_Exe         ; Set Exe Flag
SetLAACK                                ; Finish command action with ACK and Exe flag set

           IF       Firma='B'           ; Wait for Cmd-Protokoll to be finish
            call    SubInpExe           ; Wait for Cmd-tail

            movlw   ~Rx_Tail+1          ; Test for Cmd-tail (FEh)
            addwf   RxData,W            ; IF tail->FEh+2=00h->C ....
            movf    RxData,W
            bcf     ADCON0,DataRdy      ; Free RxProcess
            skpnc                       ; Was it Rx-Tail?
            goto    TestErr             ;   Yes: send Errorcode

            movlw   Rx_Hdr
            subwf   RxData,W            ; IF Rx_Hdr->FFh-FFH=00H->C ....
            skpc                        ;   Yes: Leave Error, get Adress
            goto    RetTALAC            ;   No:  Return to Hdr

            movlw   00h                 ; Leave Error, read new Adress!
            movwf   Cmd
            goto    GetAdr              ;
           ENDIF

SLAA_      bsf      Cmd,Cmd_ACK         ; Set ACK Flag
           goto     RetTALAC


          ENDIF                         ; Listen all ]
;
;--------------------------------------------------------------------
;
; SubInpExe-Roundabout
; ----------------------
; Tests several Flags and desires. If an event occurs, the according
; routine may be executed:
; If Rx data is ready return to Cmd-Process
; If Cmd_NACK is set return to Cmd-Process to send NACK and restart Cmd-Proc.
; If Cmd_ACK is set Tx process is loaded with ACK data
; If Tx process is free, last computed data is sent and new data is calcul.
; If Cmd_Test flag is set, a Debugger data is added to be sent

                                        ; Check data input, output read and executes
                                        ; waiting data output sub.
                                        ; ==========================================

SubInpExe btfsc     ADCON0,DataRdy      ; Test if data is received?
Ret1      return                        ; If DataRdy from Rx return to Cmd Execution

          btfsc     Cmd,Cmd_NACK        ; Is there an error pending (NACK)?
Ret2      return                        ; return to Cmd_Exe routines, data will be ejected!

          btfsc     Cmd,Cmd_ACK         ; Is there a ACK? Test before other data is sent
          call      SubRetACK           ; Sub to send ACK...

          btfsc     ADCON0,DataRdy      ; Test if data is received?
Ret3      return                        ; If DataRdy from Rx return to Cmd Execution

          btfsc     Cmd,Cmd_NACK        ; Is there a NACK?
Ret4      return                        ; return to Cmd_Exe routines, data will be ejected!

          btfss     PStat,TxFlag        ; Is TxProcess free?
          call      SubTstOut           ; computed data output and calc new data

          IF        _WDT=Yes
           clrwdt                       ; Reset watch dog timer
          ENDIF
          goto      SubInpExe           ; LOOP: New Data ready test


;
;--------------------------------------------------------------------
;
; SubRetACK
; ---------
;
                                        ; Return ACK data
                                        ; ===============
SubRetACK btfsc     Cmd,Cmd_NACK        ; Did NACK occur ?
          return                        ; Yes

          movlw     Tx_ACK              ; Load Tx process with ACK
          call      SubSendW            ; Send w via Tx
          bcf       Cmd,Cmd_ACK         ; Reset ACK-Flag
          return

;
;--------------------------------------------------------------------
;
; SubSendW
; --------
;
; TxFlag is tested befor sending W
;
SubSendWT btfsc     Cmd,Cmd_Tst         ; If Cmd_Tst flag is set, send ASCII-HEX
          goto      SubSndWHx
SubSendW  btfss     PStat,TxFlag        ; Wait till Tx free
          goto      SSendW
                                        ; #
          IF        _WDT=Yes
           clrwdt                       ; Reset WDT
          ENDIF

          goto      SubSendW
SSendW    movwf     TxData
          bsf       PStat,TxFlag        ; Sets command to start TxProcess
          return

;
;--------------------------------------------------------------------
;
; SubSndWHx
; ---------
;
; TxFlag is tested befor sending W as two ASCII values
;

SubSndWHx movwf     W1                  ; push
          swapf     W1,W                ; Get high nibble
          call      SendHalf            ; Send high nibble
          movf      W1,W                ; Get low nibble
SendHalf  andlw     0Fh                 ; Clear high
          addlw     30h                 ; Add ASCII '0'
WaitFree  btfss     PStat,TxFlag        ; Wait till Tx free
          goto      HxFree
                                        ; #
          IF        _WDT=Yes
           clrwdt                       ; Reset WDT
          ENDIF

          goto      WaitFree
HxFree    movwf     TxData              ; Send low nibble
          btfss     TxData,3            ; Is >=8 ?
          goto      SndHlf              ;   no: send
          btfsc     TxData,2            ; Is >=0Ch ?
          goto      AddHx               ;   Yes: add correction
          btfsc     TxData,1            ; Is >=0Ah ?
          goto      AddHx               ;   Yes: add correction
          goto      SndHlf              ;   Yes: send

AddHx     movlw     7
          addwf     TxData,F            ; Correct A-F
SndHlf    bsf       PStat,TxFlag        ; Sets command to start TxProcess
          return

;
;--------------------------------------------------------------------
;
; SubTestDataOut
; --------------
;
; (TxFlag is cleared at first data)
;

                                        ; Test if any Cmd has to output data
                                        ; ==================================
SubTstOut btfss     Cmd,Cmd_Exe         ; Is Command ready to be executed?
          return                        ;  No: go back

          btfsc     Cmd,Cmd_ACK         ; Is ACK to be sent?
          return                        ;  Yes: go back

          IF        Firma='B'
           btfss    Cmd,Cmd_Sht         ; Shoot X ?
           goto     STOCont             ;  No: Continue w. test
           bsf      Cmd,Cmd_NACK        ;  Yes for Benninger: Set Syntax Error
           return
          ELSE
           btfsc    Cmd,Cmd_Sht         ; Shoot X ?
           goto     Sht_Out             ;  Yes
          ENDIF

STOCont   btfsc     Cmd,Cmd_GDX         ; Get Data X calcs ?
          goto      GDX_Out             ;  Yes
          btfsc     Cmd,Cmd_SOG         ; Set Offset/Gain ?
          return                        ;  Yes: not needed
          btfsc     Cmd,Cmd_SNV         ; Set Niveau ?
          return                        ;  Yes: not needed
          goto      TstExtOut           ; ELSE Extended commands

STOSErr   bsf       Cmd,Cmd_NACK        ; Set Syntax Error
          return


;
;--------------------------------------------------------------------
;
; Execute-SubProcesses
; --------------------
; During SubInpExe-time with no Input this Exe-Subs will be started
; according to the loaded command

          IF        Firma='S'           ; Exe Sub Shoot X
                                        ; ===============
Sht_Out    movf     W1,W                ; Get old data
           call     SubSendWT           ; Send W via Tx
           call     GBufW               ; Gets buffer into W1 according to ChNivSel
           movwf    W1
           goto     TstDbg
          ENDIF

;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

                                        ; Exe Sub GetDataValue/Mean
                                        ; =========================
GDX_Out   btfsc     Cmd,GDX_All         ; IF GD-All THEN
          goto      OutAll              ;  Out all



                                        ; Out One value
                                        ; =============
OutVal    btfsc     ChNivSel,3          ; IF Channel < 7 THEN
          goto      OneOG
                                        ;  Get One Channel (Mean/Value)
                                        ;  ----------------------------
OneCh     call      GBufW               ;  Get Channel value into W
          goto      SendOne             ; ELSE

                                        ;  Get Offset or Gain
                                        ;  ------------------
OneOG     btfsc     Cmd,GDX_Val         ;  IF Val => Offset
          call      GetOW               ;    Get Offset value into W
          btfss     Cmd,GDX_Val         ;  ELSEIF Mean => Gain
          call      GetGW               ;    Get Gain value into W
                                        ;  ENDIF

SendOne   call      SubSendWT           ; Send W via Tx
          goto      EndGDX



                                        ; Out All of the Channels
                                        ; =======================
OutAll    movlw     0f0h                ; Mask select nibble off...
          andwf     ChNivSel,F          ; ...ChNivSel, clear it=first channel

          btfss     Cmd,GDX_Val         ; IF ALL is selected AND
          goto      AllMean             ;   IF not MEAN is selected THEN

                                        ;   Out all data values(8)
                                        ;   ----------------------
AllValue  call      GBufW               ;   Get next channels data into W
          call      SubSendWT           ;   Send W via Tx
          incf      ChNivSel,F          ;   Next channel is selected
          btfss     ChNivSel,3          ;   Is value 8 reached ?
          goto      AllValue            ;    No: loop for next

          goto      EndGDX              ; ELSE (MEAN is selected)

                                        ;   Calc Mean of all values
                                        ;   -----------------------
AllMean   btfsc     PStat,TxFlag        ;   Wait till Tx free
          goto      AllMean
          call      GBufW               ;    Get First channels data into W
          skpnc                         ;    IF Error THEN
          goto      AMSend              ;      Send Under- or Overflow (W)
                                        ;    ENDIF
          movwf     TxData              ;    Add Result to buffer
          incf      ChNivSel,F          ;    Next channel is selected

AMLoop    call      GBufW               ;    Get next channels data into W
          skpnc                         ;    IF Error THEN
          goto      AMSend              ;      Send Under- or Overflow (W)
                                        ;    ENDIF
          incf      TxData,F            ;    Round up
          skpnz                         ;    Was an overflow?
          decf      TxData,F            ;      Yes: ignore it
          addwf     TxData,F            ;    Add Result to buffer
          rrf       TxData,F            ;    Calc Mean
          incf      ChNivSel,F          ;    Next channel is selected
          btfss     ChNivSel,3          ;    Is value 8 reached ?
          goto      AMLoop              ;     No: loop for next

          movf      TxData,W            ;   Get buffer into W
AMSend    call      SubSendWT           ;   Send W via Tx

;         goto      EndGDX              ; ENDIF

EndGDX    movlw     070h                ; End of Command (leave ACK,NACK & Tst)
          andwf     Cmd,F
          bsf       Cmd,Cmd_ACK         ; Send ACK at the End of the Commands
          goto      TstDbg

;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


TstDbg    btfss     Cmd,Cmd_Tst         ; IF Test-Flag is set a debug routine is added after each cmd!
          return
                                        ; Debug routine:
                                        ; ==============
TestEnd   movlw     0Dh                 ; Add CR/LF
          Call      SubSendW
          movlw     0Ah
          Call      SubSendW

          return


;---------------------------------------------------------------------------

                                        ; Extended commands
                                        ; =================
                                        ; Do quit subroutine with return (not goto TstDbg)
TstExtOut nop
          btfsc     W1,EXT_SCX          ; Set channel X ?
          goto      SCX_Out             ;  Yes

          btfsc     W1,EXT_ZRO          ; ZRO ?
          goto      ZRO_Out             ;  Yes

          btfsc     W1,EXT_GER          ; Get Error ?
          goto      GER_Out             ;  Yes

                                        ; SyntaxError when Ext = 00h.
          clrf      Cmd                 ; End of Ext
          bsf       Cmd,Cmd_NACK        ; Imaginary set NACK for Syntax error
          return

;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
                                        ; Set channel x
                                        ; =============
SCX_Out   movf      ADRES,W             ; Get AD result
          IF        _ADInv=Yes
           xorlw    0FFh                ; Invert result
          ENDIF
          call      SubSendWT           ; Send W via Tx

          movlw     070h                ; End of Command (leave ACK,NACK & Tst)
          andwf     Cmd,F
          bsf       Cmd,Cmd_ACK         ; Send ACK at the End of the Commands
          goto      TstDbgExt


;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
                                        ; ZERO
                                        ; ====
ZRO_Out   movf      Buf0,W              ; Get Buffer 0 value and ...
          movwf     Off0                ; stores it into Offset 0
          call      SubSendWT
          movf      Buf1,W              ; Get Buffer 1 value and ...
          movwf     Off1                ; stores it into Offset 1
          call      SubSendWT
          movf      Buf2,W              ; Get Buffer 2 value and ...
          movwf     Off2                ; stores it into Offset 2
          call      SubSendWT
          movf      Buf3,W              ; Get Buffer 3 value and ...
          movwf     Off3                ; stores it iNto Offset 3
          call      SubSendWT
          movf      Buf4,W              ; Get Buffer 4 value and ...
          movwf     Off4                ; stores it into Offset 4
          call      SubSendWT
          movf      Buf5,W              ; Get Buffer 5 value and ...
          movwf     Off5                ; stores it into Offset 5
          call      SubSendWT
          movf      Buf6,W              ; Get Buffer 6 value and ...
          movwf     Off6                ; stores it into Offset 6
          call      SubSendWT
          movf      Buf7,W              ; Get Buffer 7 value and ...
          movwf     Off7                ; stores it into Offset 7
          call      SubSendWT

          btfss     W1,ZRO_GRst         ; Test for gain reset
          goto      TstDbgExt

          movlw     80h                 ; Set Gain value to 1.0
          movwf     Gain0               ; stores it into Gain 0
          movwf     Gain1               ; stores it into Gain 1
          movwf     Gain2               ; stores it into Gain 2
          movwf     Gain3               ; stores it iNto Gain 3
          movwf     Gain4               ; stores it into Gain 4
          movwf     Gain5               ; stores it into Gain 5
          movwf     Gain6               ; stores it into Gain 6
          movwf     Gain7               ; stores it into Gain 7

          movlw     070h                ; End of Command (leave ACK,NACK & Tst)
          andwf     Cmd,F
          bsf       Cmd,Cmd_ACK         ; Send ACK at the End of the Commands
          goto      TstDbgExt


;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
                                        ; Get Error
                                        ; =========
GER_Out   nop
          IF        Firma='B'
           movlw    Rx_Addr             ; Send own Address
           call     SubSendWT           ; Send W via Tx
          ENDIF

          call      TestCut             ; Test for separate cut flag output
          skpnc                         ;  C set for cut
          goto      GERCut              ;  Yes
                                        ; ELSE

GERNErr   movlw     Tx_NErr-0D0h        ; No error is there (Errors are coded 2xh)
          call      SubSendWT           ; Send W via Tx
          goto      EndGER


GERCut    movlw     Tx_Cut-0D0h         ; Send data for Cut (Error is coded 2xh)
          call      SubSendWT           ; Send W via Tx

          clrf      W1                  ; Init loop counter
GCLoop    clrc
          rrf       ChCut,F             ; Get next carry
          skpc                          ;  If carry then send channel
          goto      NoCh
          movf      W1,W                ; Get channel
          addlw     30h                 ; Make ASCII
          call      SubSendWT           ; Send W via Tx

NoCh      incf      W1,F                ; point to next channel
          btfss     W1,3                ; is loop finish ?
          goto      GCLoop              ; no

          bcf       PORTB,CutPin        ; Reset output pin Cut
          movlw     b'11111'            ; High imp. for DChPin output(=input)
          tris      TRISA               ; Reset DChPin for not active Daisy Chain
;         goto      EndGER              ;


EndGER    movlw     070h                ; End of Command (leave ACK,NACK & Tst)
          andwf     Cmd,F
          bsf       Cmd,Cmd_ACK         ; Send ACK at the End of the Commands
          goto      TstDbgExt


;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

TstDbgExt btfss     Cmd,Cmd_Tst         ; IF Test-Flag is set a debug routine is added after each cmd!
          return
                                        ; Debug routine:
                                        ; ============

          rrf       ChNivSel,W          ; ChNiv___
          andlw     78h
          call      SubSndWHx           ; Send W as Hex

          btfss     INTCON,ADIE         ; Test if one is selected
          goto      OneSel              ; If yes: send Channel

Running   movlw     3Fh                 ; No: Send 3Fh...
          goto      SndSel

OneSel    movfw     ChNivSel            ; Ch___Sel
          andlw     07h
          addlw     30h
SndSel    call      SubSndWHx           ; Send W as Hex

          goto      TestEnd

;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
__End     data      'ENDEENDE'
          END
