;
;                                 LPGS.ASM
;         "Laser Pattern Generator Shutter" based on:
;          -MpK 0.9- with interrupt handler for PIC16LC71
;
;*******************************************************************
;
;Creator : M.Kellenberger, 18.11.93, NTB-Buchs CH
;System  : Boxer Editor, MPALC, MPSTART
;
;History MpK:
;V 0.0    21.10.93 mik        Systemlabel start with "_"!
;V 0.1    22.10.93 mik        System design and description
;V 0.1a   25.10.93 mik        Add "#" to user-changeable lines
;V 0.1b   15.11.93 mik        Made systembody faster (20MHz)
;V 0.1c   16.11.93 mik        MpK Kernel via simulater tested!
;V 0.2    25.11.93 mik        RxError, RxBusy and Cmds!
;V 0.3    13. 1.94 mik        RS-484 Enable implemented
;V 0.4    21. 1.94 mik        Compiler flags _Tst,_MHz,_RS implemented
;V 0.5    27. 1.94 mik        Compiler flags _Tst into TstDbg implemented
;V 0.6    28. 1.94 mik        Compiler flags _WDT implemented
;V 0.7     8. 2.94 mik        Ext_Cmd is 001T'0000 (Flag T is no also for EXT Cmds)
;         10. 2.94 mik        reset condition into Cmd
;V 0.8    21. 2.94 mik        CRC added, _Peven, _Parity deleted
;V 0.9    21. 2.94 mik        CRCs/r added
;
;User programmers revision:
;V 1.0a   13. 1.94 jw         First attempt
;V 1.0b   27. 1.94 jw,mik     MpK corrected, single RS-422 selected
;V 1.0c   31. 1.94 jw,mik     MpK "clrf Cmd" corrected, RCV expanded, _IntPORTRB corrected
;V 1.1    31. 1.94 mik        RCV expanded, _IntPORTRB corrected, ADCON1 set to 11
;                             MP programmed
;V 1.2     7. 2.94 jw,mik     SubSendHs implemented, SSV corrected
;                             MP programmed
;V 1.3    14. 2.94 jw,mik     Tested, ShttrStrt/-Rst implemented
;         18. 2.94 mik        MP programmed
;V 1.4    21. 2.94 mik        SMain,SInt deleted,CRC implemented,SCV Full corrected
;         18. 2.94 mik        MP programmed
;V 1.5    23. 2.94 mik        Axis selected in _IntPORTRB via OldPRB, Cmd RST implemented
;                             MP programmed
;V 1.6    24. 2.94 mik        RCV expanded, CV8hl and CV9hl dismissed
;                             MP programmed
;V 1.7     1. 3.94 mik        SCN implemented, MP programmed
;V 1.8     1. 3.94 mik        SCN changed,SGO implemented, MP programmed
;
;*******************************************************************
;
; Description:
; ============
;   1. Multi process Kernel System body:
;   ------------------------------------
;         The MpK consists of
;           predefined register- and macrosets
;           initialization of HW and status/option registers
;           one (empty) MAIN (user) process,
;           an INTerruptHANDLER, which branches to the 4
;           handling processes of each interrupt source.
;           RTCC-int handler cycles through 4 processes: Tx, Rx0, ?? and Rx1.
;           and may launch the AD-conversion repeatedly.
;
;         All this empty modules may be filled by the user
;         (programmer) and described in the following paragraph!
;
;   2. LPGS
;   -------
;         Kontroller zur Shuttersteuerung und Positionsabfrage.
;         Die Digitalsignale von bis zu 4 Achsen werden berwacht und
;         eine davon ausgewhlte Achse beeinflusst einen Zhler.
;         Zur Kommunikation mit dem uP stehen verschiedene Befehle zur verfgung:
;
;         Einfache Befehle:
;         - Mit dem Befehl SSV werden die Startkoordinaten zum uP gesendet. Dieser
;           setzt den Zhler, startet die berwachung der gewhlten Achse und
;           setzt den Shutter entsprechend aus (oder ein).
;
;         - Mit dem Befehl SCV werden die Shutter-Invertier-Koordinaten zum uP
;           gesendet. Diese werden in einen Buffer gelegt und dienen als
;           Vergleichswerte fr den Zhler, um bei Gleichheit den Shutter zu
;           invertieren. Bis zu neun Koordinaten knnen vorab gesendet werden.
;
;         - Mit dem Befehl RCV knnen der Zhlerstand und gewisse Flags abgefragt
;           werden. Weiter wird ein selektiertes (Bit0-2) Koordinatenwertepaar
;           zurckgesendet.
;           Mittels dem gesetzten Testbit im Befehl wird die Art der
;           Rckmeldung bestimmt: Testbit reset = 4 Byte Hex CnTH,CntL,Cnt,Flags
;                                 Testbit set = 8/10 Byte ASCII ",",Cnt,Flags (CR/LF)
;                                 (Rem.:
;                                        Abhngig vom Assemblerfalg _Tst ist CR/LF angefgt!)
;
;         - Der Befehl RST setzt alle Zhler und den Buffer (8) auf Null, schliesst
;           den Shutter und sperrt die berwachungseingnge.
;
;         LaserPaGe Spezial-Befehle:
;         - Der Befehl SCN ist speziell auf die LaserPaGe Maschine abgestimmt,
;           um schnell und sicher die 2 Shutterwechselpositionen bermitteln
;           zu knnen. Das zurcksenden der Daten erhht die Sicherheit.
;
;         - Der Befehl SGO gibt nach "SCN" die selektierte Achse zur berwachung frei.
;
;         Verbesserungen:
;         - Shutterfreigabe (Port=Input) durch Befehl RST.
;         - Bestimmung des Shutterzustandes durch internes Flag.
;
;
;
; Notes to the source:
; --------------------
;         1. Labels with the starter "_" are system labels
;         2. # mark changeable values
;         3. Compiler/Assembler vars are:
;            _RS      = 232/485
;            _RxInv   = Yes/No
;            _TxInv   = Yes/No
;            _MHz   = 20,16,...
;            _Tst   = Yes/No
;            _WDT   = Yes/No
;
;*******************************************************************
;
;Assembler options:
;

          LIST C=100,E=0,N=60,F=INHX8M,P=16C71,R=DEC,L=ON

Yes       set       1         ; Do use this for compiler directives!
No        set       0         ; Do not use "YES" or "NO" for them!

_Tst      set       No        ; Yes for test sample:  (V0.8)
                              ;         Adds 0Dh0Ah if testflag in Cmd is set
                              ;         No CRC needed
                              ;         Reset conditions are sent after every reset
                              ; No for serie!

_WDT      set       Yes       ; Yes for active WDT and reset sending code
                              ; No for no WDT!

_MHz      set       16        ; 20 for 20MHz quarz connection
                              ; 16 for 16MHz quarz connection
;
;*******************************************************************
;
;
; ID VALUES                   ; Burn this within "MPSTART.EXE"
;
; 2000h   3FFFh               ; Code Revision Number
; 2001h   3FFFh               ; Code Manufacturing Number
; 2002h   3FFFh               ; Code ...
; 2003h   3FFFh               ; Code ...
;
;CONFIGURATION WORD: (2007H)
;
;Select   FOSC1     FOSC0     ; OSC Selection Fuses
;         0         0         ;  LP
;         0         1         ;  XT
;  >      1         0         ;  HS _MHz=20..5
;         1         1         ;  RC (default)
;
;         WDTE                ; WDT Enable Fuse
;  #      0                   ;  disable
;         1                   ;  enable (default)
;
;         PWRTE               ; Power Up Timer Fuse
;         0                   ;  disable
;  >      1                   ;  enable (default)
;
;         CP                  ; Code Protection Fuse
;         0                   ;  On
;  >      1                   ;  Off (default)
;
;
;*******************************************************************
;
;Includes
;

          INCLUDE   "PicReg.equ"
          INCLUDE   "Pic71.mac"

;
;*******************************************************************
;
;Constants
;

Presc     equ       b'000'              ; Prescaler Value   000=1:2
                                        ;                   001=1:4
                                        ;                      =
                                        ;                   111=1:256

          if        _MHz=20             ; RTCC set value (minus corr. cycles for HW (2)delay)
Rtcc      equ       (130)-2             ; ...for 20MHz
          else
Rtcc      equ       (104)-2             ; ...for 16MHz
          endif
                                        ; Calculation: (# for serial if baudrate)
                                        ; ------------
                                        ; >>>>>>>>> 16MHz t=0.25us <<<<<<<<<<<
                                        ;104 * 0.25us  = 26us
                                        ;  4 * 26us    =104us => 9.6kBaud
                                        ; ##

                                        ; 52 * 0.25us  = 13us
                                        ;  4 * 13us    = 52us => 19.2kBaud

                                        ; 26 * 0.25us  = 6.5us
                                        ;  4 *  6.5us  = 26us => 38.4kBaud

                                        ; >>>>>>>>> 20MHz t=0.2us <<<<<<<<<<<<
                                        ; 65 *  0.2us  = 13us
                                        ;  4 * 13us    = 52us => 19.2kBaud

                                        ; 32 *  0.2us  = 6.4us (adjust to 6.5us)
                                        ;  4 *  6.5us  = 26us => 38.4kBaud


;Adj      equ       (0)+2               ; Adjust value (X * 0.2us plus 2)

;SRate    equ       100h-(5)*10h        ; Sample rate for (multiplexed) AD-conversion
                                        ;
                                        ; Calculation:
                                        ; ------------
                                        ; Process timeslice t=13us
                                        ;  4 *  13us => (19,2kHz)
                                        ;  5 *  13us => (15,4kHz)
                                        ; ##
                                        ;  6 *  6.5us => (13,9kHz)

                                        ; Process timeslice t=6.5us
                                        ;  8 *  6.5us => (19,2kHz)
                                        ; 10 *  6.5us => (15,4kHz)
                                        ; 11 *  6.5us => (13,9kHz)

PCnt0     equ       00h                 ; PStat Reg. Process counter bit 0
PCnt1     equ       01h                 ; PStat Reg. Process counter bit 1
RxFlag    equ       02h                 ; PStat Reg. 1=Rx is active
TxFlag    equ       03h                 ; PStat Reg. 1=Tx is active
                                        ; PStat Reg. Bit 4-7: Countdown for sample rate

DataRdy   equ       ADCnusd             ; DataReady Flag in reg. ADCON0:
                                        ; # Must be reset via user
                                        ;   when data is read !!!

RxStart   equ       00h                 ; TxRxFlg Reg. 1=Rec. start bit
RxStop    equ       01h                 ; TxRxFlg Reg. 1=Rec. stop bit
RxSel     equ       02h                 ; TxRxFlg Reg. Select Rx process 0=01,1=11
TxStart   equ       03h                 ; TxRxFlg Reg. 1=Tx is active
TxStop    equ       04h                 ; TxRxFlg Reg. 1=Tx is active
                                        ; TxRxFlg Reg. Bit 5-7: bit counter Tx

                                        ; Bits in STATUS (Main!) or SaveST (Interrupt routine!):
RxError   equ       RP1                 ; Flag to mark Rx Error start/stop/parity
RxBusy    equ       IRP                 ; Flag to mark Rx busy (INTterupt while DataRdy=1)


                                        ; Port Pins:
                                        ; ----------

                                        ; PORTB:
                                        ; ------
RxPin     equ       00h                 ; Pin RS-232-C Receiver
TxPin     equ       01h                 ; Pin RS-232-C Transmitter

ShttrRst  equ       02h                 ; Pin Shutter Reset
ShttrStrt equ       03h                 ; Pin Shutter Start
CHBX      equ       04h                 ; Pin CHB Trig. (IntPORTRB)
CHBY      equ       05h                 ; Pin CHB Trig. (IntPORTRB)
CHBZ      equ       06h                 ; Pin CHB Trig. (IntPORTRB)
CHBR      equ       07h                 ; Pin CHB Trig. (IntPORTRB)

                                        ; PORTA:
                                        ; ------
CHAX      equ       00h                 ; Pin CHA Direction X
CHAY      equ       01h                 ; Pin CHA Direction Y
CHAZ      equ       02h                 ; Pin CHA Direction Z
CHAR      equ       03h                 ; Pin CHA Direction R


                                        ; Protokol:
                                        ; =========

                                        ; RxData:
                                        ; -------
Rx_Hdr    equ       0FFh                ; Unique data on Rx line to indicate Cmd start
                                        ; ( It must be 0FFh !!)

                                        ; Cmd bits: (Compared in the order bit 4 to bit 0)
                                        ; ==============================================

                                        ; Process flow bits:
                                        ; ------------------
Cmd_Exe   equ       07h                 ;  Command is executed
Cmd_NACK  equ       06h                 ;  NACK must be sent immediatly
Cmd_ACK   equ       05h                 ;  ACK must be sent immediatly

                                        ; Command bits (001x xxxx):
                                        ; -------------------------
Cmd_EXT   equ       04h                 ;  Extended Command:1Xh=Test of the following
Cmd_Tst   equ       04H                 ;                       routines by callinf TstDbg routine
                                        ;                       after every 'Data-out'
                                        ;                   10h=Extended commands:

Cmd_RCV   equ       03h                 ;  ReturnStatus;OK/NACK
                                        ;    Bit 0-2 are pointer to desired buffer

Cmd_SCV   equ       02h                 ;  SetCoordinateValue(Vh,Vl);OK/NACK

Cmd_SSV   equ       01h                 ;  SetStartValue01(Axis,Vh,Vl);OK/NACK
SSV_Sht   equ       00h                 ;    Shutter start value(0/1)

Cmd_RST   equ       00h                 ;  FreeShutter;OK/NACK

Ext_SCN   equ       01h                 ;  Scan01(Axis,CntH/L,CV1H/L,CV2H/L);OK/NACK,...
SCN_Sht   equ       00h                 ;    Shutter start value(0/1)

Ext_SGO   equ       00h                 ;  SGO;OK/NACK,...

                                        ; Reset conditions:
                                        ; -----------------
_RstC     equ       'R'                 ; Reset
_WdtSlpC  equ       'S'                 ; WDT occured in sleep mode
_WdtToC   equ       'T'                 ; WDT occured during running
_MClrSC   equ       'M'                 ; Reset pin was pulled in sleep mode

                                        ; TxData:
                                        ; -------
Tx_Rst    equ       080h+_RstC          ; Reset
Tx_WDTSLP equ       080h+_WdtSlpC       ; WDT occured in sleep mode
Tx_WDTTO  equ       080h+_WdtToC        ; WDT occured during running
Tx_MCLRS  equ       080h+_MClrSC        ; Reset pin was pulled in sleep mode

Tx_SErr   equ       0FEh                ; NACK: Syntax Error or CRC occured (V0.8)
Tx_RxBusy equ       0FDh                ; NACK: Rx is still busy while new data is pushing
Tx_RxErr  equ       0FCh                ; NACK: Rx start/stop bit error occured
Tx_ACK    equ       0FBh                ; ACK:  ''
Tx_OflErr equ       0FAh                ; NACK: Buffer overflow

Tx_DataMx equ       0FAh                ; 00h-FAh=data


;
;*******************************************************************
;
;Variables
;

SaveW     equ       0Ch                 ; Interrupt save register W
SaveST    equ       0Dh                 ;    "       "   STATUS
RxData    equ       0Eh                 ; Received byte
TxData    equ       0Fh                 ; Byte to send
TxRxFlg   equ       10h                 ; Tx & Ryx Flags and counter
CRCr      equ       11h                 ; CRC Test byte receive (V0.8)
CRCs      equ       12h                 ; CRC Test byte send (V0.8)
Cmd       equ       13h                 ; Command register:           00h=no cmd
ExtCmd    equ       14h                 ; Extended Cmd File,else
PStat     equ       15h                 ; Process register:
                                        ; Bit
                                        ; 0,1:    RTCC Process Counter:         00=Tx
                                        ;                                       01=[Rx]
                                        ;                                       10=#(Adjust)
                                        ;                                       11=[Rx]
                                        ; 2,3:    Rx/TxFlags: 1 if receiving/sending
                                        ; 4:      Rx process pointer 0=RTCC proc 01
                                        ;                                1=RTCC proc 11
                                        ; 5-7:    Sample rate count down
                                        ;
FSRMain   equ       16h                 ; Save for FSR during IntHandler processes
FSRInt    equ       17h                 ; Save for FSR during Main process

CntH      equ       18h                 ; Counter High
CntL      equ       19h                 ; Counter Low
Cnt       equ       1Ah                 ; Counter for 10 sub-steps

Flags     equ       1Bh                 ; Axis, Ofl Flags, WDT
OldPRB    equ       1Ch                 ; Old PortB Values

W1        equ       1Dh                 ; Work registers
W2        equ       1Eh

Shutter   equ       1Fh                 ; Future..

CV0h      equ       20h                 ; Coordinate value 0 for Shutter high
CV0l      equ       21h                 ; Coordinate value 0 for Shutter low
CV1h      equ       22h                 ; Coordinate value 1 for Shutter high
CV1l      equ       23h                 ; Coordinate value 1 for Shutter low
CV2h      equ       24h                 ; Coordinate value 2 for Shutter high
CV2l      equ       25h                 ; Coordinate value 2 for Shutter low
CV3h      equ       26h                 ; Coordinate value 3 for Shutter high
CV3l      equ       27h                 ; Coordinate value 3 for Shutter low
CV4h      equ       28h                 ; Coordinate value 4 for Shutter high
CV4l      equ       29h                 ; Coordinate value 4 for Shutter low
CV5h      equ       2Ah                 ; Coordinate value 5 for Shutter high
CV5l      equ       2Bh                 ; Coordinate value 5 for Shutter low
CV6h      equ       2Ch                 ; Coordinate value 6 for Shutter high
CV6l      equ       2Dh                 ; Coordinate value 6 for Shutter low
CV7h      equ       2Eh                 ; Coordinate value 7 for Shutter high
CV7l      equ       2Fh                 ; Coordinate value 7 for Shutter low

                                        ; Bit value in Flags register
                                        ; ===========================
Ofl       equ       07h                 ; Overflow flag during Cmd_SCV
ShtState  equ       06h                 ; State of shutter
Preset    equ       05h                 ; State of shutter
                                        ; ..
AxisR     equ       03h                 ; Axis R
AxisZ     equ       02h                 ; Axis Z
AxisY     equ       01h                 ; Axis Y
AxisX     equ       00h                 ; Axis X
;
;*******************************************************************
;
;Vectors (PIC16C71)
;

          org       PIC71
_Reset    goto      _StartUp            ; Reset startup routine

                                        ; PIC 71: new defined vectors:
                                        ; ----------------------------
_WDTSLP   goto      _WdtSlp             ; # Reenter after WDT wake-up from SLEEP
_WDTTO    goto      _WdtTo              ; # Reenter after WDT time-out
_MCLRSLP  goto      _MClrSlp            ; # Reenter after /MCLR wake-up from Sleep

                                        ; (HINT: user must firstly clear flags with
                                        ; "clrwdt" and set GIE!!)

_IntV     goto      _IntHndlr           ; Interrupt handler (Reenter after INT
                                        ;   wake-up from Sleep)

_Vers     data      "MpK V0.9/LPGS V1.8/"; Mark "Version/...

          data      "jw,mk,(c)NTB,CH-Buchs/"

          if        _MHz=20             ; ...xxMHz/..
          data      "20"
          else
            if      _MHz=16
          data      "16"
            else
          data      "??"                ; #
            endif
          endif

          data      "MHz/RS232,422"

_Baud     data      "/9.6kBd"           ; # mark also baud rate

          if        _Tst=Yes
          data      "/Tst"
          else
          data      "+CRC"
          endif

;
;*******************************************************************
;
; Subroutines which must be at PCH=0
; ----------------------------------
;
SubGetCVW2 clrc
          rlf       Cmd,W               ; Hole pointer auf CV (*2)
          andlw     07h                 ; maskiere CV
          addwf     PC,F                ; Selektiere richtige CV
          movf      CV0l,W              ; Hole CV0l
          goto      IntoW2
          movf      CV1l,W              ; Hole CV1l
          goto      IntoW2
          movf      CV2l,W              ; Hole CV2l
          goto      IntoW2
          movf      CV3l,W              ; Hole CV3l
          goto      IntoW2
          movf      CV4l,W              ; Hole CV4l
          goto      IntoW2
          movf      CV5l,W              ; Hole CV5l
          goto      IntoW2
          movf      CV6l,W              ; Hole CV6l
          goto      IntoW2
          movf      CV7l,W              ; Hole CV7l
IntoW2    movwf     W2

          clrc
          rlf       Cmd,W               ; Hole pointer auf CV (*2)
          andlw     07h                 ; maskiere CV
          addwf     PC,F                ; Selektiere richtige CV
          movf      CV0h,W              ; Hole CV0h
          goto      IntoW
          movf      CV1h,W              ; Hole CV1h
          goto      IntoW
          movf      CV2h,W              ; Hole CV2h
          goto      IntoW
          movf      CV3h,W              ; Hole CV3h
          goto      IntoW
          movf      CV4h,W              ; Hole CV4h
          goto      IntoW
          movf      CV5h,W              ; Hole CV5h
          goto      IntoW
          movf      CV6h,W              ; Hole CV6h
          goto      IntoW
          movf      CV7h,W              ; Hole CV7h
IntoW     return
;
;*******************************************************************
;
; Start up after any reset or power up...
;
                                        ; Detects Reset conditions
                                        ; and reenters accordingly
                                        ; ------------------------
_StartUp  btfsc     STATUS,PD           ; if sleep detected THEN
          goto      nosleep
          btfss     STATUS,TO           ;  if WDT time-out THEN
          goto      _WDTSLP             ;   Goto WDT-Sleep reentrance
          goto      _MCLRSLP            ;  else Goto /MCLR-Sleep reentrance
                                        ;  endif
nosleep   btfss     STATUS,TO           ; ELSif WDT time-out THEN
          goto      _WDTTO              ;  Goto WDT time-out reentrance
                                        ; endif

          movlw     Tx_Rst              ; Send Tx_Rst at Reset if bit 7 is set
          movwf     TxData
          movlw     _RstC               ; Mark Reset Condition
          movwf     Cmd
          goto      _PwrOnRst

; -----------------------------------------------------------------------

_WdtSlp   movlw     Tx_WDTSLP           ; Send Tx_WDTSLP at Reset if bit 7 is set
          movwf     TxData
          movlw     _WdtSlpC            ; Mark WDT in sleep mode Condition
          movwf     Cmd
          goto      _Dummy

; -----------------------------------------------------------------------

_WdtTo    movlw     Tx_WDTTO            ; Send Tx_WDTTO at Reset if bit 7 is set
          movwf     TxData
          movlw     _WdtToC             ; Mark WDT timeout Condition
          movwf     Cmd
          goto      _Dummy

; -----------------------------------------------------------------------

_MClrSlp  movlw     Tx_MCLRS            ; Send Tx_MCLRS at Reset if bit 7 is set
          movwf     TxData
          movlw     _MClrSC             ; Mark Reset pin pulled in sleep Condition
          movwf     Cmd
          goto      _Dummy
;
;***************************************************************
;
; PowerOn- or Reset-Condition
; ---------------------------
; Put system into defined condition
; Hardware initialization:
;
_Dummy    if        _WDT=Yes            ; and unused entry for other power up cond.
          clrwdt
          endif
;
; Hardware initialization:
;
                                        ; TRISA Register: #
_PwrOnRst movlw     b'11111'            ; Out:    -
          tris      TRISA               ; In:     0-3: CHA...
                                        ;         4:   -

;         movlw     b'11111'            ; Reset Port A: CHA
;         movwf     PORTA
          movf      PORTA,W
                                        ; TRISB Register: (#)
          movlw     b'11110001'         ; Out:    3,        ShttrStrt Pin
                                        ;         2,        ShttrRst Pin
                                        ;         1:        TxPin
          tris      TRISB               ; In:     0:        RxPin
                                        ;         7,6,5,4:  CHB Pins
          movlw     b'00000111'         ; Reset Port B: CHB,ShttrStrt=0,-Rst=1,TxPin=1
          movwf     PORTB
          movf      PORTB,W

          movlw     CV0h                ; Load Indirect Pointer of Main to first buffer
          movwf     FSR
          movlw     CV0h                ; Load Indirect Pointer of System to first buffer
          movwf     FSRInt

          movlw     b'10001000'         ; option Register init: (#)
          option                        ; RBPU=dis, INTEDG=falling, RTS=Intern,
                                        ; RTE=X, PSA=WDT,(Presc=000)

          movlw     b'00000000'         ; ADCON0 Register: OFF
          movwf     ADCON0              ; ADCS=00,x,CHS=00,GO=0,ADif=0,ADON=OFF


          movlw     b'11'               ; ADCON1 Register: #
          Page1                         ; PCFG=xxxxxx11:RA0-3=Digital in
          movwf     ADCON1
          Page0
;
; Init system regs.:
;
          movlw     b'0011'             ; (SRate+)
          movwf     PStat               ; Bit 7-4:          Countdown sample rate
                                        ; Bit 3,2:          TxFlag, RxFlag,
                                        ; Bit 1,0:          Proc 11 (->Tx 00)

          movlw     b'00001000'
          movwf     TxRxFlg             ; Init Tx / Rx flags
                                        ; Bit 7-5:          Tx bit counter=000
                                        ; Bit 4:  Flag TxStop
                                        ; Bit 3:  Flag TxStart  = 1
                                        ; Bit 2:  Flag RxSel proc 01,11 select
                                        ; Bit 1:  Flag RxStop
                                        ; Bit 0:  Flag Rxtart
;
;
; RTCC and prescaler initialization
;
          clrwdt                        ; Set defined conditions to flags, WDT
                                        ; and the prescaler
          Page1
          movf      OPT,W               ; Set prescaler:
          Page0
          iorlw     Presc               ; Internal,low-high,RTCC,'constant'
          option                        ; (OPT was set to xx000000b within _InitHW)
                                        ; First time RTCC runs up from 00h to 0FFh
;
; Enable and disable interrupts:
;
          movlw     b'10110000'         ; # En-/disables Interrupts:
          movwf     INTCON              ;   GIE,/ADIE,RTIE,INTE,/RBIE,rtcc,int,rb

          goto      _Main               ; Goto Main Programm
;
;*******************************************************************
;
; Main interrupt handler which branches depending on interrupt source
; -------------------------------------------------------------------
; All interrupts are branching to this routine. It firstly saves the
; work- and the statusfile, than detects the occured interrupt and
; jumps to the according subroutine.
; The way the user queues the calls that makes the priority. Last has
; most priority, if one after the other is tested ("_RetIntX").
; If the user wants equal priority, just reenter always at "_RetInt"!
; To maintain 100% STATUSreg recovery, the zeroflag must be testet
; and re-set after reloading the workfile!
;
                                        ; Interrupt Handler (GIE=0)
                                        ; -----------------
                                        ; higher priority interrupts which occur
                                        ; during handler are executed if detected
                                        ; after RetIntX of the previous interrupt!

_IntHndlr movwf     SaveW               ; Save W and STATUS
          movf      STATUS,W
          movwf     SaveST
          movf      FSR,W               ; Exchange FSR
          movwf     FSRMain
          movf      FSRInt,W
          movwf     FSR
                                        ; Interrupt branch depending on source
                                        ; and reenter-test:
                                        ; (Most important proc last!)
                                        ; ------------------------------------
_RetInt
;         btfsc     ADCON0,ADif         ; if AD completion THEN
;         goto      _IntAD              ;  goto AD-Handler

_RetInt1
          btfsc     INTCON,RBIF         ; elseif Port B change THEN
          goto      _IntPORTRB          ;  goto RB-Handler

_RetInt2  btfsc     INTCON,INTF         ; elseif INT THEN
          goto      _INT                ;  goto INT-Handler

_RetInt3  btfsc     INTCON,RTIF         ; if RTCC overflow THEN
          goto      _IntRTCC            ;  goto RTCC-Handler
                                        ; endif

_RetInt4  movf      FSR,W               ; Exchange FSR
          movwf     FSRInt
          movf      FSRMain,W
          movwf     FSR
          movf      SaveST,W            ; Reload old STATUS and
          movwf     STATUS
          movf      SaveW, W            ; Z-Flag changed!!
          btfsc     SaveST,Z            ; Test old Z
          setz                          ; and re-set it...
          btfss     SaveST,Z
          clrz
EOI       retfie                        ; Enable global interrupt and
                                        ; return to processes
;
;*******************************************************************
;
; Interrupt on completion of A/D-conversion (wake-up)
; ---------------------------------------------------
;
_IntAD    bcf       ADCON0,ADIF         ; Reset flag
          btfss     INTCON,ADIE         ; if AD is disabled THEN
          goto      _RetInt1            ;   return
                                        ; endifF

                                        ; #...

          goto      _RetInt1            ; End of AD interrupt
;
;*******************************************************************
;
; Interrupt on changes on port RB4-7 (wake-up)
; --------------------------------------------
;
_IntPORTRB movf     PORTB,W             ; Read PORTB
          bcf       INTCON,RBIF         ; Reset flag
          btfss     INTCON,RBIE         ; if RB is disabled THEN
          goto      EndIPRB             ;   return
                                        ; endif

                                        ; Zhler und Vergleichfunktion
                                        ; um den Shutter zu Steuern
                                        ; -------------------------
TestAxis  xorwf     OldPRB,F            ; get chanched channel

TestX     btfss     Flags,AxisX         ; Is Axis X selected ?
          goto      TestY
          btfss     OldPRB,CHBX
          goto      EndIPRB             ; Nothing selected!
          btfss     PORTA,CHAX          ; wenn Pin CHAX von PORTA = 0 dann
          goto      TestFallX           ;         gehe zu Label TestFall
          btfsc     PORTB,CHBX          ; wenn Pin CHBX von PORTB = 1 dann
          goto      IncCnt              ;         gehe zu Label IncCnt
          goto      DecCnt              ; sonst gehe zu Label DecCnt
TestFallX btfss     PORTB,CHBX          ; wenn Pin CHBX von PORTB = 0 dann
          goto      IncCnt              ;         gehe zu Label IncCnt
          goto      DecCnt              ; sonst gehe zu Label DecCnt

TestY     btfss     Flags,AxisY         ; Is Axis Y selected ?
          goto      TestZ
          btfss     OldPRB,CHBY
          goto      EndIPRB             ; Nothing selected!
          btfss     PORTA,CHAY          ; wenn Pin CHAY von PORTA = 0 dann
          goto      TestFallY           ;         gehe zu Label TestFall
          btfsc     PORTB,CHBY          ; wenn Pin CHBY von PORTB = 1 dann
          goto      IncCnt              ;         gehe zu Label IncCnt
          goto      DecCnt              ; sonst gehe zu Label DecCnt
TestFallY btfss     PORTB,CHBY          ; wenn Pin CHBY von PORTB = 0 dann
          goto      IncCnt              ;         gehe zu Label IncCnt
          goto      DecCnt              ; sonst gehe zu Label DecCnt

TestZ     btfss     Flags,AxisZ         ; Is Axis Z selected ?
          goto      TestR
          btfss     OldPRB,CHBZ
          goto      EndIPRB             ; Nothing selected!
          btfss     PORTA,CHAZ          ; wenn Pin CHAZ von PORTA = 0 dann
          goto      TestFallZ           ;         gehe zu Label TestFall
          btfsc     PORTB,CHBZ          ; wenn Pin CHBZ von PORTB = 1 dann
          goto      IncCnt              ;         gehe zu Label IncCnt
          goto      DecCnt              ; sonst gehe zu Label DecCnt
TestFallZ btfss     PORTB,CHBZ          ; wenn Pin CHBZ von PORTB = 0 dann
          goto      IncCnt              ;         gehe zu Label IncCnt
          goto      DecCnt              ; sonst gehe zu Label DecCnt

TestR     btfss     Flags,AxisR         ; Is Axis R selected ?
          goto      EndIPRB
          btfss     OldPRB,CHBR
          goto      EndIPRB             ; Nothing selected!
          btfss     PORTA,CHAR          ; wenn Pin CHAR von PORTA = 0 dann
          goto      TestFallR           ;         gehe zu Label TestFall
          btfsc     PORTB,CHBR          ; wenn Pin CHBR von PORTB = 1 dann
          goto      IncCnt              ;         gehe zu Label IncCnt
          goto      DecCnt              ; sonst gehe zu Label DecCnt
TestFallR btfss     PORTB,CHBR          ; wenn Pin CHBR von PORTB = 0 dann
          goto      IncCnt              ;         gehe zu Label IncCnt
;         goto      DecCnt              ; sonst gehe zu Label DecCnt

DecCnt    movf      PORTB,W             ; Save old sate
          movwf     OldPRB
          movlw     01h                 ; schreibe 01h ins W Register
          subwf     Cnt,W               ; W = Cnt - W
          skpc                          ; wenn Cnt = FFh dann
          movlw     09h                 ;         schreibe 09h ins W Register
          andlw     0Fh                 ; Allflliger Fehler korigieren ->Z
          movwf     Cnt                 ; schreibe den Wert vom W Register in Cnt
          skpz                          ; wenn Cnt <> 0 dann
          goto      _RetInt2            ;   sonst gehe verlasse die Interrupt Routine
          movlw     01h                 ; schreibe 01h ins W Register
          subwf     CntL,f              ; CntL = CntL - 1
          skpc                          ; wenn Cnt = FFh dann
          decf      CntH,f              ;   CntH = CntH - 1
          goto      TestKH              ; gehe zu Label TestKH

IncCnt    movf      PORTB,W             ; Save old sate
          movwf     OldPRB
          incf      Cnt,W               ; Cnt + 1
          andlw     0Fh                 ; Allflliger Fehler korigieren
          movwf     Cnt                 ; schreibe den Wert zurck ins Cnt
          sublw     0Ah                 ; W = 0Ah - W bzw. (W = 10 - W)
          skpz                          ; wenn W <> 0 dann
          goto      _RetInt2            ;   gehe zu Label TestCnt
          clrf      Cnt                 ; setze Cnt auf 0 zurck
          incf      CntL,f              ; CntL = CntL + 1
          skpnz                         ; wenn CntL > max
          incf      CntH,f              ;   CntH = CntH + 1

                                        ; Vergleiche CntHL und buffer
                                        ; ---------------------------
TestKH    movf      INDADR,W            ; schreibe die aktuelle Koordinate in W
          subwf     CntH,W              ; W = CntH - W
          skpz                          ; wenn W <> 0 dann
          goto      _RetInt2            ;   verlasse die Interrupt Routine
          incf      FSR,f               ; Schiebe den Zeiger des Buffers um 1 weiter
          movf      INDADR,W            ; schreibe die aktuelle Koordinate in W
          subwf     CntL,W              ; W = CntL - W
          skpnz                         ; wenn W = 0 dann
          goto      Shttr               ;   gehe zu Label Shutter
          decf      FSR,f               ; sonst setze den Zeiger wieder zurck
          goto      _RetInt2            ; End of Port RB interrupt

Shttr     call      SubShttrInv         ; Kehre Shutter um
          decf      FSR,f               ; setze den Zeiger um eins zurck
          clrf      INDADR              ; lsche das high Byte der Koordinate
          incf      FSR,f               ; setze de Zeiger wieder um 1 nach vorne
          clrf      INDADR              ; lsch das low Byte der Koordinate
          incf      FSR,f               ; Schiebe den Zeiger des Buffers um 1 weiter
          btfss     FSR,4               ; Und ist Bit 4 gesetzt?
          goto      _RetInt2            ; End of Port RB interrupt

          movlw     CV0h                ; schreibe CV0h ins W Register
          movwf     FSR                 ; schreibe den Wert des W Registers in FSR
          goto      _RetInt2            ; End of Port RB interrupt

EndIPRB   movf      PORTB,W             ; Save old sate
          movwf     OldPRB
          goto      _RetInt2            ; End of Port RB interrupt
;
;*******************************************************************
;
; Extern INTerrupt on RB0 pin input (wake-up)
; -------------------------------------------
; Is used to determine RxStart. A falling edge activates the interrupt.
; The sub disables itself during next 8 bits until data is received.
; The RxProcess is set free by TxFlag and TxStart. To catch the input
; level at the right time next RxProcess time slice is counted and set
; to flag RxSel.
;
_INT      bcf       INTCON,INTF         ; Reset flag
          btfss     INTCON,INTE         ; if INT is disabled THEN
          goto      _RetInt3            ;   return
                                        ; endif

          bcf       INTCON,INTE         ; Disable next interrupts until _RxStop

          bsf       PStat,RxFlag        ; Set RxFlag
          bsf       TxRxFlg,RxStart     ; Set pointer to start bit

          incf      PStat,F             ; Because _INT is before _IntRTCC
          btfsc     PStat,PCnt1         ; If Proc00/01 is active then set RxSel=0
          bsf       TxRxFlg,RxSel
          btfss     PStat,PCnt1         ; else set RxSel=1
          bcf       TxRxFlg,RxSel
          decf      PStat,F

          goto      _RetInt3            ; End of INT interrupt
;
;*******************************************************************
;
; Interrupt on Real Time Clock overflow:
; --------------------------------------
; Every (#) Xus the RTC has its overflow to detemine the right time intervall
; for Tx- and RxProcesses. The user must consider of having enough spare
; cycle to avoid the RTC overflowing twice in case of several interrupt
; sources occur!
; Before jumping to the 4 intervall processes a separate AD-Start sub is
; testet for its call. This allows to start the AD-conversion at every
; 1*Xus up to 16*Xus (sample rate)!
; If a uncommon X-tal is used, an adjust process may be provided to inc or dec
; some RTC-steps to adjust the roundabout time of the 4 processes!
; Otherwise this process may be used by the user. #
;
_IntRTCC  bcf       INTCON,RTIF         ; Reset flag
;         btfss     INTCON,RTIE         ; if RTCC is disabled THEN
;         goto      _RetInt4            ;   return
                                        ; endif

          movlw     100h-Rtcc           ; Set RTCC start value from constant (Xus)
          addwf     RTCC,f              ; #

          skpc                          ; Test if there was a cycle lost?
          goto      NoLostC
                                        ; LostCycle
                                        ; ---------
          movlw     0ffh
          movwf     RTCC                ; Set it near before INT

NoLostC   movlw     11h                 ; Countdown of sample rate and
          addwf     PStat,F             ; set process counter + 1 (overflow bit 2?)
;         skpnc
;         call      _ADProc             ; Proc AD: Start conversion

                                        ; Check processes 00 to 11
                                        ; ------------------------
          btfsc     PStat,PCnt0         ; Test
          goto      _RxProc             ; Proc 01 or 11: Rx depending on RxSel
          btfsc     PStat,PCnt1
          goto      _RetInt4            ; Not used:
;         goto      _AdjProc            ; Proc 10: # (Adjust)
          movlw     4h                  ; fix overflow bit 2
          subwf     PStat,F             ;
          goto      _TxProc             ; Proc 00: Tx
;
;*******************************************************************
;
; AD process:
; -----------
; Starts AD-conversion and resets sample rate countdown
;
;_ADProc  bsf       ADCON0,GO           ; Start conversion
;         movlw     SRate               ; Re-set countdown value
;         addwf     PStat,F
;         return
;
;*******************************************************************
;
; Tx process 00:If TxFlag is set data TxData will be sent via pin RB1.
;                   According to TxRxFlgs:
;                   TxStart:start(low)
;                   TxData: (Count at TxRxFlg)
;                   1xH:      bit0
;                   3xH:      bit1
;                   5xH:      bit2
;                   7xH:      bit3
;                   9xH:      bit4
;                   BxH:      bit5
;                   DxH:      bit6
;                   FxH:      bit7 -> Carry sets TxStop
;                   TxStop:   stop(high),reset TxFlag to mark end of Tx.
;                                          and sets TxStart for next transmission
; To set a new TxData check if TxFlag=0!
;
                                        ; Transmitter every 4*Xus
_TxProc   btfsc     PStat,TxFlag        ; Is transmitter active?
          goto      _TxOK               ;   yes: send next data
          goto      _RetInt4            ;        return

_TxOK     btfsc     TxRxFlg,TxStart     ; Is start bit to be sent?
          goto      _TxStart            ; Yes
          btfss     TxRxFlg,TxStop      ; Is data to be sent?
          goto      _TxData             ; Yes

_TxStop   bsf       PORTB,TxPin         ; Set output high (STOP)
          bcf       TxRxFlg,TxStop      ; Clear TxStop
          bcf       PStat,TxFlag        ; Clear TxFlag
          bsf       TxRxFlg,TxStart     ; Set TxStart for next transfer
          goto      _RetInt4
                                        ; Send START
_TxStart  bcf       PORTB,TxPin         ; Set output low (START)
          bcf       TxRxFlg,TxStart     ; Start is sent
          goto      _RetInt4
                                        ; Send DATA
_TxData   if        _Tst=Yes            ; ---------
          clrc                          ; Set rotatebits to zero
          endif
          rrf       TxData,F            ; bit to carry
          skpnc
          bsf       PORTB,TxPin         ; Set output according to carry
          skpc
          bcf       PORTB,TxPin         ;
          movlw     20h                 ; Increment counter...
          addwf     TxRxFlg,F           ; count up to bit 8...
          skpnc                         ; is last bit sent?
          bsf       TxRxFlg,TxStop      ; Yes->Set Stop flag
          goto      _RetInt4
;
;*******************************************************************
;
; Rx process 01,11: If RxFlag is ser and the correct process according
; to RxSel is running, data RxData will be sent via pin RB1:
;                   According to TxRxFlags:
;                   RxStart:start(low), set first bit of RxData high
;                   RxData: bit0..7, shift RxData till first bit is high
;                                        ->set RxStop
;                   RxStop:   stop(high),reset RxFlag, enable INTerrupt,
; # User must clear flag DataRdy after datafetch at file ADCON0
; to mark data is read!
;
                                        ; Receiver process every (other) 4*Xus
_RxProc   btfss     PStat,RxFlag        ; Is receiver active?
          goto      _RetInt4            ; no:return
          btfsc     PStat,PCnt1         ; Is it the selected timeslice?
          goto      _Rx11               ;
_Rx01     btfss     TxRxFlg,RxSel       ; PCnt=0:Sel=?
          goto      _RxOK               ; Sel=0   =>ok, go ahead
          goto      _RetInt4            ; Sel=1 =>return
_Rx11     btfss     TxRxFlg,RxSel       ; PCnt1=1:Sel=?
          goto      _RetInt4            ; Sel=0 =>return

_RxOK     btfsc     TxRxFlg,RxStart     ; Is start bit first
          goto      _RxStart            ; Yes
          btfss     TxRxFlg,RxStop      ; Is stop bit next?
          goto      _RxData             ; No: ->Data

_RxStop   btfss     PORTB,RxPin         ; Is signal high?
          goto      _RxError            ; Low: error
                                        ; High:ok...
_RxStopE  bsf       INTCON,INTE         ; Enable INT interrupt for new data
          bsf       ADCON0,DataRdy      ; Set data ready flag
          bcf       PStat,RxFlag        ; Clear RxFlag
          bcf       TxRxFlg,RxStop      ; Clear RxStop
          goto      _RetInt4
                                        ; An error occured:
_RxError  bsf       SaveST,RxError      ; Mark Rx Error (Reset in Main after sent)
          bsf       Cmd,Cmd_NACK        ; Mark into Cmd-Word NACK
          goto      _RxStopE            ; # Re-enable RxProcess after sending NACK!!

_RxStart  bcf       TxRxFlg,RxStart     ; Start bit passed
                                        ; ===============================
                                        ; # Execute "bcf ADCON0,DataRdy"
                                        ;   after data is read and before
                                        ;   new databits are comming in!!
                                        ; ===============================
          btfsc     ADCON0,DataRdy      ; Test for old data is read?
          goto      _RxBusy             ; Too slow.... >>> error!!!
          btfsc     PORTB,RxPin         ; Is signal low?
          goto      _RxError            ; High: error
          movlw     80H
          movwf     RxData              ; set rotate flag
          goto      _RetInt4
                                        ; An error occured:
_RxBusy   bsf       Cmd,Cmd_NACK        ; Mark into Cmd-Word NACK
          bsf       SaveST,RxBusy       ; Mark Rx Busy (Reset in Main after sent)
          goto      _RxStopE            ; # Re-enable RxProcess after sending NACK!!


_RxData   btfsc     PORTB,RxPin         ; Receive data
          setc
          btfss     PORTB,RxPin
          clrc
          rrf       RxData,F            ; Carry to file
          skpnc
          bsf       TxRxFlg,RxStop      ; Last bit is received -> get stop bit now
          goto      _RetInt4
;
;*******************************************************************
;
; Adjust process 10:
; ------------------
; This process may be used for adjustment of the roundabout time of
; the 4 Process (and the sample rate). By inc- or decrementing the RTCC
; any adjustment may be done!
; # If not used this is a free user definable process:
;
;_AdjProc movlw     Adj                 ; Adjust RTCC to the av. of Xus/process
;         addwf     RTCC,F
                                        ; OR #
;         goto      _RetInt4
;
;*******************************************************************
;
; Main Processes
; --------------
; MpK Rem.:         7 stack levels are free. Shared vars may not be used without
;                   disabling GIE (NoInt,DoInt)!
; Zwo pseude processes are conncted and runned by a roundabout angle
; point: SUB SubInputExe. during the wait of cmd-process for cmd input,
; cmd flags and cmd output are executed via output sub processes.
;
_Main     if        _Tst=Yes
                                        ; Send Reset Info if WDT is used
          bsf       PStat,TxFlag
          endif
                                        ; Clear used Variables
                                        ; ====================
;         movlw     _RstC               ; Was other than normal reset?
;         subwf     Cmd
;         skpz                          ; Yes:
;         goto      EndRstCnd           ;   skip settings

          clrf      ExtCmd              ; Clear regs:...
          clrf      CRCr
          clrf      CRCs
          clrf      OldPRB
          clrf      Cnt
          clrf      CntH
          clrf      CntL
                                        ; Reset CV-Buffer
                                        ; ===============
          NoInt                         ; Enter shared range
          movlw     CV0h                ; Reset= 00h
          movwf     FSR                 ; Schleifen resetten
          movwf     FSRInt              ;    "
          DoInt                         ; Leave shared range

          clrf      Flags               ; No Axis,no Ofl

          clrf      CV0h                ; Coordinate value 0 for Shutter high Set Gain Channel 0
          clrf      CV0l                ; Coordinate value 0 for Shutter low  Set Gain Channel 1
          clrf      CV1h                ; Coordinate value 1 for Shutter high Set Gain Channel 2
          clrf      CV1l                ; Coordinate value 1 for Shutter low  Set Gain Channel 3
          clrf      CV2h                ; Coordinate value 2 for Shutter high Set Gain Channel 4
          clrf      CV2l                ; Coordinate value 2 for Shutter low  Set Gain Channel 5
          clrf      CV3h                ; Coordinate value 3 for Shutter high Set Gain Channel 6
          clrf      CV3l                ; Coordinate value 3 for Shutter low  Set Gain Channel 7
          clrf      CV4h                ; Coordinate value 4 for Shutter high Set Offset of Channel 0
          clrf      CV4l                ; Coordinate value 4 for Shutter low  Set Offset of Channel 1
          clrf      CV5h                ; Coordinate value 5 for Shutter high Set Offset of Channel 2
          clrf      CV5l                ; Coordinate value 5 for Shutter low  Set Offset of Channel 3
          clrf      CV6h                ; Coordinate value 6 for Shutter high Set Offset of Channel 4
          clrf      CV6l                ; Coordinate value 6 for Shutter low  Set Offset of Channel 5
          clrf      CV7h                ; Coordinate value 7 for Shutter high Set Offset of Channel 6
          clrf      CV7l                ; Coordinate value 7 for Shutter low  Set Offset of Channel 7

EndRstCnd clrf      Cmd                 ; Rem.: TxData, RxData are undef...
;
;--------------------------------------------------------------------
;
; Cmd-Process
; -----------
; Receives Cmd via Rx-Line, checks Header (0FFH), Address
; and saves Cmd. If there is an direct Cmd-action to do at a specific
; cmd, it will be performed immediatly.
;
GetHdr                                  ; Wait for Header
RetNACK   call      SubInpExe           ; Wait for Cmd_Hdr

          movlw     Rx_Hdr
          subwf     RxData,W            ; if Rx_Hdr->FFh-FFh=00h(C) ....

          bcf       ADCON0,DataRdy      ; ( Free RxProcess

          btfsc     Cmd,Cmd_NACK        ; Is there a NACK?
          goto      BrnchNACK           ; Send NACK and Restart CmdExecution

HdrOK     skpc                          ; THEN get next data ...
          goto      GetHdr              ; else...wait for new header (sync)

                                        ; Wait for Cmd
                                        ; ============
GetCmd    movlw     00h                 ; Clear Cmd for new Command. This stops ExeSubs!
          movwf     Cmd
          bcf       STATUS,RxError      ; Reset Rx Errors..
          bcf       STATUS,RxBusy

          call      SubInpExe           ; Wait for Cmd
          movfw     RxData              ; Get RxData
          bcf       ADCON0,DataRdy      ; Free RxProcess
          btfsc     Cmd,Cmd_NACK        ; Is there a NACK?
          goto      BrnchNACK           ; Restart CmdExecution

                                        ; (Default is '001x xxxx')
          movwf     CRCr                ; Set first CRC (V0.8)
          movwf     Cmd                 ; Save Cmd
          btfsc     Cmd,07h             ; Test for Syntax Error
          goto      SyntaxErr
          btfsc     Cmd,06h             ; Test for Syntax Error
          goto      SyntaxErr
          btfss     Cmd,05h             ; Test for Syntax Error
          goto      SyntaxErr
                                        ; No SyntaxError:
          movlw     b'00011111'         ; Clear first 3 bits anyway...
          andwf     Cmd,F               ; ... of Cmd
          goto      TstActCmd           ; Test for action command now!

RetTAC    btfsc     Cmd,Cmd_NACK        ; Test for Syntax Error
          goto      BrnchNACK
          goto      GetHdr              ; Endless loop
;
;--------------------------------------------------------------------
;
; Syntax Error occured:
; ---------------------
; Only the NACK-Flag is set!
;
SyntaxErr bsf       Cmd,Cmd_NACK        ; Imaginary set NACK for Syntax error
          bcf       STATUS,RxError      ; Reset Rx Errors..
          bcf       STATUS,RxBusy
;         goto      BrnchNACK           ; (no other flag is set)
;
;--------------------------------------------------------------------
;
; Error branch if Cmd_NACK is set:
;
; If Cmd_NACK is set this routine should be accessed.
; Tree errors may be sent back to host:
;         if RxError is set, start/stop bit failure occured
;         if RxBusy is set, Tx did send to fast
;         else a Syntax error were found
; The according Error-Data will be sent back to host and
; Cmd-Process restarts at GetHdr...
;
BrnchNACK bcf       Cmd,Cmd_NACK        ; Clear NACK Flag

TestErr   btfsc     STATUS,RxError      ; Did RxError occur ?
          goto      TxRxError           ;  Yes
          btfsc     STATUS,RxBusy       ; Did Rxbusy occur ?
          goto      TxRxBusy            ;  Yes
          btfsc     Flags,Ofl           ; Did Overflow occur ?
          goto      TxOflErr            ;  Yes

          goto      TxSxErr             ; Else Syntax Error... !

TxOflErr  movlw     Tx_OflErr           ; Mark for Ofl
          bcf       Flags,Ofl           ; Reset Flag
          goto      RxEn                ; Free Rx Process again

                                        ; MPK:
TxRxError movlw     Tx_RxErr            ; Mark for RxError
          bcf       STATUS,RxError      ; Reset Flag
          goto      RxEn                ; Free Rx Process again

TxRxBusy  movlw     Tx_RxBusy           ; Mark for RxBusy
          bcf       STATUS,RxBusy       ; Reset Flag

RxEn      DoInt                         ; Enable INT interrupt for new data
          goto      SendErr

TxSxErr   movlw     Tx_SErr             ; Mark for syntax error

SendErr   call      SubSendW
          movlw     00h                 ; Resets Exe-, ACK-(, NACK-) and Cmd-Flags
          movwf     Cmd
          goto      RetNACK             ; Restarts Cmd receiving process!
;
;****************************************************************************
;
; Action Command routines
; -----------------------
; If a Cmd as an action part(*), it will no be performed once. After that
; the Cmd-Process waits for new Cmd.
;
; (*) The action part may include additional Data fetching, action execution
; and/or first CmdSub call.
;
TstActCmd btfsc     Cmd,Cmd_RCV         ; Return Coordinate Value ?
          goto      RCV_Cmd             ;  Yes

          btfsc     Cmd,Cmd_SCV         ; Set Coordinate value ?
          goto      SCV_Cmd             ;  Yes

          btfsc     Cmd,Cmd_SSV         ; Set Start Value/Axis ?
          goto      SSV_Cmd             ;  Yes

          btfsc     Cmd,Cmd_RST         ; Reset Shutter ?
          goto      RST_Cmd             ;  Yes

          goto      EXT_Cmd             ; Spezial routine

SErr      clrf      Cmd                 ; Error Return Address: Clear cmd and set error:
          bsf       Cmd,Cmd_NACK        ; Imaginary set NACK for Syntax error
          bcf       STATUS,RxError      ; Reset Rx Errors..
          bcf       STATUS,RxBusy
          goto      RetTAC              ; (no other flag is set)
;
;--------------------------------------------------------------------
;
; Get data and execute direct action commands
;
;
                                        ; Cmd direct action
                                        ; Return Coordinate Value
                                        ; =======================
RCV_Cmd   call      SubGetCRCr          ; Get and Test CRC
          skpc                          ; C set if CRCr not equal
          goto      SetACKEXE           ;  Set Cmd_Exe,Cmd_ACK flags and goto GetHdr

          goto      SErr                ; Set NACK: Syntax/CRCError

;
;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; Sets Coordinate value into next bufferspace
;
;
                                        ; Set Coordinate Value
                                        ; ====================
SCV_Cmd   call      SubInpExe           ; liest das nchste Byte ein
          movf      RxData,W            ; schreibe RxData ins W Register (high Byte)
          bcf       ADCON0,DataRdy      ; Free RxProcess
          btfsc     Cmd,Cmd_NACK        ; Is there a NACK?
          goto      RetTAC              ; Restart CmdExecution

          addwf     CRCr,F              ; add CRC
          movwf     W1                  ; Test, ob FSR frei ist ?
          movf      INDADR,W            ; TestHigh
          skpz
          goto      Full                ;

          movf      W1,W                ; Hole Wert
          movwf     INDADR              ; schreibe das W Register in den Ringbuffer
          incf      FSR,f               ; Schiebe den Zeiger des Buffers um 1 weiter

          call      SubInpExe           ; liest das nchste Byte ein
          movf      RxData,W            ; schreibe RxData ins W Register (low Byte)
          bcf       ADCON0,DataRdy      ; Free RxProcess
          btfsc     Cmd,Cmd_NACK        ; Is there a NACK?
          goto      RetTAC              ; Restart CmdExecution

          addwf     CRCr,F              ; add CRC
          movwf     W1                  ; Test, ob FSR frei ist ?
          movf      INDADR,W            ; TestHigh
          skpz
          goto      DecFull             ; Lschen letzter Daten

          movf      W1,W                ; Hole Wert
          movwf     INDADR              ; schreibe das W Register in den Ringbuffer

                                        ; CRC berprfung:
          call      SubGetCRCr          ; Get and Test CRC
          skpc                          ; C set if CRCr not equal
          goto      SCVCRCok            ;  OK

          clrf      INDADR              ; Lscht low Data
          decf      FSR,F               ; Pointer zurcksetzen
          clrf      INDADR              ; Lscht high Data
          goto      SErr                ; Set SErr/CRC

SCVCRCok  incf      FSR,F               ; Schiebe den Zeiger des Buffers um 1 weiter
          btfss     FSR,4               ; Is also Bit 4 set?
          goto      SetACKEXE           ;  Set Cmd_Exe,Cmd_ACK flags and goto GetHdr

          movlw     CV0h                ; sonst schreibe CV0h ins W Register
          movwf     FSR                 ; schreibe den Wert des W Registers in FSR
          goto      SetACKEXE           ;  Set Cmd_Exe,Cmd_ACK flags and goto GetHdr


DecFull   decf      FSR,F               ; Pointer zurck
          clrf      INDADR              ; Lscht letzte Daten
Full      bsf       Flags,Ofl           ; Set Overflow
;
;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; Sets start value into CntHL and enables _IntPORTRB
;
;
                                        ; Set Start Value
                                        ; ===============
SSV_Cmd   bcf       INTCON,RBIE         ; disable _IntPORTRB
          call      SubInpExe           ; liest das nchste Byte ein
          movf      RxData,W            ; schreibe RxData ins W Register (Achse)
          bcf       ADCON0,DataRdy      ; Free RxProcess
          btfsc     Cmd,Cmd_NACK        ; Is there a NACK?
          goto      RetTAC              ; Restart CmdExecution

          addwf     CRCr,F              ; add CRC
          andlw     b'00001111'         ; Maskiere Achsenwahl
          movwf     Flags               ; schreibe den Wert des W Registes ins Flag

          call      SubInpExe           ; liest das nchste Byte ein
          movf      RxData,W            ; schreibe RxData ins W Register (high Byte)
          bcf       ADCON0,DataRdy      ; Free RxProcess
          btfsc     Cmd,Cmd_NACK        ; Is there a NACK?
          goto      RetTAC              ; Restart CmdExecution

          addwf     CRCr,F              ; add CRC
          movwf     CntH                ; schreibe den Wert des W Registes in CntH

          call      SubInpExe           ; liest das nchste Byte ein
          movf      RxData,W            ; schreibe RxData ins W Register (low Byte)
          bcf       ADCON0,DataRdy      ; Free RxProcess
          btfsc     Cmd,Cmd_NACK        ; Is there a NACK?
          goto      RetTAC              ; Restart CmdExecution

          addwf     CRCr,F              ; add CRC
          movwf     CntL                ; schreibe den Wert des W Registes in CntL

          clrf      Cnt                 ; setzt den Inneren Zhler Cnt auf 0

                                        ; CRC berprfung:
          call      SubGetCRCr          ; Get and Test CRC
          skpc                          ; C set if CRCr not equal
          goto      SSVCRCok            ;  OK
                                        ;  Sonst : SErr/CRC setzen
          clrf      CntH                ; Lsche Cnts
          clrf      CntL
          goto      SErr

SSVCRCok  btfss     Cmd,SSV_Sht         ; Shutter preset:
          call      SubShttrRst         ; Aus
          btfsc     Cmd,SSV_Sht
          call      SubShttrStrt        ; Ein

          movf      PORTB,W             ; Save Port B state
          movwf     OldPRB              ;
          bsf       INTCON,RBIE         ; Enable _IntPORTRB

          goto      SetACKEXE           ; Set Cmd_Exe,Cmd_ACK flags and goto GetHdr
;
;--------------------------------------------------------------------
;
; Free Shutter and disables _IntPORTRB
;
;
                                        ; Reset
                                        ; =====
RST_Cmd   call      SubGetCRCr          ; Get and Test CRC
          skpnc                         ; C set if CRCr not equal
          goto      SErr

          bcf       INTCON,RBIE         ; Disable _IntPORTRB
          call      SubShttrRst         ; Reset Shutter
          clrf      Flags
          clrf      CV0h                ; Coordinate value 0 for Shutter high Set Gain Channel 0
          clrf      CV0l                ; Coordinate value 0 for Shutter low  Set Gain Channel 1
          clrf      CV1h                ; Coordinate value 1 for Shutter high Set Gain Channel 2
          clrf      CV1l                ; Coordinate value 1 for Shutter low  Set Gain Channel 3
          clrf      CV2h                ; Coordinate value 2 for Shutter high Set Gain Channel 4
          clrf      CV2l                ; Coordinate value 2 for Shutter low  Set Gain Channel 5
          clrf      CV3h                ; Coordinate value 3 for Shutter high Set Gain Channel 6
          clrf      CV3l                ; Coordinate value 3 for Shutter low  Set Gain Channel 7
          clrf      CV4h                ; Coordinate value 4 for Shutter high Set Offset of Channel 0
          clrf      CV4l                ; Coordinate value 4 for Shutter low  Set Offset of Channel 1
          clrf      CV5h                ; Coordinate value 5 for Shutter high Set Offset of Channel 2
          clrf      CV5l                ; Coordinate value 5 for Shutter low  Set Offset of Channel 3
          clrf      CV6h                ; Coordinate value 6 for Shutter high Set Offset of Channel 4
          clrf      CV6l                ; Coordinate value 6 for Shutter low  Set Offset of Channel 5
          clrf      CV7h                ; Coordinate value 7 for Shutter high Set Offset of Channel 6
          clrf      CV7l                ; Coordinate value 7 for Shutter low  Set Offset of Channel 7

          movlw     CV0h                ; Set both FSRs to CV0h
          movwf     FSR
          movwf     FSRInt

          clrf      CntH                ; Clearance of counters
          clrf      CntL
          clrf      Cnt

          goto      SetACKEXE           ; Set Cmd_Exe,Cmd_ACK flags and goto GetHdr
;
;---------------------------------------------------------------------
; Extended commands
; =================
;
EXT_Cmd   call      SubInpExe           ; Wait for next data
          movf      RxData,W            ; Get Data into W
          bcf       ADCON0,DataRdy      ; Free RxProcess
          btfsc     Cmd,Cmd_NACK        ; Is there a NACK?
          goto      RetTAC              ; Restart CmdExecution

          addwf     CRCr,F              ; add CRC
          movwf     ExtCmd              ; Save Ext into ExtCmd

                                        ; Test Ext action
TstActExt                               ; ---------------
          btfsc     ExtCmd,Ext_SCN      ; Scan ?
          goto      SCN_Ext             ;  Yes

          btfsc     ExtCmd,Ext_SGO      ; Shutter Go ?
          goto      SGO_Ext             ;  Yes

SEErr     clrf      ExtCmd
          goto      SErr                ; SyntaxError when Ext = 00h.
;
;--------------------------------------------------------------------
;
; Extended Command actions
;
SCN_Ext   bcf       INTCON,RBIE         ; Disable _IntPORTRB
          call      SubShttrRst         ; Reset Shutter
          clrf      Flags
          clrf      CV0h                ; Coordinate value 0 for Shutter high Set Gain Channel 0
          clrf      CV0l                ; Coordinate value 0 for Shutter low  Set Gain Channel 1
          clrf      CV1h                ; Coordinate value 1 for Shutter high Set Gain Channel 2
          clrf      CV1l                ; Coordinate value 1 for Shutter low  Set Gain Channel 3
          clrf      CV2h                ; Coordinate value 2 for Shutter high Set Gain Channel 4
          clrf      CV2l                ; Coordinate value 2 for Shutter low  Set Gain Channel 5
          clrf      CV3h                ; Coordinate value 3 for Shutter high Set Gain Channel 6
          clrf      CV3l                ; Coordinate value 3 for Shutter low  Set Gain Channel 7
          clrf      CV4h                ; Coordinate value 4 for Shutter high Set Offset of Channel 0
          clrf      CV4l                ; Coordinate value 4 for Shutter low  Set Offset of Channel 1
          clrf      CV5h                ; Coordinate value 5 for Shutter high Set Offset of Channel 2
          clrf      CV5l                ; Coordinate value 5 for Shutter low  Set Offset of Channel 3
          clrf      CV6h                ; Coordinate value 6 for Shutter high Set Offset of Channel 4
          clrf      CV6l                ; Coordinate value 6 for Shutter low  Set Offset of Channel 5
          clrf      CV7h                ; Coordinate value 7 for Shutter high Set Offset of Channel 6
          clrf      CV7l                ; Coordinate value 7 for Shutter low  Set Offset of Channel 7

          movlw     CV0h                ; Set both FSRs to CV0h
          movwf     FSR
          movwf     FSRInt

GetAxis   call      SubInpExe           ; liest das nchste Byte ein
          movf      RxData,W            ; schreibe RxData ins W Register (Achse)
          bcf       ADCON0,DataRdy      ; Free RxProcess
          btfsc     Cmd,Cmd_NACK        ; Is there a NACK?
          goto      SCNErr              ; Restart CmdExecution

          addwf     CRCr,F              ; add CRC
          andlw     b'00001111'         ; Maskiere Achsenwahl
          movwf     Flags               ; schreibe den Wert des W Registes ins Flag

GetCntH   call      SubInpExe           ; liest das nchste Byte ein
          movf      RxData,W            ; schreibe RxData ins W Register (high Byte)
          bcf       ADCON0,DataRdy      ; Free RxProcess
          btfsc     Cmd,Cmd_NACK        ; Is there a NACK?
          goto      SCNErr              ; Restart CmdExecution

          addwf     CRCr,F              ; add CRC
          movwf     CntH                ; schreibe den Wert des W Registes in CntH

GetCntL   call      SubInpExe           ; liest das nchste Byte ein
          movf      RxData,W            ; schreibe RxData ins W Register (low Byte)
          bcf       ADCON0,DataRdy      ; Free RxProcess
          btfsc     Cmd,Cmd_NACK        ; Is there a NACK?
          goto      SCNErr              ; Restart CmdExecution

          addwf     CRCr,F              ; add CRC
          movwf     CntL                ; schreibe den Wert des W Registes in CntL
          clrf      Cnt                 ; setzt den Inneren Zhler Cnt auf 0

GetCV0H   call      SubInpExe           ; liest das nchste Byte ein
          movf      RxData,W            ; schreibe RxData ins W Register (high Byte)
          bcf       ADCON0,DataRdy      ; Free RxProcess
          btfsc     Cmd,Cmd_NACK        ; Is there a NACK?
          goto      SCNErr              ; Restart CmdExecution

          addwf     CRCr,F              ; add CRC
          movwf     INDADR              ; schreibe das W Register in den Ringbuffer
          incf      FSR,f               ; Schiebe den Zeiger des Buffers um 1 weiter

GetCV0L   call      SubInpExe           ; liest das nchste Byte ein
          movf      RxData,W            ; schreibe RxData ins W Register (low Byte)
          bcf       ADCON0,DataRdy      ; Free RxProcess
          btfsc     Cmd,Cmd_NACK        ; Is there a NACK?
          goto      SCNErr              ; Restart CmdExecution

          addwf     CRCr,F              ; add CRC
          movwf     INDADR              ; schreibe das W Register in den Ringbuffer
          incf      FSR,F               ; Schiebe den Zeiger des Buffers um 1 weiter

GetCV1H   call      SubInpExe           ; liest das nchste Byte ein
          movf      RxData,W            ; schreibe RxData ins W Register (high Byte)
          bcf       ADCON0,DataRdy      ; Free RxProcess
          btfsc     Cmd,Cmd_NACK        ; Is there a NACK?
          goto      SCNErr              ; Restart CmdExecution

          addwf     CRCr,F              ; add CRC
          movwf     INDADR              ; schreibe das W Register in den Ringbuffer
          incf      FSR,f               ; Schiebe den Zeiger des Buffers um 1 weiter

GetCV1L   call      SubInpExe           ; liest das nchste Byte ein
          movf      RxData,W            ; schreibe RxData ins W Register (low Byte)
          bcf       ADCON0,DataRdy      ; Free RxProcess
          btfsc     Cmd,Cmd_NACK        ; Is there a NACK?
          goto      SCNErr              ; Restart CmdExecution

          addwf     CRCr,F              ; add CRC
          movwf     INDADR              ; schreibe das W Register in den Ringbuffer

                                        ; CRC berprfung:
          call      SubGetCRCr          ; Get and Test CRC
          skpc                          ; C set if CRCr not equal
          goto      SCNCRCok            ;  OK

SCNErr    movlw     CV0h                ; Set both FSRs to CV0h
          movwf     FSR
          movwf     FSRInt

          clrf      Flags               ; Lsche Zhler und Achse
          clrf      CntH
          clrf      CntL

          clrf      CV0h                ; Lsche buffer
          clrf      CV0l
          clrf      CV1h
          clrf      CV1l
          call      SubShttrRst         ; Aus
          goto      SEErr               ; NACK:CRC,Restart CmdExecution

SCNCRCok  incf      FSR,F               ; Schiebe den Zeiger des Buffers um 1 weiter
          btfss     ExtCmd,SCN_Sht      ; Shutter preset:
          bcf       Flags,Preset        ; Aus
          btfsc     ExtCmd,SCN_Sht
          bsf       Flags,Preset        ; Ein

          goto      SetACKEXE           ; Set Cmd_Exe,Cmd_ACK flags and goto GetHdr

;
;--------------------------------------------------------------------
;
; Shutter Go
;
;
                                        ; Cmd direct action
SGO_Ext   call      SubGetCRCr          ; Get and Test CRC
          skpnc                         ; C set if CRCr not equal
          goto      SEErr               ; Set NACK: Syntax/CRCError

          movf      PORTB,W             ; Save Port B state
          movwf     OldPRB              ;
          bsf       INTCON,RBIE         ; Enable _IntPORTRB

          btfss     Flags,Preset        ; Shutter preset:
          call      SubShttrRst         ; Aus
          btfsc     Flags,Preset
          call      SubShttrStrt        ; Ein

;         goto      SetACKEXE           ; Set Cmd_Exe,Cmd_ACK flags and goto GetHdr


;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

SetACKEXE                               ; Finish command action with ACK and Exe flag set
SetA_E_   bsf       Cmd,Cmd_ACK         ; Set ACK Flag
          bsf       Cmd,Cmd_Exe         ; Set Exe Flag
          goto      RetTAC
;
;--------------------------------------------------------------------
;
; SubInpExe-Roundabout
; ----------------------
; Tests several Flags and desires. If an event occurs, the according
; routine may be executed:
; If Rx data is ready return to Cmd-Process
; If Cmd_NACK is set return to Cmd-Process to send NACK and restart Cmd-Proc.
; If Cmd_ACK is set Tx process is loaded with ACK data
; If Tx process is free, last computed data is sent and new data is calcul.
; If Cmd_Test flag is set, a Debugger data is added to be sent

                                        ; Check data input, output read and executes
                                        ; waiting data output sub.
                                        ; ==========================================
SubInpExe btfsc     ADCON0,DataRdy      ; Test if data is received?
Ret1      return                        ; If DataRdy from Rx return to Cmd Execution

          btfsc     Cmd,Cmd_NACK        ; Is there an error pending (NACK)?
Ret2      return                        ; return to Cmd_Exe routines, data will be ejected!

          btfsc     Cmd,Cmd_ACK         ; Is there a ACK? Test before other data is sent
          call      SubRetACK           ; Sub to send ACK...

          btfsc     ADCON0,DataRdy      ; Test if data is received?
Ret3      return                        ; If DataRdy from Rx return to Cmd Execution

          btfsc     Cmd,Cmd_NACK        ; Is there a NACK?
Ret4      return                        ; return to Cmd_Exe routines, data will be ejected!

          btfss     PStat,TxFlag        ; Is TxProcess free?
          call      SubTstOut           ; computed data output and calc new data

          if        _WDT=Yes
          clrwdt                        ; Reset watch dog timer
          endif
          goto      SubInpExe           ; LOOP: New Data ready test
;
;--------------------------------------------------------------------
;
; SubRetACK
; ---------
;
                                        ; Return ACK data
                                        ; ===============
SubRetACK btfsc     Cmd,Cmd_NACK        ; Did NACK occur ?
          return                        ; Yes

          movlw     Tx_ACK              ; Load Tx process with ACK
          call      SubSendW            ; Send w via Tx
          bcf       Cmd,Cmd_ACK         ; Reset ACK-Flag
          return
;
;--------------------------------------------------------------------
;
; SubSendW
; --------
;
; TxFlag is tested befor sending W
;
SubSendWT btfsc     Cmd,Cmd_Tst         ; If Cmd_Tst flag is set, send ASCII-HEX
          goto      SubSndWHx
SubSendW  btfss     PStat,TxFlag        ; Wait till Tx free
          goto      SSendW
          if        _WDT=Yes
          clrwdt                        ; Reset WDT
          endif
          goto      SubSendW
SSendW    movwf     TxData
          bsf       PStat,TxFlag        ; Sets command to start TxProcess
          return
;
;--------------------------------------------------------------------
;
; SubSndWHx
; ---------
;
; TxFlag is tested befor sending W as two ASCII values
;
SubSndWHx movwf     W1                  ; push
          swapf     W1,W                ; Get high nibble
          call      SendHalf            ; Send high nibble
          movf      W1,W                ; Get low nibble
SendHalf  andlw     0Fh                 ; Clear high
          addlw     30h                 ; Add ASCII '0'
WaitFree  btfss     PStat,TxFlag        ; Wait till Tx free
          goto      HxFree
          if        _WDT=Yes
          clrwdt                        ; Reset WDT
          endif
          goto      WaitFree
HxFree    movwf     TxData              ; Send low nibble
          btfss     TxData,3            ; Is >=8 ?
          goto      SndHlf              ;   no: send
          btfsc     TxData,2            ; Is >=0Ch ?
          goto      AddHx               ;   Yes: add correction
          btfsc     TxData,1            ; Is >=0Ah ?
          goto      AddHx               ;   Yes: add correction
          goto      SndHlf              ;   No: send

AddHx     movlw     7
          addwf     TxData,F            ; Correct A-F
SndHlf    bsf       PStat,TxFlag        ; Sets command to start TxProcess
          return
;
;--------------------------------------------------------------------
;
; SubGetCRC
; ---------
;
; Gets and tests CR-Code. Sets C if not equal
;
SubGetCRCr if       _Tst=Yes
          clrc                          ; No CRC
          else
          call      SubInpExe           ; liest das nchste Byte ein
          movf      RxData,W            ; schreibe RxData ins W Register (low Byte)
          bcf       ADCON0,DataRdy      ; Free RxProcess
          btfss     Cmd,Cmd_NACK        ; Is there a NACK?
          goto      NoNACK              ; Restart CmdExecution

GCRCErr   setc
          return

NoNACK    subwf     CRCr,W              ; Are they equal ?
          skpz                          ;
          goto      GCRCErr             ;  No: error

          clrc                          ;  Yes: go ahead
          endif
          return
;
;--------------------------------------------------------------------
;
; SubShttrInv
; -----------
;
; Shutter output will be inverted to present state
;
SubShttrInv btfsc   PORTB,ShttrStrt     ; Ist Start gesetzt ?
          goto      SubShttrRst         ;   Reset aktivieren
          btfsc     PORTB,ShttrRst      ; Ist Reset gesetzt ?
          goto      SubShttrStrt        ;   Start aktivieren
          goto      SubShttrRst         ; SONST Reset aktivieren
;
;--------------------------------------------------------------------
;
; SubShttrRst
; -----------
;
; Shutter output will be reseted
;
SubShttrRst bcf     Flags,ShtState      ; Mark state
          bcf       PORTB,ShttrStrt     ; Start pin low
          bsf       PORTB,ShttrRst      ; Reset pin high
          return
;
;--------------------------------------------------------------------
;
; SubShttrStrt
; ------------
;
; Shutter output will be set (start)
;
SubShttrStrt bsf    Flags,ShtState      ; Mark state
          bcf       PORTB,ShttrRst      ; Reset pin low
          bsf       PORTB,ShttrStrt     ; Start pin high
          return
;
;--------------------------------------------------------------------
;
; SubTestDataOut
; --------------
;
; (TxFlag is cleared at first data)
;
                                        ; Test if any Cmd has to output data
                                        ; ==================================
SubTstOut btfss     Cmd,Cmd_Exe         ; Is Command ready to be executed?
          return                        ;  No: go back

          btfsc     Cmd,Cmd_ACK         ; Is ACK to be sent?
          return                        ;  Yes: go back

          btfsc     Cmd,Cmd_RCV         ; Return Coordinate Value ?
          goto      RCV_Out             ;  Yes
          btfsc     Cmd,Cmd_SCV         ; Set Coord. Val ?
          goto      SCV_Out             ;  Yes
          btfsc     Cmd,Cmd_SSV         ; Set Start Val ?
          goto      SSV_Out             ;  Yes
          btfsc     Cmd,Cmd_RST         ; Free Shutter ?
          goto      RST_Out             ;  Yes

          goto      EXT_Out             ;  Yes

;
;--------------------------------------------------------------------
;
; Execute-SubProcesses
; --------------------
; During SubInpExe-time with no Input this Exe-Subs will be started
; according to the loaded command
;
                                        ; Exe Sub Shoot X
                                        ; ===============

RCV_Out   movf      Flags,W             ; schreibe Flags/Axis ins W Register
          if        _Tst=No
          movwf     CRCs                ; Save for CRC
          endif
          call      SubSendWT           ; Sende Flags


          movf      CntH,W              ; schreibe CntH ins W Register
          if        _Tst=No
          addwf     CRCs,F              ; Calc CRC
          endif
          call      SubSendWT           ; Sende CntH

          movf      CntL,W              ; schreibe CntL ins W Register
          if        _Tst=No
          addwf     CRCs,F              ; Calc CRC
          endif
          call      SubSendWT           ; Sende CntL

          movf      Cnt,W               ; schreibe Cnt ins W Register
          if        _Tst=No
          addwf     CRCs,F              ; Calc CRC
          endif
          call      SubSendWT           ; Sende Cnt


          call      SubGetCVW2          ; Holt CVwerte via Cmd ins W und W2
          if        _Tst=No
          addwf     CRCs,F              ; Calc CRC
          endif
          call      SubSendWT           ; Sende CVh

          movf      W2,W                ; hole low Wert
          if        _Tst=No
          addwf     CRCs,F              ; Calc CRC
          endif
          call      SubSendWT           ; Sende CVl

          if        _Tst=No
          movf      CRCs,W              ; Get CRC
          call      SubSendWT           ; Sende CRC
          endif

          movlw     070h                ; End of Command
          andwf     Cmd,F
          bsf       Cmd, Cmd_ACK        ; Send ACK at the End of the Commands
          goto      TstDbg
;
;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
                                        ; Exe Set Coord. Val
                                        ; ==================

SCV_Out   movlw     070h                ; End of Command
          andwf     Cmd,F
;         bsf       Cmd, Cmd_ACK        ; Send ACK at the End of the Commands
          goto      TstDbg
;
;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
                                        ; Exe Set Start Value
                                        ; ===================

SSV_Out   movlw     070h                ; End of Command
          andwf     Cmd,F
;         bsf       Cmd, Cmd_ACK        ; Send ACK at the End of the Commands
          goto      TstDbg              ; Protocol has no returned data, clear Cmd
;
;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
                                        ; Free Shutter
                                        ; ============

RST_Out   movlw     070h                ; End of Command
          andwf     Cmd,F
;         bsf       Cmd, Cmd_ACK        ; Send ACK at the End of the Commands
;         goto      TstDbg
;
;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
TstDbg    btfss     Cmd,Cmd_Tst         ; if Test-Flag is set a debug routine is added after each cmd!
          return
                                        ; Debug routine:
                                        ; ==============
          if        _Tst=Yes
          movlw     0Dh
          Call      SubSendW
          movlw     0Ah
          Call      SubSendW
          endif

          return
;
;---------------------------------------------------------------------------
;
; Extended commands output
; ========================
; Do quit subroutine with return (not goto TstDbg)
;
EXT_Out   btfsc     ExtCmd,Ext_SCN      ; Scan ?
          goto      SCN_Out             ;  Yes

          btfsc     ExtCmd,Ext_SGO      ; Shutter Go ?
          goto      SGO_Out             ;  Yes

          goto      SEErr               ; SyntaxError when Ext = 00h.

;
;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
                                        ; Ext Cmd output
                                        ; ==============
SCN_Out   movf      Flags,W             ; schreibe Flags/Axis ins W Register
          if        _Tst=No
          movwf     CRCs                ; Save for CRC
          endif
          call      SubSendWT           ; Sende Flags


          movf      CntH,W              ; schreibe CntH ins W Register
          if        _Tst=No
          addwf     CRCs,F              ; Calc CRC
          endif
          call      SubSendWT           ; Sende CntH

          movf      CntL,W              ; schreibe CntL ins W Register
          if        _Tst=No
          addwf     CRCs,F              ; Calc CRC
          endif
          call      SubSendWT           ; Sende CntL

          movf      CV0h,W              ; schreibe CV0h ins W Register
          if        _Tst=No
          addwf     CRCs,F              ; Calc CRC
          endif
          call      SubSendWT           ; Sende Cnt

          movf      CV0l,W              ; schreibe CV0l ins W Register
          if        _Tst=No
          addwf     CRCs,F              ; Calc CRC
          endif
          call      SubSendWT           ; Sende CVh

          movf      CV1h,W              ; schreibe CV1h ins W Register
          if        _Tst=No
          addwf     CRCs,F              ; Calc CRC
          endif
          call      SubSendWT           ; Sende Cnt

          movf      CV1l,W              ; schreibe CV1l ins W Register
          if        _Tst=No
          addwf     CRCs,F              ; Calc CRC
          endif
          call      SubSendWT           ; Sende CVh

          if        _Tst=No
          movf      CRCs,W              ; Get CRC
          call      SubSendWT           ; Sende CRC
          endif

          movlw     070h                ; End of Command
          andwf     Cmd,F
          clrf      ExtCmd
          bsf       Cmd,Cmd_ACK         ; Send ACK at the End of the Commands
          goto      TstDbg
;
;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
                                        ; Exe Set Coord. Val
                                        ; ==================

SGO_Out   movlw     070h                ; End of Command
          andwf     Cmd,F
          clrf      ExtCmd
;         bsf       Cmd, Cmd_ACK        ; Send ACK at the End of the Commands
          goto      TstDbg
;
;
;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
__LastInstr data    "ENDE"
          end

