;
;                             #.ASM
;                             =============
;          -MpK 1.0- with interrupt handler for PIC16LC71
;
;**************************************************************************
;
;Creator : M.Kellenberger, 18.11.93, IMS, Neu Technikum Buchs,CH-9470 Buchs
;          ( TEL.: ++49-81-755 34 44; E-MAIL: kellenbe@ntb.ch )
;System  : PC, MPALC, MPSTART, PDC.BAT (uses 4DOS.COM)
;Target  : Microchip PIC 16C71
;Editor  : Boxer Version 6.0, Tab expanded to 10 spaces!
;Support : Only on this original source
;Errors  : Please report with comment and jrn-file
;Warnings: max. 8 Warnings are "normal" during compilation:
;          MPK.asm(918,0);Warning:..: Filereg masked to appropriate number of bits
;          MPK.asm(927,0);Warning:..: Filereg masked to appropriate number of bits
;          MPK.asm(942,0);Warning:..: Filereg masked to appropriate number of bits
;          MPK.asm(971,0);Warning:..: Filereg masked to appropriate number of bits
;          MPK.asm(1504,0);Warning:..: Value truncated to 8 bits
;          MPK.asm(1592,0);Warning:..: Value truncated to 8 bits
;          MPK.asm(1804,0);Warning:..: Value truncated to 8 bits
;          MPK.asm(1911,0);Warning:..: Value truncated to 8 bits
;Waranty : None
;Copyright by M.Kellenberger, NTB, CH.
;          (Source may be copied freely only with this header)
;
;
;History MpK:
;V 0.0    21.10.93 mik        Definitions, systemlabel start with "_"!
;V 0.1    22.10.93 mik        System design and description
;V 0.1a   25.10.93 mik        Add "#" to user-changeable lines
;V 0.1b   15.11.93 mik        Made systembody faster (20MHz)
;V 0.1c   16.11.93 mik        MpK Kernel via simulater tested!
;V 0.2    25.11.93 mik        RxError, RxBusy and Cmds!
;V 0.3    13. 1.94 mik        RS-485 Enable implemented
;V 0.4    21. 1.94 mik        Compiler flags _Tst,_MHz,_RS implemented
;V 0.5    27. 1.94 mik        Compiler flags _Tst into TstDbg implemented
;V 0.6    28. 1.94 mik        Compiler flags _WDT implemented
;V 0.7     8. 2.94 mik        Ext_Cmd is 001T'0000 (Flag T is now also for EXT Cmds)
;         10. 2.94 mik        reset condition into Cmd
;V 0.8    21. 2.94 mik        CRC added, _Peven, _Parity deleted
;V 0.9    21. 2.94 mik        CRCs/r added
;
;                             Three projects were (are) launched with this body...
;                               - Intelligent sensor to messure thread tension. (sensation!)
;                                 Gave the kick to design MpK!
;                               - Front end for a "laser pattern generator" (step tracker)
;                               - Switched capacitor controller. (->private company)
;                             ...beta test passed!
;
;                             Remarks: - "easy to learn and use"      (J.Wagner)
;                                      - "good programming philosophy"(L.Grecco)
;                                      - "short time to release"      (G.Staufert)
;                                      - "don't bother about basic stuff,
;                                         it's managed by the MpK"    (L.Grecco)
;                                      - "they got it"                (M.Kellenberger)
;
;V 1.0    10. 8.94 mik        Last test, more text, first public release!
;
;User programmers revision:
;V 1.0    ##.##.## ___        First attempt: #
;
;*******************************************************************
;
; Description:
; ============
;
;   1. Multi process Kernel System body (design and tested only for PIC16C71):
;   --------------------------------------------------------------------------
;
;    The MpK is designed as an (empty) programm body for several perposes:
;          * Management of all interrupt launched processes.
;          * Management for cyclic processes via RTCC (one timebase).
;          * Usage of as little amount of PIC resources as possible for the MpK.
;          * Serial communication up to 19.6kBd (half/full duplex).
;          * Simple serial protocol. Ability to address 253 nodes (RS485) optionaly.
;          * So far 1(or2)-byte-command supported -> may cause any action in the
;            main process. (see chapter 2)
;          * Quick realisation of user programms with that (and other) needs..
;
;
;         The MpK consists of (a user defineable body for):
;          * Predefined register- and macrosets (see also new "PicReg.equ" and
;            "Pic71.mac")
;          * Proper initialization of HW and status/option registers,
;          * An INTerruptHANDLER (with simple priority managment),
;          * One process for each interrupt source (managed by INTHANDLER),
;          * 4 cyclic processes: (the cycle time base is managed by
;                                 the RTCC-int.handler)
;            - Tx, Rx0, Rx1:
;                for a half/full duplex communication (RS232/485->driver needed)
;            - Adjust- or user-process:
;                for cyclic time adjustment or an other user defined cyclic
;                process that base on that RTCC timebase.
;
;          * May launch the AD-conversion repeatedly
;              (calculated out of the cycle timebase)
;          * One (empty) MAIN (user) process
;              (suggestion: round-robin-scheduler, see chapter 2).
;          * Serial communication is half/full duplex and may be setup easly.
;              (acts almost like a built in SCI)
;          * Semaphore (flags) for data transfer to/from MpK.
;
;          Options:
;          * One FSR during interrupt cycle and one other FSR during main cycle.
;          * Simple implementation of a protocol (-> chapter 2).
;          * Simple implementation of CRC for sending and receiving line.
;          * Simple implementation of 254 addressed nodes.
;
;
;         Resources used by MpK:
;          * LESS than 600 bytes programm memory when used all features.
;          * INT interrupt as serial data receiver.
;          * RB0,RB1 (and optional RB3 for TxEn) pin.
;          * RTCC and its interrupt as cycle time base generator.
;          * Prescaler (in some cases not, see calculations).
;          * Unused Flags RP1, IRP and "ADCn(ot)used" (see PicReg.equ).
;          * Ports RP0, RB1 for Rx and Tx.
;          * Port RP2 TxEnable if used for driver, else free.
;          * Only 7 Byte MpK reserved fileregisters including 1 command bytes.
;          * Two more reserved bytes if two FSRs are desired.
;          * Two more reserved bytes if CRC (Rx and Tx) is desired.
;          * WDT is embedded.
;          * All other interrupts are embedded.
;
;
;
;
;
;
;    2. Philosophy of the MpK
;    ------------------------
;         The idea for this MpK came up when I had to design a controller for
;         my thread sensors. I wanted to create source code that is easy to
;         expand and maintain. Also there was the idea to split up the
;         whole problem into two parts:
;         - The background running processes (put into MpK):
;                   SCI, AD sampler, MUX switcher, periodic actions, ...
;         - The main program (pseudo processes):
;                   Protocol, cmd actions, data analising, ...
;         All should be usable for other projects!
;
;
;         2.1 Process structure
;         ---------------------
;         (see MpK_Proc.doc)
;         The heart of MpK is the Interrupt brancher called _IntHndlr.
;         Any interrupt will endup here and depending to its priority
;         (set via the position of comparison and reentrance) cause a
;         branch to its associated handler process:
;
;         Interrupt Process    Action
;         ----------------------------------------------------------
;         PortRB    _IntPORTRB Caused by any change on RB4-7 inputs.
;         A/D       _IntAD     Caused by the end of conversation of
;                              the built in A/D converter.
;         INT       _INT       Caused by a falling edge at RB0. This
;                              Input is used as Rx-pin. The "_INT"
;                              sets RxFlag and selects the nearest
;                              RxProc (Proc01 or Proc11) by RxSel.
;                              Any further INT will be cut off by
;                              disabling INT (for the next received
;                              byte).
;         RTC       _IntRTCC   Is used to maintain a periodically
;                              switching process launcher:
;                              By incrementing PCnt from 00b to 11b
;                              4 processes may be launched in a cyrcle:
;
;                   _TxProc    (Proc00) If TxFlag is set the process will
;                              serial out the data in TxData file. After
;                              the last stop bit, TxFlag will be resetted
;                              to signal "ready for next data".
;                   _RxProc    (Proc01) If INT has occured between Proc11
;                              and Proc01 this process is selected by _INT
;                              (/RxSel).
;                              The process tries to get the received bit into
;                              file RxData. Any errors are marked.
;                              After the stop bit RxFlag is resetted and the
;                              flaf DataRdy will be set to signal available
;                              data in RxData file.
;                              After read the data out of RxData the user must
;                              manually reset the flag DataRdy to signal "free
;                              line" for RxProc.
;                   _AdjProc   (Proc10) This process may be used for any time
;                              adjustment for this cycle or used otherwise.
;                   _RxProc    (Proc11) If INT has occured between Proc01
;                              and Proc11 this process is selected by _INT
;                              (RxSel) as active RxProc. See above for details.
;
;                   _ADProc    There is another counter to start the AD-
;                              converter periodically: SRate.
;                              This counter also depends on the timebase in
;                              _IntRTCC.
;
;         ---       _Main      This process is not capable to be split up into
;                              several process by interrupt. So if more than
;                              one is desired (as suggested) a Round-Robbin-
;                              Scheduler must be installed. See below.
;
;
;
;         2.2 HW-requirements
;         -------------------
;           Only a battery (6V), the crystal (16MHz), a reset network (R 10k,C 10uF)
;           and a serial driver (Max232|SN79176|SN79178) are necessary for
;           its minimal usage. Of course a PIC16C71 is also required...
;
;         2.3 Main process
;         ----------------
;           This process is not capable to be split up into several process
;           by interrupt. So if more than one is desired (as suggested) a
;           Round-Robbin-Scheduler must be installed. The suggested way is
;           because of the need of a serial communication in a absolutly
;           slave mode. No action will be done without any causing command.
;           Any background actions has to be put into a process launched by
;           _IntHandlr or _IntRTCC.
;
;           2.3.1 Communication to MpK
;           --------------------------
;            There shall be three ways of communicate with the MpK-Main:
;
;            Flags:/Comm:  F-Dplx/2p   F-Dplx/Bus   H-Dplx/2p   H-Dplx/Bus
;            --------------------------------------------------------------
;             __RS         232         232(=422)    485         485
;             __TxEn       No          Yes/No *     Yes         Yes
;             __Addr       No          Yes          No          Yes
;             (Addresses   -           0,1-253      -           0,1-253)
;             (Tail byte   No          Yes          Yes         Yes)
;
;            The flag __RS switches between "Full duplex" and "Half duplex".
;            If RS485 is used with 4 lines (=422 driver), Full duplex is
;            selected by __RS=232 and * __TxEn=Yes/No according to the drivers
;            needs!
;            The pin TxEn is switched for an Enable-pin at any driver device.
;            The simple protocol maintains only a master-SLAVE communication.
;            MpK is ment to be the slave (here).
;
;           2.3.2 Round Robin Scheduler as switch center
;           --------------------------------------------
;            There is a branching routine called SubInputExe
;            that switches to any Sub by flags. All these subs
;            must return within a small timeslice! Only this
;            way any realtime reaction can be achiefed.
;            The whole serial protocol can be set up this way.
;            Any command sets flags and afterwards the flags
;            cause the SubInputExe to branch to the according
;            CmdSub. There is a special programming style used
;            for the receiving part called CmdProc. Actually
;            this process runs off sequently and always waits
;            for the next data. Thats way _Main starts with
;            this process and during waiting time the SubInputExe
;            is called!
;
;           2.3.3 Simple protocol
;           ---------------------
;            A byte based protocol is assumed to be simple and easy:
;
;            >>>> Hdr,[Adr],[Ext],Cmd,[Data],[CRC],[Tail]
;            <<<< NACK
;            <<<< ACK,[Data,[TestData],[CRC],ACK]
;
;            ------ Master call -----
;            A unique header (FFh) is received first. Any lost slave may be hang
;            in again at this header.
;            Next the optional Address follows. 00h is an ListenAll node.
;            Cmd byte is next. Each bit represents a Cmd. By reading it from
;            MSB to LSB the command is selected by 1b and all followed bits
;            may be used for this command:
;
;            XXXT1yyyb is first command with 3 bits free (yyyb)
;            XXXT01yyb is second command with 2 bits free
;            XXXT001yb is third command with 1 bit free
;            XXXT0001b is last command with no free bit.
;
;            Remark on XXX:
;              The bit 7,6,5 are fixed to "001...." during sending to achieve
;              easy teminal testing "21h('!')..3Fh('?').
;              This bits are also used for Cmd-Flags within the Cmd file.
;
;              XXX=001 during sending, XXX=Exe/Ack/NAck-Flags) within Cmd
;
;            Bit 4 is reserved to indicate a test mode. It's used to suppress
;            CRC, convert output to ASCII and add any test data and CR/LF
;            for terminal testing.
;            So there are 4 commands to be selected. If more than 5 commands are
;            required and Ext-Cmd followed by an other cmd byte may be used
;            by sending 20h (00100000b).
;            Then additional 8 commands are available and so on...
;            After read this command the the DoAction-part of the received
;            cmd is executed. Any additional data may be read in and put off
;            or what ever. So every received data must be processed immediately!
;            An optional CRC data may be sent after that.
;            A Tail mark (00h) is sent when using addressable mode (__RS=485).
;
;            ----- Slave response -----
;            If there is any error a NACK will be sent and the communication
;            breaks up here. In Full duplex mode this NACK will be sent
;            immediately not waiting till tail is received!!!!
;
;            Normally a ACK indicates "Got cmd" and if any data are asked
;            they will be sent back immediatly.
;            On setted T-flag in Cmd file any test datas are added.
;            The CRC is added optionally too.
;            An (pseudo) ACK indicates the end of response.
;
;            Remark: There is no other reaction of the master except the next cmd.
;
;           2.3.4 Receiving "process"/actions
;           ---------------------------------
;            When a cmd byte is received within the Cmd-process it will branch
;            to the associated "DoAction" part. There the next datas must be
;            received manually. Also the CRC(r) must be programmed manually by
;            adding a Statement to the source. The final CRC comparison must be done
;            at the end of its data stream (SubGetCRCr).
;
;           2.3.5 Sending "process"/actions
;           -------------------------------
;            The cmd byte that is received within the Cmd-process is stored and
;            causes the RR-Scheduler to switch to the associated "CmdSub-Action"
;            part. They are designed to send any analysed/processed/bufferd
;            data back to the master. Any CRC(s) must be incremented manually by
;            adding a statement to the source. The CRC must then also be sent at
;            the end of its data stream.
;
;
;         2.4 Other features
;         ------------------
;           There or some other features implemented
;
;           2.4.1 __Tst compiler/assembler flag
;           -----------------------------------
;            Ability to put all "test run" routines into dependence of that
;            flag.
;            CR/LF, reset behavior and CRC are affected.
;
;           2.4.2 Test flag (Cmd bit.4)
;           ---------------------------
;            Some routines are delivered to maintain a test mode.
;            - TstDbg:   Routine to add test data.
;                        _Tst=Yes:Addition of CR/LF to any sending data.
;            - SubSendW: Sending of the W file with conversion to ASCII-HEX.
;                        (2byte: 'A'->'40')
;
;           2.4.3 Reset behavior
;           --------------------
;            With __Tst=Yes after every reset the MpK sends a reset code to the
;            master. In this way the reset behavior may be tested.
;             'r': Reset
;             'S': WDT occured in sleep mode
;             'W': WDT occured during running
;             'R': Reset pin was pulled
;            This affects RS485 networking of course!!!!
;
;
;           2.4.4 WDT usage
;           ---------------
;            With the flag __WDT the watch dog usage may be switched.
;            Within the RR-Scheduler the WDT will be reseted periodically.
;
;           2.4.5 New pseudo vectores
;           -------------------------
;            There are 3 new pseudo vectores for the start up behavoirs.
;            It's just to have a similar handling as for the normal reset!
;              * _WDTSLP : Reenter after WDT wake-up from SLEEP
;              * _WDTTO  : Reenter after WDT time-out
;              * _MCLRSLP: Reenter after /MCLR wake-up from Sleep
;            These vectores must point to the according entry point.
;
;           2.4.6 Usage of two FSRs
;           -----------------------
;            The MpK is designed to run off in the background by using the
;            interrupt capability of the '71.
;            If the user needs to have the FSR file within the main process
;            that register should never be changed by background processes!!
;            If a user specified background processes need to have an own FSR
;            (bg-proc: buffer write; main: buffer read) this will be achieved
;            by the two files FSRMain and FSRInt. At every Interrupt the FSR will
;            be setup with the according value.
;
;
;         2.5 Timing calculations
;         -----------------------
;          The _IntRTCC uses a time base to increment the PCnt pointer. This
;          time base gives the baud rate and the lowest AD start cycle.
;          Calculations are made at the constant settings.
;          Prescaler is used to step down by 2 or more for lower cycles.
;          The reset value (const Rtcc) for the RTCC is calculated out
;          of the instruction cycle time t. (16MHz=250ns)
;          (2 t's must be subracted due to some misbehavior of the RTCC!)
;          The time_base to launch Proc00..11 is: Rtcc*t. (16MHz,104=26us)
;          The baud rate is calculated out of 4*Rtcc*t. (16MHz=9.6kBd)
;
;           2.5.1 Time base adjustment
;           --------------------------
;            If there is no chance to get correct baud rates just correct the
;            time base once a cycle by adding or subtracting t's from the RTCC
;            within the Proc10 (_AdjProc).
;
;
;         2.6 Code optimization
;         ---------------------
;           To gain additional space and speed optimization according to the
;           users program needs may be done. Some hints:
;
;           - Some redundant nops are included to visualize labels before
;             IF statements within MPSIM. Delete "nop" if not appreciated.
;           - Some not used Compiler flags may be wiped out with their inluded
;             parts. (Better listings)
;           - Put unused MpK proc into comments to gain speed
;	  - If never-disabled interrupts are used, get ride of the
;             test sequenz in front of every int proc.
;           - Optimize _IntHndlr: FSR exchange, Z-Flag
;           - Get ride of all the unused MpK file regs (FSR..., CRC., Ext)
;
;
;
;   All the empty modules (marked with #) may be filled up by the user
;   (programmer) and described in the following chapter!
;
;    3. #
;    ---------
;         This program ...
;
;
;
;
; NOTES (to the source):
; ----------------------
;         1. Labels with the starter "_" are MpK system labels
; >>>>>>> 2. <#> marks user changeable values/parts <<<<<<<<<<<<<<<<<<<<<<
;             Browse through the source code and replace all <#> with your
;             code and description !!!
;         3. Compiler/Assembler vars are:
;            __RS      = 232/485
;            __TxEn    = Yes/No  \
;            __Addr    = Yes/No   > __Tail automatically
;            __RxInv   = Yes/No  /
;            __TxInv   = Yes/No
;            __ADInv   = Yes/No
;            __MHz     = 20,16,12,...
;            __Tst     = Yes/No
;            __WDT     = Yes/No
;         4. ".DS" is used when it occured as suitable
;
;
;*******************************************************************
;
;Assembler options: (# change flags !)
;


          LIST C=100,E=0,N=60,F=INHX8M,P=16C71,R=DEC,L=ON

Yes       set       1         ; !!! DO use this for compiler directives !!!
No        set       0         ; !!! Do NOT use "YES" or "NO" for them   !!!

__RS      set       485       ; 232 for point-to-point-connection via RS232 connection
                              ; (Full duplex! RS422 drivers are also possible)
                              ; 485 for (multi-)2-wire-RS485-bus (optional output pin TxEn)
                              ; (Half duplex!)

__Addr    set       Yes       ; Yes for addressable nodes (RS485 or RS422)
                              ; No for no addressing and tailing code

__TxEn    set       Yes       ; Yes to enable a TxEnable output pin (Used for RS485/422)
                              ; No to disable a TxEnable output pin

__Tst     set       No        ; Yes for test sample:
                              ;         Adds 0Dh0Ah if testflag in Cmd is set
                              ;         No CRC needed (assumed as correct)!
                              ;         Reset conditions are sent after every reset (__RS232)!
                              ;         Halt on lost cycle within RTCC interrupt handler!
                              ; No for serie!

__RxInv   set       No        ; Yes for Rx Bits to be inverted
                              ; No for Rx Bits not to be inverted

__TxInv   set       No        ; Yes for Tx Bits to be inverted
                              ; No for Tx Bits not to be inverted

__ADInv   set       Yes       ; Yes for ADRES to be inverted
                              ; No for ADRES not to be inverted

__WDT     set       Yes       ; Yes for active WDT
                              ; No for no WDT!

__MHz     set       16        ; 16 for 16MHz osc. connection
                              ; (or 12 for 12MHz osc. connection)
                              ; (or what else ....)
;-----------------------------------------------------------------------
; Automatic set:
__Tail    set       Yes       ; A tail byte is needed when wiring a bus or half duplex mode:
          IF        __RS=232
          IF        __Addr=No
__Tail    set       No
          ENDIF
          ELSE
__TxEn    set       Yes       ; TxEn must be "Yes" on __RS=485 !
          ENDIF
;
;*******************************************************************
;
; Prommer information:
; --------------------
;
; ID VALUES                   ; Burn this within "MPSTART.EXE"
;
; 2000h   3FFFh               ; Code Revision Number
; 2001h   3FFFh               ; Code Manufacturing Number
; 2002h   3FFFh               ; Code ...
; 2003h   3FFFh               ; Code ...
;
;CONFIGURATION WORD: (2007H)
;
;Select   FOSC1     FOSC0     ; OSC Selection Fuses
;         0         0         ;  LP
;         0         1         ;  XT
;  >      1         0         ;  HS a must ! (__MHz=20..5)
;         1         1         ;  RC (default)
;
;         WDTE                ; WDT Enable Fuse
;         0                   ;  disable
;  #      1                   ;  enable (default)
;
;         PWRTE               ; Power Up Timer Fuse
;         0                   ;  disable
;  #      1                   ;  enable (default)
;
;         CP                  ; Code Protection Fuse
;         0                   ;  On
;  #      1                   ;  Off (default)
;
;
;*******************************************************************
;
;Includes
;

          INCLUDE   "PicReg.equ"        ; modified for PIC16C71
          INCLUDE   "Pic71.mac"         ; page switching

;
;*******************************************************************
;
;Constants  (only change calculated values within brackets)
;
; Timebase:
; =========
;
Presc     equ       (b'000')            ; Prescaler Value   000=1:2   #
                                        ;                   001=1:4
                                        ;                      =
                                        ;                   111=1:256

                                        ; Calculation: (also for serial IF baudrate)
                                        ; ------------
                                        ; RTCC set value (minus corr. cycles for HW (2)delay)

          IF        __MHz=20            ; >>>>>>>>> 20MHz t=0.2us <<<<<<<<<<<<
Rtcc      equ       (130)-2             ;130 *  0.2us  = 26us (timebase)
                                        ;  4 * 26us    =104us => 9.6kBaud
                                        ;---------------------------------
                                        ; 65 *  0.2us  = 13us (timebase)
                                        ;  4 * 13us    = 52us => 19.2kBaud
          ENDIF

          IF        __MHz=16            ; >>>>>>>>> 16MHz t=0.25us <<<<<<<<<<<
Rtcc      equ       (104)-2             ;104 * 0.25us  = 26us (timebase)
                                        ;  4 * 26us    =104us => 9.6kBaud
          ENDIF

          IF        __MHz=12            ; >>>>>>>>> 12MHz t=0.333us <<<<<<<<<<<<
Rtcc      equ       (78)-2              ; 78 * 0.333us = 26us (timebase)
                                        ;  4 * 13us    =104us => 9.6kBaud
          ENDIF
;         IF
                                        ; ...
;         ENDIF


Adj       equ       (0)+2               ; Adjust value (plus (X * t) within 4 timebases)

                                        ; Calculation:
                                        ; ------------
SRate     equ       100h-(6)*10h        ; Sample rate for (multiplexed) AD-conversion
                                        ;
                                        ; Process timebase t=26us
                                        ;  2 *  26us => (19,2kHz)
                                        ;  3 *  26us => (13,9kHz)

                                        ; Process timebase t=13us
                                        ;  4 *  13us => (19,2kHz)
                                        ;  5 *  13us => (15,4kHz)
                                        ;  6 *  13us => (13,9kHz)
                                        ; ##

                                        ; ----------------------------------------------
; MpK reg bits:
; =============

PCnt0     equ       00h                 ; PStat Reg. Process counter bit 0
PCnt1     equ       01h                 ; PStat Reg. Process counter bit 1
RxFlag    equ       02h                 ; PStat Reg. 1=Rx is active
TxFlag    equ       03h                 ; PStat Reg. 1=Tx is active
                                        ; PStat Reg. Bit 4-7: Countdown for sample rate

DataRdy   equ       ADCnusd             ; DataReady Flag in reg. ADCON0:
                                        ; # Must be reset via user
                                        ;   when data is read !!!

RxStart   equ       00h                 ; TxRxFlg Reg. 1=Rec. start bit
RxStop    equ       01h                 ; TxRxFlg Reg. 1=Rec. stop bit
RxSel     equ       02h                 ; TxRxFlg Reg. Select Rx process 0=01,1=11
TxStart   equ       03h                 ; TxRxFlg Reg. 1=Tx is active
TxStop    equ       04h                 ; TxRxFlg Reg. 1=Tx is active
                                        ; TxRxFlg Reg. Bit 5-7: bit counter Tx

                                        ; Bits in STATUS (Main!) or SaveST (Interrupt routine!):
RxError   equ       RP1                 ; Flag to mark Rx Error start/stop/parity
RxBusy    equ       IRP                 ; Flag to mark Rx busy (INTterupt while DataRdy=1)


; MpK Port Pins:
; ==============

                                        ; PORTB:
RxPin     equ       00h                 ; Pin RS-232-C Receiver
TxPin     equ       01h                 ; Pin RS-232-C Transmitter
          IF        __TxEn=Yes
TxEn      equ       02h                 ; Pin Enable Tx of any drivers need
          ENDIF

; User Port Pins:
; ===============

                                        ; PORTB:
                                        ; #:

                                        ; PORTA:
                                        ; #:


; Protocol:
; =========

                                        ; RxData:
                                        ; -------
Rx_Hdr    equ       0FFh                ; Unique data on Rx line to indicate Cmd start
                                        ; ( It must be 0FFh !!)

          IF        __Tail=Yes          ; Addressable nodes:
Rx_Tail   equ       0FEh                ; Data on Rx line to indicate Cmd end
          ENDIF                         ; ( It must be 0FEh !!)

          IF       __Addr=Yes		; Addressable nodes:
Rx_Addr   equ      20h                  ; Pic node address (01h-FDh) #
Rx_LAll   equ      00h                  ; Pic node address for all to listen (no ACK)
                                        ; ( It must be 00h !!)
          ENDIF

                                        ; Cmd bits: (Compared in the order bit 4 to bit 0)
                                        ; ==============================================

                                        ; Process flow bits:
                                        ; ------------------
Cmd_Exe   equ       07h                 ;  Command is executed
Cmd_NACK  equ       06h                 ;  NACK must be sent immediatly
Cmd_ACK   equ       05h                 ;  ACK must be sent immediatly

                                        ; Command bits (001x xxxx):
                                        ; -------------------------
Cmd_Tst   equ       04h                 ; Out-routines call TstDbg routines after every 'Data-out'
Cmd_1xxx  equ       03h                 ;  #
Cmd_01xx  equ       02h                 ;  #
Cmd_001x  equ       01h                 ;  #
Cmd_0001  equ       00h                 ;  #
                                        ; 20h=Extended commands:

                                        ; Extended commands:
                                        ; ------------------
Ext_XXh   equ       07h                 ;  # (all 8 bits may be used)

                                        ; Listen All commands (same constuction as Cmd):
                                        ; ----------------------------------------------

LAC_Tst   equ       04h                 ; LOut-routines call TstDbg routine after every 'Data-out'
LAC_1xxx  equ       03h                 ;  #
LAC_01xx  equ       02h                 ;  #
LAC_001x  equ       01h                 ;  #
LAC_0001  equ       00h                 ;  #

                                        ; 20h=Extended LA-commands:
                                        ;  (User may define own LACmdExt's...)

; TxData:
; =======

          IF        __Tst=Yes           ; This bytes are sent after reset condition if its bit7 is set
Tx_Rst    equ       'r'                 ; Reset
Tx_WDTSLP equ       'S'                 ; WDT occured in sleep mode
Tx_WDTTO  equ       'W'                 ; WDT occured during running
Tx_MCLRS  equ       'R'                 ; Reset pin was pulled
          ENDIF

Tx_SErr   equ       0FEh                ; NACK: Syntax Error occured
Tx_RxBusy equ       0FDh                ; NACK: Rx is still busy while new data is pushing
Tx_RxErr  equ       0FCh                ; NACK: Rx start/stop bit error occured
Tx_ACK    equ       0FBh                ; ACK:  ''
                                        ; #...

Tx_DataMx equ       0FBh                ; 00h-FAh=data


;
;*******************************************************************
;
;Variables
;

SaveW     equ       0Ch                 ; Interrupt save register W
SaveST    equ       0Dh                 ;    "       "   STATUS
RxData    equ       0Eh                 ; Received byte
TxData    equ       0Fh                 ; Byte to send
TxRxFlg   equ       10h                 ; Tx & Rx Flags and counter
Cmd       equ       11h                 ; Command register:           00h=no cmd
PStat     equ       12h                 ; Process register:
                                        ; Bit
                                        ; 0,1:    RTCC Process Counter:         00=Tx
                                        ;                                       01=[Rx]
                                        ;                                       10=#(Adjust)
                                        ;                                       11=[Rx]
                                        ; 2,3:    Rx/TxFlags: 1 if receiving/sending
                                        ; 4:      Rx process pointer 0=RTCC proc 01
                                        ;                                1=RTCC proc 11
                                        ; 5-7:    Sample rate count down
                                        ;
FSRMain   equ       13h                 ; Save for FSR during IntHandler processes
FSRInt    equ       14h                 ; Save for FSR during Main process
CRCs      equ       15h                 ; CR-Check during sending
CRCr      equ       16h                 ; CR-Check during receiving
                                        ;
                                        ; # regs:
                                        ; ---------
First     equ       17h                 ; First file reg

W1        equ       17h                 ; temporary work register

                                        ; ......

Last      equ       2Fh                 ; Last file reg
;
;*******************************************************************
;
;Vectors (PIC16C71)
;

          ORG       PIC71
_ResetV   goto      _StartUp            ; Reset startup routine

                                        ; PIC 71: new defined vectors:
                                        ; ----------------------------
_WdtSlpV  goto      _WdtSlp             ; # Reenter after WDT wake-up from SLEEP
_WdtToV   goto      _WdtTo              ; # Reenter after WDT time-out
_MClrSlpV goto      _MClrSlp            ; # Reenter after /MCLR wake-up from Sleep

                                        ; (HINT: user must firstly clear flags with
                                        ; "clrwdt" and set GIE!!)

_IntV     goto      _IntHndlr           ; Interrupt handler (Reenter after INT
                                        ;   wake-up from Sleep)

;
;*******************************************************************
;
;Version marks (still visible when compiled/dumped)
;
_Vers     data      "MpK V1.0/#/"       ; Mark "MpK-Ver/User-Ver/...

          data      "(c)NTB,CH-Buchs/#/"

          IF        __MHz=20            ; ...xxMHz/..
          data      "20"
          ELSE
           IF       __MHz=16
           data     "16"
           ELSE
            IF      __MHz=12
            data    "12"
            ELSE
            data    "??"                ; # (other rates)
            ENDIF
           ENDIF
          ENDIF

          data      "MHz/RS"
          IF        __RS=232            ; ..RS.../..
          data      "232"
          ELSE
          data      "485"
          ENDIF
          IF        __Addr=Yes
          data      "Node: #"           ; # (Rx_Addr)
          ENDIF

          data      "/nP/"              ; no parity !

          data      "9.6kBd"            ; # (user baud rate)

_Flags    IF        __Tst=Yes           ; Mark used Flags as comment
           data     "Tst/"
          ENDIF
          IF        __RxInv=Yes
           data     "RxINV/"
          ENDIF
          IF        __TxInv=Yes
           data     "TxINV/"
          ENDIF
          IF        __ADInv=Yes
           data     "ADINV/"
          ENDIF
          IF        __WDT=Yes
           data     "WDT"
          ENDIF

;
;*******************************************************************
;
; Subroutines which must be at PCH=0 for ind gotos and calls
; ----------------------------------------------------------
;
;         ORG       40h                 ; If you want to be aware of total code securety
                                        ; (-> check that location)

                                        ; #
          return
;
;*******************************************************************
;
; Start up after any reset or power up...
;
                                        ; Detects Reset conditions
                                        ; and reenters accordingly
                                        ; ------------------------
_StartUp  btfsc     STATUS,PD           ; IF sleep detected THEN
          goto      nosleep
          btfss     STATUS,TO           ;  IF WDT time-out THEN
          goto      _WdtSlpV            ;   Jump WDT-Sleep vector
          goto      _MClrSlpV           ;  ELSE jump /MCLR-Sleep vector
                                        ;  ENDIF
nosleep   btfss     STATUS,TO           ; ELSIF WDT time-out THEN
          goto      _WdtToV             ;  Jump WDT time-out vector
                                        ; ENDIF

          IF        __Tst
          movlw     Tx_Rst              ; Send Tx_Rst code at Reset
          movwf     TxData
          ENDIF

          goto      _PwrOnRst

;
;*******************************************************************
;
; Reentrance (dummy) for reset conditions
;

_WdtSlp   IF        __Tst
          movlw     Tx_WDTSLP           ; Send Tx_WDTSLP at Reset
          movwf     TxData
          ENDIF

                                        ; #

          goto      _Dummy

; -----------------------------------------------------------------------

_WdtTo    IF        __Tst
          movlw     Tx_WDTTO            ; Send Tx_WDTTO at Reset
          movwf     TxData
          ENDIF

                                        ; #

          goto      _Dummy

; -----------------------------------------------------------------------

_MClrSlp  IF        __Tst
          movlw     Tx_MCLRS            ; Send Tx_MCLRS at Reset
          movwf     TxData
          ENDIF

                                        ; #

          goto      _Dummy


;
;***************************************************************
;
; PowerOn- or Reset-Condition
; ---------------------------
; Put system into defined condition
; Hardware initialization:
;

_Dummy    IF        __WDT=Yes           ; Entry for other unused power up cond.
          clrwdt
          ENDIF
                                        ; Ordinary entry...
;
; PinHardware initialization:
;
                                        ; TRISA Register: #
_PwrOnRst movlw     b'11111'            ; Out:    -
          tris      TRISA               ; In:     0-3:      AD-Inputs,
                                        ;         4:        digital input pin
          movlw     b'11111'            ; Reset Port A: Preset as "1"
          movwf     PORTA
          movf      PORTA,W
                                        ; TRISB Register: (#)
          movlw     b'00000001'         ; Out:    7,6,5,4,3:unused
                                        ;         2:        TxEn or unused
                                        ;         1:        TxPin
          tris      TRISB               ; In:     0:        RxPin
          movlw     b'00000011'         ; Reset Port B: 7-3=0,TxEn=0,Tx/RxPin=1
          movwf     PORTB
          movf      PORTB,W

          movlw     b'10001000'         ; OPTION Register init: (#)
          OPTION                        ; RBPU=dis, INTEDG=falling, RTS=Intern,
                                        ; RTE=X, PSA=WDT,(Presc=000)

          movlw     b'11000001'         ; ADCON0 Register: #
          movwf     ADCON0              ; ADCS=11,x,CHS=00,GO=0,ADIF=0,ADON=1


          movlw     b'11'               ; ADCON1 Register: #
          Page1                         ; PCFG=11:RA0-3=digital in
          movwf     ADCON1
          Page0

;
; Init system regs.:
;

          movlw     SRate+b'0011'
          movwf     PStat               ; Bit 7-4:          Countdown sample rate
                                        ; Bit 3,2:          TxFlag, RxFlag,
                                        ; Bit 1,0:          Proc 11 (->Tx 00)

          movlw     b'00001000'
          movwf     TxRxFlg             ; Init Tx / Rx flags
                                        ; Bit 7-5:Tx bit counter=000
                                        ; Bit 4:  Flag TxStop
                                        ; Bit 3:  Flag TxStart  =1
                                        ; Bit 2:  Flag RxSel proc 01,11 select
                                        ; Bit 1:  Flag RxStop
                                        ; Bit 0:  Flag Rxtart

;
;
; RTCC and prescaler initialization
;

          clrwdt                        ; Set defined conditions to flags, WDT
                                        ; and the prescaler
          Page1
          movf      OPT,W               ; Set prescaler:
          Page0
          iorlw     Presc               ; Internal,low-high,RTCC,'constant'
          OPTION                        ; (OPT was set to xx000000b within _InitHW)
                                        ; First time RTCC runs up from 00h to 0FFh

;
;
; Enable and disable interrupts:
;
          movlw     b'11110000'         ; # En-/disables Interrupts:
          movwf     INTCON              ;   'GIE,ADIE,RTIE,INTE,RBIE,rtcc,int,rb'

          goto      _Main               ; Goto Main Programm


;
;*******************************************************************
;
; Main interrupt handler which branches depending on interrupt source
; -------------------------------------------------------------------
; All interrupts are branching to this routine. It firstly saves the
; work- and the statusfile, than detects the occured interrupt and
; jumps to the according subroutine.
; The way the user queues the calls that makes the priority. Last has
; most priority, if one after the other is tested ("_RetIntX").
; If the user wants equal priority, just reenter always at "_RetInt"!
; To maintain 100% STATUSreg recovery, the zeroflag must be testet
; and re-set after reloading the workfile!
;

                                        ; Interrupt Handler (GIE=0)
                                        ; -----------------
                                        ; higher priority interrupts which occur
                                        ; during handler are executed if detected
                                        ; after RetIntX of the previous interrupt!

_IntHndlr movwf     SaveW               ; Save W and STATUS
          movf      STATUS,W
          movwf     SaveST

          movf      FSR,W               ; Exchange FSRs
          movwf     FSRMain             ; (# implement only if used)
          movf      FSRInt,W
          movwf     FSR

                                        ; Interrupt branch depending on source
                                        ; and reenter-test:
                                        ; ------------------------------------
                                        ; Time critical design (used here to get high baud rates):
                                        ;   If your application has a short timebase (often interrupts
                                        ;   to this handler by the RTCC) then use this way of programming.
                                        ;   Every interrupt is checked after the other (also by jumping
                                        ;   to the next _RetIntX after a processed handler). So if the first
                                        ;   check is passed any further interrupt on that one will be ignored
                                        ;   until next RTCC-interrupt comes up!
                                        ;   (Hint: place most important proc last to be sure to get it!)
                                        ;
                                        ; Most relyable design (not used here):
                                        ;   If your application has a long timebase then it's better to change
                                        ;   the following design.
                                        ;   After each processed interrupt-handler the return address must be
                                        ;   _RetInt again. So if any interrupt will come up during handler
                                        ;   processing it also will be processed until no 9interrupt flag
                                        ;   is set anymore.
                                        ;   (Hint: place most important proc first -> priority!)
                                        ;
                                        ;
_RetInt   btfsc     ADCON0,ADIF         ; IF AD completion THEN
          goto      _IntAD              ;  goto AD-Handler

_RetInt1
          btfsc     INTCON,RBIF         ; ELSEIF Port B change THEN
          goto      _IntPORTRB          ;  goto RB-Handler

_RetInt2  btfsc     INTCON,INTF         ; ELSEIF INT THEN
          goto      _INT                ;  goto INT-Handler

_RetInt3  btfsc     INTCON,RTIF         ; IF RTCC overflow THEN
          goto      _IntRTCC            ;  goto RTCC-Handler
                                        ; ENDIF

_RetInt4  movf      FSR,W               ; Exchange FSR
          movwf     FSRInt              ; (# only if used)
          movf      FSRMain,W
          movwf     FSR

          movf      SaveST,W            ; Reload old STATUS and
          movwf     STATUS
          movf      SaveW, W            ; Z-Flag changed!!

                                        ; Correct the Z-flag
          btfsc     SaveST,Z            ; Test old Z
          setz                          ; and re-set it...
          btfss     SaveST,Z
          clrz
EOI       retfie                        ; Enable global interrupt and
                                        ; return to processes

;
;*******************************************************************
;
; Interrupt on completion of A/D-conversion (wake-up)
; ---------------------------------------------------
; #

_IntAD    bcf       ADCON0,ADIF         ; Reset flag
          btfss     INTCON,ADIE         ; IF AD is disabled THEN
          goto      _RetInt1            ;   return
                                        ;   (Alternativly use _RetInt !)
                                        ; ENDIFF

          IF        __ADInv=Yes
InvRes     movlw    0FFh                ; Invert ADRES
           xorwf    ADRES,F             ;
          ENDIF
                                        ; #...

          goto      _RetInt1            ; End of AD interrupt
                                        ; # Goto next interrupt test...
                                        ; (Alternativly use _RetInt !)

;
;*******************************************************************
;
; Interrupt on changes on port RB4-7 (wake-up)
; --------------------------------------------
; #

_IntPORTRB movf     PORTB,W             ; Read PORTB
          bcf       INTCON,RBIF         ; Reset flag
          btfss     INTCON,RBIE         ; IF RB is disabled THEN
          goto      _RetInt2            ;   return
                                        ;   (Alternativly use _RetInt !)
                                        ; ENDIF

                                        ; #

          goto      _RetInt2            ; End of Port RB interrupt
                                        ; # Goto next interrupt test...
                                        ; (Alternativly use _RetInt !)

;
;*******************************************************************
;
; Extern INTerrupt on RB0 pin input (wake-up)
; -------------------------------------------
; Is used to determine RxStart. A falling edge activates the interrupt.
; The sub disables itself during next 8 bits until data is received.
; The RxProcess is set free by TxFlag and TxStart. To catch the input
; level at the right time next RxProcess time slice is counted and set
; to flag RxSel.
;

_INT      bcf       INTCON,INTF         ; Reset flag
          btfss     INTCON,INTE         ; IF INT is disabled THEN
          goto      _RetInt3            ;   return
                                        ;   (Alternativly use _RetInt !)
                                        ; ENDIF

          bcf       INTCON,INTE         ; Disable next interrupts until _RxStop

          bsf       PStat,RxFlag        ; Set RxFlag
          bsf       TxRxFlg,RxStart     ; Set pointer to start bit

          incf      PStat,F             ; Because _INT is before _IntRTCC
          btfsc     PStat,PCnt1         ; If Proc00/01 is active then set RxSel=0
          bsf       TxRxFlg,RxSel
          btfss     PStat,PCnt1         ; else set RxSel=1
          bcf       TxRxFlg,RxSel
          decf      PStat,F

          goto      _RetInt3            ; End of INT interrupt
                                        ; # Goto next interrupt test...
                                        ; (Alternativly use _RetInt !)

;
;*******************************************************************
;
; Interrupt on Real Time Clock overflow:
; --------------------------------------
; Every (#) Xus the RTC has its overflow to detemine the right time intervall
; for Tx- and RxProcesses. The user must consider of having enough spare
; cycle to avoid the RTC overflowing twice in case of several interrupt
; sources occur!
; Before jumping to the 4 intervall processes a separate AD-Start sub is
; testet for its call. This allows to start the AD-conversion at every
; 1*Xus up to 16*Xus (sample rate)!
; If a uncommon X-tal is used, an adjust process may be provided to inc or dec
; some RTC-steps to adjust the roundabout time of the 4 processes!
; Otherwise this process may be used by the user. #
;

_IntRTCC  bcf       INTCON,RTIF         ; Reset flag
;         btfss     INTCON,RTIE         ; IF RTCC is disabled THEN
;         goto      _RetInt4            ;   return
                                        ;   (Alternativly use _RetInt !)
                                        ; ENDIF

          movlw     100h-Rtcc           ; Set RTCC start value from constant (Xus)
          addwf     RTCC,f              ; #

          skpc                          ; Test if there was a cycle lost?
          goto      NoLostC
                                        ; LostCycle!!!
          IF        __Tst=No            ; ------------
          movlw     0ffh                ; # One way to solve this: Set it near before INT
          movwf     RTCC                ; Other ways are to set an error or stop operation:
          ELSE
          halt			; During test phase stop here !
          ENDIF

NoLostC   movlw     11h                 ; Countdown of sample rate and
          addwf     PStat,F             ; set process counter + 1 (overflow bit 2?)
          skpnc
          call      _SubADProc          ; Subroutine Proc AD: Start conversion

                                        ; Check processes 00 to 11
                                        ; ------------------------
          btfsc     PStat,PCnt0         ; Test
          goto      _RxProc             ; Proc 01 or 11: Rx depending on RxSel
          btfsc     PStat,PCnt1

          goto      _AdjProc            ; Proc 10: # (Adjust)
                                        ;          Or just _RetInt4 if not used!
          movlw     4h                  ; fix overflow bit 2
          subwf     PStat,F             ;
          goto      _TxProc             ; Proc 00: Tx

;
;*******************************************************************
;
; AD process (sub):
; -----------------
; Starts AD-conversion and resets sample rate countdown
;

_SubADProc                              ; #
          bsf       ADCON0,GO           ; Start conversion
          movlw     SRate               ; Re-set countdown value
          addwf     PStat,F
                                        ; #

          return                        ; ...from sub

;
;*******************************************************************
;
; Tx process 00:If TxFlag is set data TxData will be sent via pin RB1.
;                   According to TxRxFlgs:
;                   TxStart:start(low)
;                   TxData: (Count at TxRxFlg)
;                   1xH:      bit0
;                   3xH:      bit1
;                   5xH:      bit2
;                   7xH:      bit3
;                   9xH:      bit4
;                   BxH:      bit5
;                   DxH:      bit6
;                   FxH:      bit7 -> Carry sets TxStop
;                   TxStop:   stop(high),reset TxFlag to mark end of Tx.
;                                          and sets TxStart for next transmission
; To set a new TxData check if TxFlag=0!
; Due to limited file space there is not parity check!!
;

                                        ; Transmitter every 4*Xus
_TxProc   btfsc     PStat,TxFlag        ; Is transmitter active?
          goto      _TxOK               ;   yes: send next data

          IF        __RS=485            ; If half duplex:
          IF        __TxEn=Yes
          bcf       PORTB,TxEn          ;   no:  Disable output
          ENDIF
          ENDIF                         ;        and
          goto      _RetInt4            ;        return
                                        ;        (Alternativly use _RetInt !)

_TxOK     btfsc     TxRxFlg,TxStart     ; Is start bit to be sent?
          goto      _TxStart            ; Yes
          btfss     TxRxFlg,TxStop      ; Is data to be sent?
          goto      _TxData             ; Yes

_TxStop   IF        __TxInv=Yes
           bcf      PORTB,TxPin         ; Set output high (STOP)
          ELSE
           bsf      PORTB,TxPin         ; Set output high (STOP)
          ENDIF
          bcf       TxRxFlg,TxStop      ; Clear TxStop
          bcf       PStat,TxFlag        ; Clear TxFlag
          bsf       TxRxFlg,TxStart     ; Set TxStart for next transfer
          IF        __RS=485
          bsf       INTCON,INTE         ; Enable Rx Int-process
          ENDIF
          goto      _RetInt4            ; (Alternativly use _RetInt !)

                                        ; Send START
_TxStart  IF        __RS=485
          btfsc     PStat,RxFlag        ; Is Rx receiving ?
          goto      _RetInt4            ;   Yes: Wait for its end OR set Error???

          bcf       INTCON,INTE         ; Disable Rx Int-process
          IF	__TxEn=Yes
          bsf       PORTB,TxEn          ; Enable output
          ENDIF
          ENDIF

          IF        __TxInv=Yes
           bsf      PORTB,TxPin         ; Set output low (START)
          ELSE
           bcf      PORTB,TxPin         ; Set output low (START)
          ENDIF
          bcf       TxRxFlg,TxStart     ; Start is sent
          goto      _RetInt4            ; (Alternativly use _RetInt !)

                                        ; Send DATA
_TxData   IF        __Tst=Yes
          clrc                          ; Set rotatebits to zero
          ENDIF
          rrf       TxData,F            ; bit to carry
          IF        __TxInv=Yes
           skpc
           bsf      PORTB,TxPin         ; Set output according to carry
           skpnc
           bcf      PORTB,TxPin         ;
          ELSE
           skpnc
           bsf      PORTB,TxPin         ; Set output according to carry
           skpc
           bcf      PORTB,TxPin         ;
          ENDIF
          movlw     20h                 ; Increment counter...
          addwf     TxRxFlg,F           ; count up to bit 8...
          skpnc                         ; is last bit sent?
          bsf       TxRxFlg,TxStop      ; Yes->Set Stop flag
          goto      _RetInt4            ; (Alternativly use _RetInt !)

;
;*******************************************************************
;
; Rx process 01,11: If RxFlag is ser and the correct process according
; to RxSel is running, data RxData will be sent via pin RB1:
;                   According to TxRxFlags:
;                   RxStart:start(low), set first bit of RxData high
;                   RxData: bit0..7, shift RxData till first bit is high
;                                        ->set RxStop
;                   RxStop:   stop(high),reset RxFlag, enable INTerrupt,
; # User must clear flag DataRdy after datafetch at file ADCON0
; to mark data is read!
;

                                        ; Receiver process every (other) 4*Xus
_RxProc   btfss     PStat,RxFlag        ; Is receiver active?
          goto      _RetInt4            ; no:return
                                        ; (Alternativly use _RetInt !)
          btfsc     PStat,PCnt1         ; Is it the selected timeslice?
          goto      _Rx11               ;
_Rx01     btfss     TxRxFlg,RxSel       ; PCnt=0:Sel=?
          goto      _RxOK               ; Sel=0   =>ok, go ahead
          goto      _RetInt4            ; Sel=1 =>return
                                        ; (Alternativly use _RetInt !)
_Rx11     btfss     TxRxFlg,RxSel       ; PCnt1=1:Sel=?
          goto      _RetInt4            ; Sel=0 =>return
                                        ; (Alternativly use _RetInt !)
_RxOK     btfsc     TxRxFlg,RxStart     ; Is start bit first
          goto      _RxStart            ; Yes
          btfss     TxRxFlg,RxStop      ; Is stop bit next?
          goto      _RxData             ; No: ->Data

_RxStop   IF        __RxInv=Yes
           btfsc    PORTB,RxPin         ; Is signal high?
          ELSE
           btfss    PORTB,RxPin         ; Is signal high?
          ENDIF
          goto      _RxError            ; Low: error
                                        ; High:ok...
_RxStopE  bsf       INTCON,INTE         ; Enable INT interrupt for new data
          bsf       ADCON0,DataRdy      ; Set data ready flag
          IF        __TxEn=Yes          ; If TxEn used:
          bcf       PORTB,TxEn          ; Disable sending (line must remain on "1" !!)
          ENDIF
          bcf       PStat,RxFlag        ; Clear RxFlag
          bcf       TxRxFlg,RxStop      ; Clear RxStop
          goto      _RetInt4            ; (Alternativly use _RetInt !)

                                        ; An error occured:
_RxError  bsf       SaveST,RxError      ; Mark Rx Error (Reset in Main after sent)
          bsf       Cmd,Cmd_NACK        ; Mark into Cmd-Word NACK
          goto      _RxStopE            ; # Re-enable RxProcess after sending NACK!!


_RxStart  bcf       TxRxFlg,RxStart     ; Start bit passed
                                        ; ===============================
                                        ; # Execute "bcf ADCON0,DataRdy"
                                        ;   after data is read and before
                                        ;   new databits are comming in!!
                                        ; ===============================
          btfsc     ADCON0,DataRdy      ; Test for old data is read?
          goto      _RxBusy             ; Too slow.... >>> error!!!
          IF        __RxInv=Yes
           btfsc    PORTB,RxPin         ; Is signal low?
          ELSE
           btfsc    PORTB,RxPin         ; Is signal low?
          ENDIF
          goto      _RxError            ; High: error
          movlw     80H
          movwf     RxData              ; set rotate flag
          IF        __TxEn=Yes          ; If TxEn used:
          bsf       PORTB,TxEn          ; Enable sending
          ENDIF
          goto      _RetInt4            ; (Alternativly use _RetInt !)

                                        ; An error occured:
_RxBusy   bsf       Cmd,Cmd_NACK        ; Mark into Cmd-Word NACK
          bsf       SaveST,RxBusy       ; Mark Rx Busy (Reset in Main after sent)
          goto      _RxStopE            ; # Re-enable RxProcess after sending NACK!!


_RxData   btfsc     PORTB,RxPin         ; Receive data
          IF        __RxInv=Yes
           clrc
          ELSE
           setc
          ENDIF
          btfss     PORTB,RxPin
          IF        __RxInv=Yes
           setc
          ELSE
           clrc
          ENDIF
          rrf       RxData,F            ; Carry to file
          skpnc
          bsf       TxRxFlg,RxStop      ; Last bit is received -> get stop bit now
          goto      _RetInt4            ; (Alternativly use _RetInt !)

;
;*******************************************************************
;
; Adjust process 10:
; ------------------
; This process may be used for adjustment of the roundabout time of
; the 4 Process (and the sample rate). By inc- or decrementing the RTCC
; any adjustment may be done!
; # If not used this is a free user definable process:
;

_AdjProc  movlw     Adj                 ; Adjust RTCC to the av. of Xus/process
          addwf     RTCC,F
                                        ; OR #

          goto      _RetInt4            ; (Alternativly use _RetInt !)

;
;*******************************************************************
;
; Main Processes
; --------------
; MpK Rem.:         7 stack levels are free.
;                   (Z-Flag may not be used without disabling GIE or
;                    adding extend code when leaving IntHandler!)

; Zwo pseude processes are conncted and runned by a roundabout angle
; point: SUB SubInputExe. during the wait of cmd-process for cmd input,
; cmd flags and cmd output are executed via output sub processes.


_Main     IF        __Tst=Yes
           IF       __RS=232            ; (RS485: # depends on net architecture!)
                                        ; Send Reset Info if WDT is used
           bsf      PStat,TxFlag        ; Hint: By the means of this value a conditional
                                        ; clearing part (following next) may be bypassed!!
           ENDIF
          ENDIF

                                        ; Clear/define used Variables
                                        ; ===========================
Clear     clrf      Cmd                 ; Rem.: TxData, RxData are undef...
;         clrf      Ext                 ; If there is any Extended command file reg

          clrf      CRCr                ; Clear CRCreceive and -send
          clrf      CRCs                ; (# only if used)

;         movlw     BufMain             ; # Load Main Ind Pointer to any buffer
          movwf     FSR
                                        ; Optional usage of Main and Int FSR:
;         movlw     BufInt              ; # Load Int Ind Pointer to any buffer
          movwf     FSRInt

                                        ; #

EndClear                                ; (Use for bypassing)

;--------------------------------------------------------------------
;
; Cmd-Process
; -----------
; Receives Cmd via Rx-Line, checks Header (0FFH), Address
; and saves Cmd. If there is an immed. Cmd-action to do at a specific
; cmd, it will be performed immediatly.

GetHdr    nop                           ; Wait for Header
          IF        __RS=485            ; ===============
          bsf       INTCON,GIE          ; Enable Interrupts (used if addresse was wrong)
          ENDIF

RetNACK   call      SubInpExe           ; Wait for Cmd_Hdr

          movlw     Rx_Hdr
          subwf     RxData,W            ; IF Rx_Hdr->FFh-FFh=00h(C) ....

          bcf       ADCON0,DataRdy      ; ( Free RxProcess

          IF        __RS=232            ;   Point to Point:
          btfsc     Cmd,Cmd_NACK        ;   Is there a NACK?
          goto      BrnchNACK           ;   Send NACK and Restart CmdExecution

          ELSE                          ;   Bus:
          btfss     Cmd,Cmd_NACK        ;   Is there a NACK?
          goto      HdrOK               ;   No...continue
          btfsc     Cmd,Cmd_Exe         ;   Is still a command running?
          goto      BrnchNACK           ;     Yes: Send NACK
          ENDIF                         ; )

HdrOK     skpc                          ; THEN get next data ...
          goto      GetHdr              ; ELSE...wait for new header (sync)
          clrf      CRCr                ; Reset CRC

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
          IF        __Addr=Yes          ; Bus supports addresses:
                                        ; Wait for Address
                                        ; ================
GetAdr    call      SubInpExe           ;   Wait for address data

          movlw     ~Rx_LAll            ;   Test for listen all address 00h (~00h => FFh)
Invers    addwf     RxData,W            ;   IF LAll (->00h+FFh=FFh,NC) THEN
          bcf       ADCON0,DataRdy      ;   (Free RxProcess
          movf      RxData,W            ;   Save address W := RxData)
          skpc                          ;
          goto      ListenAll           ;     BRANCH (reset DataRdy there!)
                                        ;   ENDIF

TestAdr   addwf     CRCr,F              ;   Add CRC
          sublw     Rx_Addr             ;   IF RxData(W)<>Rx_Addr THEN
          skpz
          goto      GetHdr              ;     Wait for new header (sync)
                                        ;   ENDIF
          btfss     Cmd,Cmd_NACK        ;   Is there a NACK?
          goto      GetCmd              ;    No: continue
          btfsc     Cmd,Cmd_Exe         ;    Yes: And is still a command running?
          goto      BrnchNACK           ;         Yes: Send NACK

          ENDIF
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

                                        ; Wait for Cmd
                                        ; ============
GetCmd    clrf      Cmd                 ; Clear Cmd for new Command. This stops ExeSubs!
          bcf       STATUS,RxError      ; Reset Rx Errors..
          bcf       STATUS,RxBusy

          call      SubInpExe           ; Wait for Cmd

          movfw     RxData              ; Get RxData into W
          bcf       ADCON0,DataRdy      ; Free RxProcess
          btfsc     Cmd,Cmd_NACK        ; Is there a NACK?
          goto      BrnchNACK           ; Restart CmdExecution

                                        ; (Default is '001x xxxx'!!!)
                                        ;              =========
          addwf     CRCr,F              ; Add CRC
          movwf     Cmd                 ; Save Cmd
          btfsc     Cmd,07h             ; Test for Syntax Error
          goto      SyntaxErr
          btfsc     Cmd,06h             ; Test for Syntax Error
          goto      SyntaxErr
          btfss     Cmd,05h             ; Test for Syntax Error
          goto      SyntaxErr
                                        ; No SyntaxError:
          movlw     b'00011111'         ; Clear first 3 bits anyway...
          andwf     Cmd,F               ; ... of Cmd
          goto      TstActCmd           ; Test for action command now!

RetTAC    btfsc     Cmd,Cmd_NACK        ; Test for Syntax Error
          goto      BrnchNACK
          goto      GetHdr              ; Endless loop


;
;--------------------------------------------------------------------
;
; Syntax Error occured:
; ---------------------
; Only the NACK-Flag is set!
;

SyntaxErr bsf       Cmd,Cmd_NACK        ; Imaginary set NACK for Syntax error
          bcf       STATUS,RxError      ; Reset Rx Errors..
          bcf       STATUS,RxBusy
;         goto      BrnchNACK           ; (no other flag is set)


;
;--------------------------------------------------------------------
;
; Error branch if Cmd_NACK is set:
;
; If Cmd_NACK is set this routine should be accessed.
; Tree errors may be sent back to host:
;         if RxError is set, start/stop bit failure occured
;         if RxBusy is set, Tx did send to fast
;         else a Syntax error were found
; The according Error-Data will be sent back to host and
; Cmd-Process restarts at GetHdr...



BrnchNACK bcf       Cmd,Cmd_NACK        ; Clear NACK Flag

          IF        __Tail=Yes          ; Wait for Cmd-Protocol to be finish
WaitTail  call      SubInpExe           ; Wait for Cmd-tail

          movlw     ~Rx_Tail+1          ; Test for Cmd-tail (FEh)
          addwf     RxData,W            ; IF tail->FEh+2=00h->C ....
          movf      RxData,W
          bcf       ADCON0,DataRdy      ; Free RxProcess
          skpnc                         ; Was it Rx-Tail?
          goto      TestErr             ;   Yes: send Errorcode

          movlw     Rx_Hdr
          subwf     RxData,W            ; IF Rx_Hdr->FFh-FFh=00h->C ....
          skpc                          ;   Yes: Leave Error, get Adress
          goto      WaitTail            ;   No:  Wait for next data

          movlw     00h                 ; Leave Error, read new Adress!
          movwf     Cmd
          bcf       STATUS,RxError
          bcf       STATUS,RxBusy
          IF	__Addr=Yes
          goto      GetAdr              ; Hdr OK -> wait for address
          ELSE
          goto      GetCmd              ; Hdr OK -> wait for Cmd
          ENDIF
          ENDIF

TestErr   btfsc     STATUS,RxError      ; Did RxError occur ?
          goto      TxRxError           ;  Yes
          btfsc     STATUS,RxBusy       ; Did Rxbusy occur ?
          goto      TxRxBusy            ;  Yes

                                        ; #

          goto      TxSxErr             ; Else Syntax Error... !

                                        ; #

                                        ; MPK:
TxRxError movlw     Tx_RxErr            ; Mark for RxError
          bcf       STATUS,RxError      ; Reset Flag
          goto      RxEn                ; Free Rx Process again

TxRxBusy  movlw     Tx_RxBusy           ; Mark for RxBusy
          bcf       STATUS,RxBusy       ; Reset Flag

RxEn      bsf       INTCON,INTE         ; Enable INT interrupt for new data
          goto      SendErr


TxSxErr   movlw     Tx_SErr             ; Mark for syntax error

SendErr   call      SubSendW
          movlw     00h                 ; Resets Exe-, ACK-(, NACK-) and Cmd-Flags
          movwf     Cmd
          goto      RetNACK             ; Restarts Cmd receiving process!


;
;****************************************************************************
;
; Action Command routines
; -----------------------
; If a Cmd as an action part(*), it will no be performed once. After that
; the Cmd-Process waits for new Cmd.
;
; (*) The action part may include additional Data fetching, action execution
; and/or first CmdSub call.


TstActCmd btfsc     Cmd,Cmd_1xxx        ; # Cmd 001T 1xxx ?
          goto      Cmd1xxx             ;  Yes

          btfsc     Cmd,Cmd_01xx        ; Cmd 001T 01xx ?
          goto      Cmd01xx             ;  Yes

          btfsc     Cmd,Cmd_001x        ; Cmd 001T 001x ?
          goto      Cmd001x             ;  Yes

          btfsc     Cmd,Cmd_0001        ; Cmd 001T 0001 ?
          goto      Cmd0001             ;  Yes

                                        ; ELSE:

          goto      ExtCmd              ; Cmd 001T 0000 -> Extended commands
                                        ; # or (no extended commands)
;         bsf       Cmd,Cmd_NACK        ; Imaginary set NACK for Syntax error
;         bcf       STATUS,RxError      ; Reset Rx Errors..
;         bcf       STATUS,RxBusy
;         goto      RetTAC


;
;--------------------------------------------------------------------
;
; Get data and execute immed. action commands
;
;
;

                                        ; Cmd immed. action
Cmd1xxx                                 ; =================

                                        ; #

                                        ; (If there is no cmd action (eg. data fetch)
                                        ; following the cmd, just add nop at this part.)

                                        ; Usage of CRCr:
                                        ; ==============
                                        ; ...
GetAData  call      SubInpExe           ; read next data
          movf      RxData,W            ; save data to W
          bcf       ADCON0,DataRdy      ; Free RxProcess
          btfsc     Cmd,Cmd_NACK        ; Is there a NACK?
          goto      RetTAC              ; Restart CmdExecution
          addwf     CRCr,F              ; Add CRC
          movwf     W1                  ; eg. ...

CRCtest   call      SubGetCRCr          ; Get and test CRC after all data (and before Cmd-tail)
          skpnc                         ; C is set if CRCr not equal received data!
          goto      SErr                ;  Set Cmd_Exe,Cmd_ACK flags and goto GetHdr

          goto      SetEXEACK           ;  Set Cmd_Exe,Cmd_ACK flags and goto GetHdr

;
;--------------------------------------------------------------------
;
; Get data and execute immed. action commands
;
;
;

                                        ; Cmd immed. action
Cmd01xx                                 ; =================

                                        ; #

          goto      SetEXEACK           ;  Set Cmd_Exe,Cmd_ACK flags and goto GetHdr

;
;--------------------------------------------------------------------
;
; Get data and execute immed. action commands
;
;
;

                                        ; Cmd immed. action
Cmd001x                                 ; =================

                                        ; #

          goto      SetEXEACK           ;  Set Cmd_Exe,Cmd_ACK flags and goto GetHdr

;
;--------------------------------------------------------------------
;
; Get data and execute immed. action commands
;
;
;

                                        ; Cmd immed. action
Cmd0001                                 ; =================

                                        ; #

          goto      SetEXEACK           ;  Set Cmd_Exe,Cmd_ACK flags and goto GetHdr



;---------------------------------------------------------------------------------------

                                        ; Extended commands
                                        ; =================
ExtCmd    call      SubInpExe           ; Wait for next data
          movf      RxData,W            ; Get Data into W
          bcf       ADCON0,DataRdy      ; Free RxProcess
          btfsc     Cmd,Cmd_NACK        ; Is there a NACK?
          goto      RetTAC              ; Restart CmdExecution

          movwf     W1                  ; Save Ext into W1
                                        ; (If there is enough file regs available
                                        ;  a special file named Ext may be used
                                        ;  instead)

                                        ; Test Ext action
TstActExt                               ; ---------------
          btfsc     W1,Ext_XXh          ; # Bit_ ? (8 times)
          goto      ExtXXh              ;  Yes

                                        ; # (continue like above, but 8 times...)

SErr      bsf       Cmd,Cmd_NACK        ; Imaginary set NACK for Syntax error
          bcf       STATUS,RxError      ; Reset Rx Errors..
          bcf       STATUS,RxBusy
          goto      RetTAC

;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

                                        ; Get Error
                                        ; =========

ExtXXh                                  ; # immed. action

          goto      SetEXEACK           ; Set Cmd_Exe,Cmd_ACK flags and goto GetHdr


;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

SetEXEACK nop                           ; Finish command action with ACK and Exe flag set

          IF        __Tail=Yes          ; Wait for Cmd-Protocol to be finish
          call      SubInpExe           ; Wait for Cmd-tail

          movlw     ~Rx_Tail+1          ; Test for Cmd tail (FEh)
          addwf     RxData,W            ; IF tail->FEh+2=00h->C ....
          movf      RxData,W
          bcf       ADCON0,DataRdy      ; Free RxProcess
          skpnc                         ; Was it Rx-Tail?
          goto      SetA_               ;   Yes: set ACK

          movlw     Rx_Hdr
          subwf     RxData,W            ; IF Rx_Hdr->FFh-FFh=00h->C ....
          skpc                          ;   Yes: Leave Error, get address
          goto      RetTAC              ;   No:  Return to Hdr

          clrf      Cmd                 ; Leave Error, read new address!
          IF	__Addr=Yes
          goto      GetAdr              ; Hdr OK -> wait for address
          ELSE
          goto      GetCmd              ; Hdr OK -> wait for Cmd
          ENDIF
          ENDIF

SetA_     bsf       Cmd,Cmd_Exe         ; Set Exe Flag
          bsf       Cmd,Cmd_ACK         ; Set ACK Flag
          goto      RetTAC


;
;
;
;
;
                                        ; Only if addressable:
          IF        __Addr=Yes          ; BEGIN of "Listen All"
;
;****************************************************************************
;
; Listen All BRANCH;
;
;
                                        ; Wait for LACmd
                                        ; ==============
ListenAll movlw     00h                 ; Clear Cmd for new Command. This stops ExeSubs!
          movwf     Cmd

          call      SubInpExe           ; Wait for LACmd

          movf      RxData,W            ; Get RxData
          bcf       ADCON0,DataRdy      ; Free RxProcess
          btfsc     Cmd,Cmd_NACK        ; Is there a NACK?
          bcf       Cmd,Cmd_NACK        ;   Clear flag (Don't send any NACK!...)

                                        ; (Default is '001x xxxx')
          movwf     Cmd                 ; Save Cmd
          btfsc     Cmd,07h             ; Test for Syntax Error
          goto      RetTALAC
          btfsc     Cmd,06h             ; Test for Syntax Error
          goto      RetTALAC
          btfss     Cmd,05h             ; Test for Syntax Error
          goto      RetTALAC
                                        ; No SyntaxError:
          movlw     b'00011111'         ; Clear first 3 bits anyway...
          andwf     Cmd,F               ; ... of Cmd
          goto      TstActLAC           ; Test for action command now!

RetTALAC  goto      GetHdr              ; Error return with no mark of SyntaxError

;
;----------------------------------------------------------------------------
;
; SUB: Action Command routines
; ----------------------------
; If a Cmd as an action part(*), it will no be performed once. After that
; the Cmd-Process waits for new Cmd.
;
; (*) The action part may include additional Data fetching, action execution
; and/or first CmdSub call.


TstActLAC btfsc     Cmd,Cmd_1xxx        ; # Cmd 001Txxxx ?
          goto      LAOut1xxx           ;  Yes

                                        ; #...

          bsf       Cmd,Cmd_NACK        ; Imaginary set NACK for Syntax error
          goto      RetTALAC            ; (no other flag is set)

;
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
                                        ; LA Cmd immed. action
                                        ; ====================
LAOut1xxx

;         btfss     xxxx,xxxxFlag       ; Test for output
;         goto      RetTALAC            ; No output: => Return without any output

                                        ;  #...

          goto      SetLAE_A_           ; Set EXE and ACK Flag, return


;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

SetLAE_A_ nop                           ; Finish command action with EXE and ACK flag set

          IF        __Addr=Yes          ; Wait for Cmd-Protocol to be finish
          call      SubInpExe           ; Wait for Cmd-tail

          movlw     ~Rx_Tail+1          ; Test for Cmd-tail (FEh)
          addwf     RxData,W            ; IF tail->FEh+2=00h->C ....
          movf      RxData,W
          bcf       ADCON0,DataRdy      ; Free RxProcess
          skpnc                         ; Was it Rx-Tail?
          goto      SLAE_A_             ;   Yes: set EXE and ACK

          movlw     Rx_Hdr
          subwf     RxData,W            ; IF Rx_Hdr->FFh-FFh=00h->C ....
          skpc                          ;   Yes: Leave Error, get Adress
          goto      RetTALAC            ;   No:  Return to Hdr

          movlw     00h                 ; Leave Error, read new Adress!
          movwf     Cmd
          goto      GetAdr              ;
          ENDIF

SLAE_A_   bsf       Cmd,Cmd_Exe         ; Set Exe Flag
          bsf       Cmd,Cmd_ACK         ; Set ACK Flag
          goto      RetTALAC

;
;

          ENDIF                         ; END of "Listen all"
;
;
;
;
;--------------------------------------------------------------------
;
; SubInpExe-Roundabout (Round Robin Scheduler)
; --------------------------------------------
; Tests several Flags and desires. If an event occurs, the according
; routine may be executed:
; If Rx data is ready return to Cmd-Process
; If Cmd_NACK is set return to Cmd-Process to send NACK and restart Cmd-Proc.
; If Cmd_ACK is set Tx process is loaded with ACK data
; If Tx process is free, last computed data is sent and new data is calcul.
; If Cmd_Test flag is set, a Debugger data is added to be sent

                                        ; Check data input, output read and executes
                                        ; waiting data output sub.
                                        ; ==========================================

SubInpExe btfsc     ADCON0,DataRdy      ; Test if data is received?
Ret1      return                        ; If DataRdy from Rx return to Cmd Execution

          btfsc     Cmd,Cmd_NACK        ; Is there an error pending (NACK)?
Ret2      return                        ; return to Cmd_Exe routines, data will be ejected!

          btfsc     Cmd,Cmd_ACK         ; Is there a ACK? Test before other data is sent
          call      SubRetACK           ; Sub to send ACK...

          btfsc     ADCON0,DataRdy      ; Test if data is received?
Ret3      return                        ; If DataRdy from Rx return to Cmd Execution

          btfsc     Cmd,Cmd_NACK        ; Is there a NACK?
Ret4      return                        ; return to Cmd_Exe routines, data will be ejected!

          btfss     PStat,TxFlag        ; Is TxProcess free?
          call      SubTstOut           ; computed data output and calc new data

          IF        __WDT=Yes
          clrwdt                        ; Reset watch dog timer
          ENDIF
          goto      SubInpExe           ; LOOP: New Data ready test


;
;--------------------------------------------------------------------
;
; SubRetACK
; ---------
;
;
;
                                        ; Return ACK data
                                        ; ===============
SubRetACK btfsc     Cmd,Cmd_NACK        ; Did NACK occur ?
          return                        ; Yes

          movlw     Tx_ACK              ; Load Tx process with ACK
          call      SubSendW            ; Send w via Tx
          bcf       Cmd,Cmd_ACK         ; Reset ACK-Flag
          return

;
;--------------------------------------------------------------------
;
; SubSendW(T)
; -----------
;
; Entrypoint SubSendWT checks T-Flag in Cmd. If set the value will be sent as
; converted ASCII value (2 Bytes, W1 used!).
; TxFlag is tested befor sending W
;

SubSendWT btfsc     Cmd,Cmd_Tst         ; If Cmd_Tst flag is set, send ASCII-HEX
          goto      SubSndWHx
SubSendW  btfss     PStat,TxFlag        ; Wait till Tx free
          goto      SSendW
                                        ; #
          IF        __WDT=Yes
           clrwdt                       ; Reset WDT
          ENDIF

          goto      SubSendW
SSendW    movwf     TxData
          bsf       PStat,TxFlag        ; Sets command to start TxProcess
          return

;
;--------------------------------------------------------------------
;
; SubSndWHx (uses W1 !)
; ---------
;
; TxFlag is tested befor sending W as two ASCII values
;

SubSndWHx movwf     W1                  ; push
          swapf     W1,W                ; Get high nibble
          call      SendHalf            ; Send high nibble
          movf      W1,W                ; Get low nibble
SendHalf  andlw     0Fh                 ; Clear high
          addlw     30h                 ; Add ASCII '0'
WaitFree  btfss     PStat,TxFlag        ; Wait till Tx free
          goto      HxFree

                                        ; # (Test other Flags...)

          IF        __WDT=Yes
           clrwdt                       ; Reset WDT
          ENDIF

          goto      WaitFree
HxFree    movwf     TxData              ; Send low nibble
          btfss     TxData,3            ; Is >=8 ?
          goto      SndHlf              ;   no: send
          btfsc     TxData,2            ; Is >=0Ch ?
          goto      AddHx               ;   Yes: add correction
          btfsc     TxData,1            ; Is >=0Ah ?
          goto      AddHx               ;   Yes: add correction
          goto      SndHlf              ;   Yes: send

AddHx     movlw     7
          addwf     TxData,F            ; Correct A-F
SndHlf    bsf       PStat,TxFlag        ; Sets command to start TxProcess
          return

;
;--------------------------------------------------------------------
;
; SubGetCRC
; ---------
;
; Gets and tests CR-Code. Sets C if not equal
;
SubGetCRCr nop
           IF        __Tst=Yes
           clrc                         ; No CRC
           ELSE
           call     SubInpExe           ; liest das nchste Byte ein
           movf     RxData,W            ; schreibe RxData ins W Register (low Byte)
           bcf      ADCON0,DataRdy      ; Free RxProcess
           btfss    Cmd,Cmd_NACK        ; Is there a NACK?
           goto     NoNACK              ; Restart CmdExecution

GCRCErr    setc
           return

NoNACK     subwf    CRCr,W              ; Are they equal ?
           skpz                         ;
           goto     GCRCErr             ;  No: error

           clrc                         ;  Yes: go ahead
           ENDIF
           return
;
;--------------------------------------------------------------------
;
; SubTestDataOut
; --------------
; This sub is called only if TxFlag is cleared (Line is open for sending any data)
;

                                        ; Test if any Cmd has to output data
                                        ; ==================================
SubTstOut btfss     Cmd,Cmd_Exe         ; Is Command ready to be executed?
          return                        ;  No: go back

          btfsc     Cmd,Cmd_ACK         ; Is ACK to be sent?
          return                        ;  Yes: go back

          btfsc     Cmd,Cmd_1xxx        ; Cmd 1xxx ?
          goto      Out1xxx             ;  Yes
          btfsc     Cmd,Cmd_01xx        ; Cmd 01xx?
          goto      Out01xx             ;  Yes
          btfsc     Cmd,Cmd_001x        ; Cmd 001x ?
          goto      Out001x             ;  Yes
          btfsc     Cmd,Cmd_0001        ; Cmd 0001 ?
          goto      Out0001             ;  Yes

          goto      TstExt_Out          ; Goto Extended commands
                                        ; # or
;         bsf       Cmd,Cmd_NACK        ; Set Syntax Error
;         return


;
;--------------------------------------------------------------------
;
; Execute-SubProcesses
; --------------------
; During SubInpExe-time with no Input this Exe-Subs will be started
; according to the loaded command

                                        ; Cmd 1xxx out
                                        ; ============
Out1xxx
                                        ; #
                                        ; (If there is no returning action (eg. data send)
                                        ; following the cmd, just add nop at this part.)

                                        ; Usage of CRCs
                                        ; =============
PutAData  clrf      CRCs                ; Clear CRC
                                        ; ...
          movlw     55h                 ; write a data (eg 55h) to W
          IF        __Tst=No
          addwf     CRCs,F		; Add sent value to CRCs
          ENDIF
          call      SubSendWT           ; Sende data according T-flag, adds value(s) to CRCs
                                        ; ...

SendCRCs  IF        __Tst=No            ; Send CRC after all data and before ACK:
          movf      CRCs,W              ; Get CRC into W
          call      SubSendW            ; Send CRC
          ENDIF


EndOut1   movlw     070h                ; End of Command (leave ACK,NACK & Tst)
          andwf     Cmd,F
          bsf       Cmd,Cmd_ACK         ; Send ACK at the End of the Commands
          goto      TstDbg

;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

                                        ; Cmd 01xx out
                                        ; ============
Out01xx
                                        ; #

EndOut01  movlw     070h                ; End of Command (leave ACK,NACK & Tst)
          andwf     Cmd,F
          bsf       Cmd,Cmd_ACK         ; Send ACK at the End of the Commands
          goto      TstDbg

;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

                                        ; Cmd 001x out
                                        ; ============
Out001x
                                        ; #

EndOut001 movlw     070h                ; End of Command (leave ACK,NACK & Tst)
          andwf     Cmd,F
          bsf       Cmd,Cmd_ACK         ; Send ACK at the End of the Commands
          goto      TstDbg

;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

                                        ; Cmd 0001 out
                                        ; ============
Out0001
                                        ; #

EndOt0001 movlw     070h                ; End of Command (leave ACK,NACK & Tst)
          andwf     Cmd,F
          bsf       Cmd,Cmd_ACK         ; Send ACK at the End of the Commands
          goto      TstDbg


;----------------------------------------------------------------------------


TstDbg    btfss     Cmd,Cmd_Tst         ; IF Test-Flag is set a debug routine is added after each cmd!
          return
                                        ; Debug routine:
                                        ; ==============

                                        ;  #

          IF        __Tst=Yes           ; For Test on a Terminal: Add "CR/LF"
          movlw     0Dh
          Call      SubSendW
          movlw     0Ah
          Call      SubSendW
          ENDIF

          return

;---------------------------------------------------------------------------

                                        ; Extended commands
                                        ; =================
                                        ; Do quit subroutine with return (not goto TstDbg)
TstExt_Out nop
          btfsc     W1,Ext_XXh          ; Bit_ ?
          goto      EOutXXh             ;  Yes

                                        ; #

                                        ; SyntaxError when Ext = 00h.
          bsf       Cmd,Cmd_NACK        ; Imaginary set NACK for Syntax error
          return

;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

                                        ; Ext Cmd output
                                        ; ==============
EOutXXh

                                        ; #

          movlw     070h                ; End of Command (leave ACK,NACK & Tst)
          andwf     Cmd,F
          bsf       Cmd,Cmd_ACK         ; Send ACK at the End of the Commands
          goto      TstDbgExt           ; or only TstDbg if it is the same...


;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

TstDbgExt btfss     Cmd,Cmd_Tst         ; IF Test-Flag is set a debug routine is added after each cmd!
          return
                                        ; Ext debug routine:
                                        ; ==================

                                        ; #

          IF        __Tst=Yes           ; For Test on a Terminal: Add "CR/LF"
          movlw     0Dh
          Call      SubSendW
          movlw     0Ah
          Call      SubSendW
          ENDIF

          return

;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
__LastInstr
          END

